#!/usr/bin/perl -w

##########################################################################
#                                                                        #
#  OpenFTPD Perl-Sitebot script v0.1 - (c) 2002 FEJF                     #
#                                                                        #
#  Written by FEJF with big thx to dngnand from POE-Team and Poing.      #
#  And of course topace who wrote the original sitebot.pl for openftpd.  #
#                                                                        #
#  fejf@gmx.de                                         www.openftpd.org  #
#                                                                        #
# ---------------------------------------------------------------------- #
#                                                                        #
#  This program is free software; you can redistribute it and/or modify  #
#  it under the terms of the GNU General Public License as published by  #
#  the Free Software Foundation; either version 2 of the License, or     #
#  (at your option) any later version.                                   #
#                                                                        #
##########################################################################


# TODO readregistry
# TODO more than 1 irc-server


# Specify these three variables if you want to change anything
my $DEFAULTDIRSTOSHOW	= 3;
my $MAXDIRSTOSHOW	= 10;
my $SHOWCOLORS		= 1;
my $VERSION		= "V0.1";

############ SHOULDNT HAVE TO CHANGE ANYTHING PAST HERE ##################
use strict; #always
use FindBin;
use lib "$FindBin::Bin/";
use File::Basename;
use primetools;
use Fcntl;
use POE;
use POE::Component::IRC;

# irc setup for sitebot
my $debug	= 0;
my $nick	= 'ttyS0wnz';
my $chan	= '#openftpd';
my $ircserver	= 'irc.efnet.pl';
my $chanpass	= ' ';
my $chanmodes	= ' ';

# This should locate your standarddir, either from F4A_STD or ~/.openftpdrc
if (exists $ENV{'F4A_STD'}) {
  my $standarddir = $ENV{'F4A_STD'};
}
elsif ( -e "$ENV{'HOME'}/.openftpdrc" ) {
  $standarddir = `cat $ENV{'HOME'}/.openftpdrc`;
  $standarddir =~ s/\n//;
  $standarddir =~ s/\r//;
}
else {
  print "ERROR: Cant locate your standarddir either in Environment F4A_STD or in $ENV{'HOME'}/.openftpdrc\n";
  exit 1;
}

$ENV{'F4A_COLOR'}="c";
# And change these if need be, but they should be ok
my $registry		= $standarddir."/etc/ftpd.reg";
my $pidfile		= $standarddir."/var/ftpd.pid";
my $wholog		= $standarddir."/var/who.log";
my $directorylog	= $standarddir."/var/log/directory.log";
my $sectioncfg		= $standarddir."/etc/sections.cfg";
my $request_file	= $standarddir."/var/request";
my $locatedbbin		= $standarddir."/bin/locatedb.pl";
my $affils_file		= $standarddir."/msg/affils";

#ok, so this is the lazy coders way of doing it, so sue me
my $shortsitenamereg	= `cat $registry |grep shortsitename`;
my $ftpbasedirreg	= `cat $registry |grep basedir`;
my $announcefile	= `cat $registry |grep announcefile`;
my $botpidfile		= `cat $registry |grep pidfile`;
my ($bla,$shortsitename,$ftpbasedir);
($bla,$shortsitename)=split(/\s+/,$shortsitenamereg,2);
($bla,$ftpbasedir)=split(/\s+/,$ftpbasedirreg,2);
($bla,$announcefile)=split(/\s+/,$announcefile,2);
($bla,$botpidfile)=split(/\s+/,$botpidfile,2);
if (!$shortsitename) {
	$shortsitename = "ABC";
}
else {
	chomp($shortsitename);
}
if ((!$ftpbasedir) or (!$announcefile) or (!$botpidfile)) {
	die "ftpbasedir, announcefile or botpidfile  not set correctly in your ftpd.reg";
}
chomp($ftpbasedir);
chomp($announcefile);
chomp($botpidfile);

# what dirs should i ignore on !site usage
my $usageignore="(PRiVATE|GROUPS|STAFF)";

# vars for irc-output
my $trigger		="!".$shortsitename;
my $maxcharsperline	= 450;

############ DEFINATELY SHOULDNT HAVE TO CHANGE ANYTHING PAST HERE ##################

# site_who
sub site_who {
	my ($pid, $dummy, @who);

	# open up the ftpd-pidfile
	open (PIDFILE, "< $pidfile") or die "cant open pid file";
	chomp ($pid=<PIDFILE>);
	close(PIDFILE);

	# kill the ftpd, to get the latest who.log
	kill ('USR2', $pid); 
	sleep(2);

	# open up the who.log, and push int into the hash
	open (WHOFILE, "< $wholog") or die "cant open who.log";
	while ($dummy = <WHOFILE>) 
	{
		chomp($dummy);
		if($dummy) {
			push(@who, $dummy);
		}
	}
	return (@who);
}

# do_* functions r called by the command-parser

#
# WHO Command for the ANYWHERE #
#
sub do_who {
	my $kernel = shift;
	my $to = shift;
	my ($dummy, $line, @who);
	my ($username, $groupname, $ontime, $acttime, $bytesdown, $bytesup, $activity, $speed, $path, $tagline, $someothernumber);

	@who = site_who();
	$dummy = sprintf("\[\002%s\002:\002WHO\002]-",$shortsitename);
	foreach $line (@who) {
		($username, $groupname, $ontime, $acttime, $bytesdown, $bytesup, $activity, $speed, $path, $tagline, $someothernumber) = split(/\t+/,$line,12);
		$activity =~ s/DOWNLOAD/DL/;
		$activity =~ s/UPLOAD/UL/;
		if($speed eq 0) {
			$dummy .= sprintf("[\002%s\002: %s] ",$username,$activity);
		}
		else {
			$dummy .= sprintf("[\002%s\002: %s at \002%2.1dkb/s\002] ",$username,$activity,$speed/1024);
		}
	}
	announce($kernel, $to, $dummy);
}

#
# STATS Command
#
sub do_stats {
	my $kernel = shift;
	my $to = shift;
	my $what = shift;
	my $section = shift;
	my $number = shift;
	my ($line, @statsoutput);
  
	@statsoutput=`$standarddir/bin/stats.pl $what $section $number`;
	foreach $line (@statsoutput) {
		$line = convertcolors($line);
		announce($kernel, $to, $line);
	}
}

#
# USER Command
#
sub do_user {
	my $kernel = shift;
	my $to = shift;
	my $user = shift;
	my ($dummy, $match, $line, $username, $groupname, $ontime, $acttime, @who);
	my ($bytesup, $bytesdown, $activity, $speed, $path, $tagline, $someothernumber);

	$dummy = sprintf("\[\002%s\002:\002USER\002]-",$shortsitename);
	$match=0;
	@who = site_who();
	foreach $line (@who) {
		($username, $groupname, $ontime, $acttime, $bytesup, $bytesdown, $activity, $speed, $path, $tagline, $someothernumber) = split(/\t+/,$line,12);
		$activity =~ s/DOWNLOAD/DL/;
		$activity =~ s/UPLOAD/UL/;
		if(uc($user) eq uc($username)) {
			$match=1;
			if($speed eq 0) {
				$dummy .= sprintf("[%s/%s: \002UL\002:%dmb \002DL\002:%dmb \002OnTime\002 :%s \002Current\002: %s] ",$username,$groupname,$bytesup/1048576,$bytesdown/1048576,$ontime,$activity);
			}
			else {
				$dummy .= sprintf("[%s/%s: \002UL\002:%dmb \002DL\002:%dmb \002OnTime\002 :%s \002Current\002: %s at \002%2.1dkb/s\002] ",$username,$groupname,$bytesup/1048576,$bytesdown/1048576,$ontime,$activity,$speed/1024);
			}
		}
	}
	if($match eq 0) {
		$dummy .= sprintf("\[\002%s\002 is not online]",$user);
	}
	announce($kernel, $to, "$dummy");
}

#
# BANDWIDTH
#
sub do_bw {
	my $kernel = shift;
	my $to = shift;
	my $countup=0;
	my $totalup=0;
	my $countdown=0;
	my $totaldown=0;
	my ($dummy, $line, $username, $groupname, $ontime, $acttime, $totalbandwidth, @who) = (0, 0, 0, 0, 0, 0, 0, 0);
	my ($bytesdown, $bytesup, $activity, $speed, $path, $tagline, $someothernumber, $total) = (0, 0, 0, 0, 0, 0, 0, 0);

	$dummy = sprintf("\[\002%s\002:\002BW\002]-",$shortsitename);
	@who = site_who();
	foreach $line (@who) {
		($username, $groupname, $ontime, $acttime, $bytesdown, $bytesup, $activity, $speed, $path, $tagline, $someothernumber) = split(/\t+/,$line,12);
		if($activity =~ /UL:/) {
			$countup++;
			$totalup+=$speed;
		}
		if($activity =~ /DL:/) {
			$countdown++;
			$totaldown+=$speed;
		}
		$total++;
	}
	$totalbandwidth=$totalup+$totaldown;
	$dummy .= sprintf ("[\002%d\002 UPLOADS at \002%1.1dkb/s\002 -- \002%d\002 DOWNLOADS at \002%1.1dkb/s\002 -- \002%d\002 TOTAL \002%1.1dkb/s\002]",$countup, $totalup/1024, $countdown, $totaldown/1024, $total, $totalbandwidth/1024);
	announce($kernel, $to, "$dummy");
} 

#
# REQUESTS
#
sub do_requests {
	my $kernel = shift;
	my $to = shift;
	my $number = shift;
	my ($numbertoshow, $count, $timestamp, $forusername, $forgroupname, $request, $msgstr, $dummy);

	if($number > $MAXDIRSTOSHOW ) {
		$numbertoshow=$MAXDIRSTOSHOW;
	}
	else {
		$numbertoshow=$number;
	}
 
	$count=1;
	$msgstr = sprintf("\[\002%s\002:\002REQUESTS\002]-",$shortsitename);
	if (-s $request_file) {
		if (open (REQFILE, "< $request_file")) {
			while (($count<=$numbertoshow) and ($dummy=<REQFILE>)) {
				chomp($dummy);
				($timestamp, $forusername, $forgroupname, $request) = split(/\s+/, $dummy, 4);
				$msgstr .= "[$count. \002$request\002 ($forusername/$forgroupname)] ";
				$count++;
			}
			close(REQFILE);
		}
	}
	announce($kernel, $to, "$msgstr");
} 



#
# NUKETOP/NUKERTOP Command
#
sub do_nuketop {
	my $kernel = shift;
	my $to = shift;
	my $what = shift;
	my $number = shift;
	my ($time, $section, $directory, $nuker, $nukergroup, $nukee, $nukeegroup, $line);
	my ($files, $size, $totalsize, $factor, $reason, $nukemsg, $num, %nuketop, %nukertop);

	$nukemsg=sprintf("\[\002%s\002:\002%s\002] ",$shortsitename,uc($what));
	open(NUKELOG, "< $standarddir/var/log/nukes.log") or die "Can't open nukelog!\n";
	while(<NUKELOG>) {
		chop($_);
		($time, $section, $directory, $nuker, $nukergroup, $nukee, $nukeegroup, $files, $size, $totalsize, $factor, $reason)=split(";",$_);
		if(exists $nuketop{$nukee}) {
			$nuketop{$nukee}{'size'}+=$size;
			$nuketop{$nukee}{'files'}+=$files;
		}
		else {
			$nuketop{$nukee}{'size'}=$size;
			$nuketop{$nukee}{'files'}=$files;
		}
		if(exists $nukertop{$nuker}) {
			$nukertop{$nuker}{'size'}+=$size;
			$nukertop{$nuker}{'files'}+=$files;
		}
		else {
			$nukertop{$nuker}{'size'}=$size;
			$nukertop{$nuker}{'files'}=$files;
		}
	}
	close(NUKELOG);
	$num=0;
	if($what eq "NUKETOP") {
		foreach $line (sort {$nuketop{$b}{'size'} <=> $nuketop{$a}{'size'}}(keys %nuketop)) {
			$nukemsg.="[ Nukee #".++$num.": \002$line\002 Files: \002$nuketop{$line}{'files'}\002 Size: \002". filesize_to_string($nuketop{$line}{'size'})."\002 ] "; 
			if($num==$number) {
				last;
			}
		}
	}
	else {
		foreach $line (sort {$nuketop{$b}{'size'} <=> $nuketop{$a}{'size'}}(keys %nukertop)) {
			$nukemsg.="[ Nuker #".++$num.": \002$line\002 Files: \002$nukertop{$line}{'files'}\002 Size: \002". filesize_to_string($nukertop{$line}{'size'})."\002 ] ";
			if($num == $number) {
				last;
			}
		}
	}
	announce($kernel, $to, "$nukemsg");
}

#
# NEW Command
#
sub do_new {
	my $kernel = shift;
	my $to = shift;
	my $sect = shift;
	my $number = shift;
	my ($dummy, $x, $last, $finalstr, @dirlog, $NUMBERTOSHOW);
	my ($date, $time, $username, $groupname, $section, $path);

	$dummy = sprintf("\[\002%s\002:\002NEW\002]-",$shortsitename);
	if($number > $MAXDIRSTOSHOW ) {
		$NUMBERTOSHOW=$MAXDIRSTOSHOW;
	}
	else {
		$NUMBERTOSHOW=$number;
	}

	$x=0;
	open (DIRLOG, "< $directorylog") or die "cant open directorylog";
	while (<DIRLOG>) {
		chomp($_);
		if($_) {
			if($_ !~ /[Cc][Dd][0-9]/ && $_ !~ /(sample|sbc-stats|stats$|ac3-codec)/i ) {
				push(@dirlog, $_);
			}
		}
	}
	close(DIRLOG);
	$sect=uc($sect);

	while($x < $NUMBERTOSHOW) {
		$last=pop(@dirlog);
		if(!$last) {
			$x=$NUMBERTOSHOW;
			next;
		}
		($date, $time, $username, $groupname, $section, $path) = split(/\s+/,$last,6);
		if($sect ne "ALL") {
			if(uc($section) eq $sect) {
				$x++;
				$finalstr=$finalstr . " \[\002$x.\002 " . basename($path) . "]";
			}
		}
		else {
			$x++;
			$finalstr=$finalstr . " \[\002$x.\002 " . basename($path) . "]";
		}
	}
	if(!$finalstr) {
		announce($kernel, $to, "NO releases in section: $sect");
	}
	else {
		$dummy .= sprintf("\[\002%s\002]-%s",$sect,$finalstr);
		$x++;
	}
	announce($kernel, $to, "$dummy");
}

#
# USAGE Command
#
sub do_usage {
	my $kernel = shift;
	my $to = shift;
	my ($dummy, $usage, $sec, @sections);
	my ($secname,$seccfg,$secstat,$secpath);
  
	$dummy = sprintf ("\[\002%s\002:\002USAGE\002]-",$shortsitename);

# now open the sections.cfg file, and grab all the path names of active sections
	open(SECCFG,"< $sectioncfg") or die "ERROR: unable to open sectioncfg";
	while(<SECCFG>) {
		if($_ !~ /[#].*/) {
			chomp($_);
			push(@sections,$_);
		}
	}
	close(SECCFG);
	foreach my $sec (@sections) {
		($secname,$seccfg,$secstat,$secpath) = split(/\s+/,$sec,4);
		if($secpath=~/$usageignore/) {
			next;
		}
		$usage=`du -sk $ftpbasedir$secpath`;
		($usage)=($usage=~/^(\d+)\s+.*/);
		if($usage>1024) {
			$usage=int $usage/1024;
			$dummy.= "[$secpath: \002$usage\002 MB] ";
		}
		else {
			$dummy.= "[$secpath: \002$usage\002 KB] ";
		}
	}
	announce($kernel, $to, "$dummy");
}

#
# DF Command
#
sub do_df {
	my $kernel = shift;
	my $to = shift;
	my ($dummy, $outstr, $line, @sections, @lines);

#If you use linux, use these lines.
	my $df=`df -m`;
	my $dfbase=`df -m $ftpbasedir`;

#	if you use OpenBSD, use these lines
#	$df=`df -h`;
#	$dfbase=`df -h $ftpbasedir`;

	my @baselines = split(/\n/, $dfbase);
	my ($dev,$total,$used,$free,$percent,$where)=split(/\s+/,$baselines[1],6);
	my $basefree=$free;

	$dummy = sprintf ("\[\002%s\002:\002DF\002]-",$shortsitename);

# now open the sections.cfg file, and grab all the path names of active sections
	open(SECCFG,"< $sectioncfg") or die "ERROR: unable to open sectioncfg";
	while(<SECCFG>) {
		if($_ =~ /[#].*/) {
			#ignore cuz its a comment
		}
		else {
			chomp($_);
			push(@sections,$_);
		}
	}
	close(SECCFG);

	@lines = split(/\n/, $df);
	foreach my $sec (@sections) {
		my ($secname,$seccfg,$secstat,$secpath) = split(/\s+/,$sec,4);
		my $matched=0;
		foreach $line (@lines) {
			($dev,$total,$used,$free,$percent,$where)=split(/\s+/,$line,6);
			if($where =~ /$secpath/) {
				$outstr = $outstr . "\[$secname: \002$free\002 MB] ";
				$matched=1;
			}
		}
		if($matched eq 0) {
			$outstr = $outstr . "\[$secname: \002$basefree\002 MB] ";
		}
	}
	announce($kernel, $to, "$outstr");
}

#
# SEARCH Command
#
sub do_search {
	my $kernel = shift;
	my $to = shift;
	my $searchstring = shift;
	my $section = shift;
	my ($dummy, $command, @result);

	announce($kernel, $to, "---------------------------------------------------------------------------");
	announce($kernel, $to, " $shortsitename SEARCH for $searchstring - OpenFTPD Sitebot");
	announce($kernel, $to, "---------------------------------------------------------------------------");
	$command = "$locatedbbin $searchstring $section";

	@result=`$command`;
	foreach $dummy (@result) {
		chomp($dummy);
		announce($kernel, $to, $dummy);
	}
	announce($kernel, $to, "---------------------------------------------------------------------------");
}

#
# BNC Command
#
sub do_bnc  {
	my $kernel = shift;
	my $to = shift;
	my ($dummy, @result);

	@result=`$standarddir/bin/bnctest.pl`;
	foreach $dummy (@result) {
		chomp($dummy);
		announce($kernel, $to, $dummy);
	}
}

#
# AFFILS
#
sub do_affils {
	my $kernel = shift;
	my $to = shift;

	if (-f $affils_file) {
		open (AFFILFILE, "<$affils_file");
		while(<AFFILFILE>) {
			chomp($_);
			announce($kernel, $to, $_ );
		}
		close AFFILFILE;
	}
	else {
		announce($kernel, $to, "Affils file not found error");
	}
}

#
# HELP Command
#
sub do_help {
	my $kernel = shift;
	my $to = shift;
	my ($secname,$seccfg,$secstat,$secpath, $sections);

	$sections = " Sections: ";
	open(SECCFG,"< $sectioncfg") or die "ERROR: unable to open sectioncfg";
	while(<SECCFG>) {
		if($_ =~ /[#].*/) {
			#ignore cuz its a comment
		}
		else {
			chomp($_);
			($secname,$seccfg,$secstat,$secpath) = split(/\s+/,$_,4);
			$sections.= "$secname ";
		}
	}
	close(SECCFG);
	announce($kernel, $to, "---------------------------------------------------------------------------");
	announce($kernel, $to, " $shortsitename HELP - OpenFTPD Sitebot");
	announce($kernel, $to, "---------------------------------------------------------------------------");
	announce($kernel, $to, " Command                          Description");
	announce($kernel, $to, "---------------------------------------------------------------------------");
	announce($kernel, $to, " HELP                             This help");
	announce($kernel, $to, " WHO                              WHO is online");
	announce($kernel, $to, " DF                               Disk usage information for each section");
	announce($kernel, $to, " BNC                              Shows BNC-Status");
	announce($kernel, $to, " USAGE                            Shows discusage of sections");
	announce($kernel, $to, " AFFILS                           Shows affils of your site");
	announce($kernel, $to, " BANDWIDTH or BW                  Current site bandwidth usage");
	announce($kernel, $to, " NUKETOP [NUMBER]                 Display NUMBER topnukees");
	announce($kernel, $to, " NUKERTOP [NUMBER]                Display NUMBER topnukers");
	announce($kernel, $to, " NEW [SECTION] [NUMBER]           NUMBER New Directories in SECTION");
	announce($kernel, $to, " USER <USERNAME>                  Information about USERNAME");
	announce($kernel, $to, " REQUESTS [NUM] or REQS [NUM]     Shows current requests (request 1 to NUM)");
	announce($kernel, $to, " SEARCH <WHAT> [SECTION]          Search for dirs on your ftp");
	announce($kernel, $to, " STATS <WHAT> [SECTION] [NUMBER]  NUMBER lines of stats from SECTION,");
	announce($kernel, $to, "                                  WHAT is defined as <u|g><d|w|m|y|t><u|d>");
	announce($kernel, $to, "---------------------------------------------------------------------------");
	announce($kernel, $to, $sections);
	announce($kernel, $to, "---------------------------------------------------------------------------");

}

#####################
# now the irc-stuff #
#####################

### announce( $kernel, $to, $msg );
sub announce {
	my $kernel = shift;
	my $to = shift;
	my $msg = shift;
	my ($dummy, $output, @parts);

	if (length $msg > $maxcharsperline) {
		$output="";
		@parts = split (/ /,$msg);		# TODO fix the splitting e.g. does not work correctly for !sitebnc !!!
		while($dummy = shift @parts) {
			if (length("$output.$dummy")>$maxcharsperline) {
				if (($output) and ($output !~ /^\s*$/)) {
					$kernel->post( 'openftpd', 'privmsg', $to, $output);
				}
				$output = $dummy;
			}
			else {
				$output .= " $dummy";
			}
		}
		if (($output) and ($output !~ /^\s*$/)) {
			$kernel->post( 'openftpd', 'privmsg', $to, $output);
		}
	}
	else {
		$kernel->post( 'openftpd', 'privmsg', $to, $msg);
	}
	# make IRC.pm happy !
	return 1;
}


### parse_line( $kernel, $who_did_it, $where_to_send_the_reply, $msg);
sub parse_line {
	my $kernel = shift;
	my $who = shift;
	my $to = shift;
	my $msg = shift;

	if($msg =~ /^$trigger {0,1}help/i) {
		do_help($kernel, $who);
	}
	if($msg =~ /^$trigger {0,1}who/i) {
		do_who($kernel, $to);
	}
	if($msg =~ /^$trigger {0,1}df/i) {
		do_df($kernel, $to);
	}
	if($msg =~ /^$trigger {0,1}bnc/i) {
		do_bnc($kernel, $to);
	}
	if($msg =~ /^$trigger {0,1}usage/i) {
		do_usage($kernel, $to);
	}
	if($msg =~ /^$trigger {0,1}affils/i) {
		do_affils($kernel, $to);
	}
	if($msg =~ /^$trigger {0,1}bw/i) {
		do_bw($kernel, $to);
	}
	if($msg =~ /^$trigger {0,1}nuketop/i) {
		if($msg =~ /^$trigger {0,1}nuketop (\d+)/i) {
			do_nuketop($kernel, $to, "NUKETOP", $1);
		}
		else {
			do_nuketop($kernel, $to, "NUKETOP", $DEFAULTDIRSTOSHOW);
		}
	}
	if($msg =~ /^$trigger {0,1}nukertop/i) {
		if($msg =~ /^$trigger {0,1}nukertop (\d+)/i) {
			do_nuketop($kernel, $to, "NUKERTOP", $1);
		}
		else {
			do_nuketop($kernel, $to, "NUKERTOP", $DEFAULTDIRSTOSHOW);
		}
	}
	if($msg =~ /^$trigger {0,1}new/i) {
		if($msg =~ /^$trigger {0,1}new (\S+) (\d+)/i) {
			do_new($kernel, $to, $1, $2);
		}
		elsif($msg =~ /^$trigger {0,1}new (\S+)/i) {
			do_new($kernel, $to, $1, $DEFAULTDIRSTOSHOW);
		}
		else {
			do_new($kernel, $to, "ALL", $DEFAULTDIRSTOSHOW);
		}
	}
	if($msg =~ /^$trigger {0,1}(user|speed)/i) {
		if($msg =~ /^$trigger {0,1}(user|speed) (\S+)/i) {
			do_user($kernel, $to, $2);
		}
		else {
			announce($kernel, $to, "\002USAGE\002: $trigger user <USERNAME>");
		}
	}
	if($msg =~ /^$trigger {0,1}(requests|reqs)/i) {
		if($msg =~ /^$trigger {0,1}(requests|reqs) (\d+)/i) {
			do_requests($kernel, $to, $2);
		}
		else {
			do_requests($kernel, $to, $DEFAULTDIRSTOSHOW);
		}
	}
	if($msg =~ /^$trigger {0,1}search/i) {
		if($msg =~ /^$trigger {0,1}search (\S+) (\S+)/i) {
			do_search($kernel, $who, $1, $2);
		}
		elsif($msg =~ /^$trigger {0,1}search (\S+)/i) {
			do_search($kernel, $who, $1, " ");
		}
		else {
			announce($kernel, $to, "\002USAGE\002: $trigger search <WHAT>");
		}
	}
	if($msg =~ /^$trigger {0,1}stats/i) {
		if($msg =~ /^$trigger {0,1}stats (\S+) (\S+) (\d+)/i) {
			do_stats($kernel, $who, $1, $2, $3);
		}
		elsif($msg =~ /^$trigger {0,1}stats (\S+) (\S+)/i) {
			do_stats($kernel, $who, $1, $2, $DEFAULTDIRSTOSHOW);
		}
		elsif($msg =~ /^$trigger {0,1}stats (\S+)/i) {
			do_stats($kernel, $who, $1, "global", $DEFAULTDIRSTOSHOW);
		}
		else {
			announce($kernel, $to, "\002USAGE\002: $trigger stat <WHAT> [SECTiON] [NUMBER]");
		}
	}
}

sub _signal {
	my ($kernel, $sig) = @_[KERNEL, ARG0];
	my ($dummy);

	if (-e $announcefile) {
		sysopen(ANCFILE, $announcefile, O_RDONLY) or return 0;
		# LOCK_EX = 2
		flock(ANCFILE, 2) or return 0;
		while ($dummy = <ANCFILE>) {
			if ($dummy =~ /!$/) {
				next;
			}
			if($debug) {
				print "(II) Writing to $chan : $dummy\n";
			}
			announce($kernel, "$chan", "$dummy");
		}
		# LOCK_UN = 8
		flock(ANCFILE, 8);
		close(ANCFILE);
		unlink $announcefile;
	}
	if ($sig eq "INT") {
		announce($kernel, "$chan", "quit requested by SIGINT");
		$kernel->call( 'openftpd', 'quit', 'Someone does not like me...' );
		die "(II) got SIGINT -> exiting";
	}
	# make the signal handler happy !
	return 1;
}

sub _start {
	my ($kernel) = $_[KERNEL];

	$kernel->alias_set( 'sitebot' );
	$kernel->post( 'openftpd', 'register', 'all');
	$kernel->post( 'openftpd', 'connect', { Debug    => $debug,
						Nick     => $nick,
						Server   => $ircserver ||
						            'irc.efnet.pl',
						Port     => $ARGV[1] || 6667,
						Username => 'openftpd',
						Ircname  => "OpenFTPD Sitebot $VERSION by FEJF",
						}
	);
	$kernel->sig( HUP => '_signal');
	$kernel->sig( INT => '_signal');
	open(PIDFILE, ">$botpidfile") or die "Can't open $botpidfile";
	print PIDFILE $$;
	close(PIDFILE);
	print "OpenFTPD Perl Sitebot $VERSION ready to serve $shortsitename.\n\nTrying to connect to the net.\n";
}

sub irc_001 {
	my ($kernel) = $_[KERNEL];

	$kernel->post( 'openftpd', 'join', "$chan", "$chanpass" );
	$kernel->post( 'openftpd', 'mode', "$chan", "$chanmodes" );
	$kernel->post( 'openftpd', 'privmsg', "$chan", "OpenFTPD Perl Sitebot $VERSION ready to serve $shortsitename." );
#	$kernel->post( 'openftpd', 'topic', "$topic" );
}

sub irc_disconnected {
	my ($server) = $_[ARG0];

	print "Lost connection to server $server.\n";
}

sub irc_error {
	my $err = $_[ARG0];

	print "Server error occurred! $err\n";
}

sub irc_socketerr {
	my $err = $_[ARG0];

	print "Couldn't connect to server: $err\n";
}

sub _stop {
	my ($kernel) = $_[KERNEL];

	print "Control session stopped.\n";
	$kernel->call( 'openftpd', 'quit', 'Someone does not like me...' );
	unlink($botpidfile);
}

sub irc_public {
	my ($kernel, $who, $chan, $msg) = @_[KERNEL, ARG0 .. ARG2];

	$who =~ /^(.+)!/ and $who = $1;
	if($debug) {
		print "(II) irc_public: who $who, chan $chan, msg $msg\n";
	}
	parse_line($kernel, $who, $chan, $msg);
}

sub irc_notice {
	my ($kernel, $who, $to, $msg) = @_[KERNEL, ARG0 .. ARG2];

	$who =~ /^(.+)!/ and $who = $1;
	if($debug) {
		print "(II) irc_notice who $who, msg $msg\n";
	}
	parse_line($kernel, $who, $who, $msg);
}

sub irc_msg {
	my ($kernel, $who, $to, $msg) = @_[KERNEL, ARG0 .. ARG2];

	$who =~ /^(.+)!/ and $who = $1;
	if($debug) {
		print "(II) irc_msg who $who, msg $msg\n";
	}
	parse_line($kernel, $who, $who, $msg);
}


POE::Component::IRC->new( 'openftpd' ) or die "Can't instantiate new IRC component!\n";
POE::Session->new( 'main' => [qw(_start _stop _signal irc_001 irc_disconnected irc_socketerr irc_error irc_public irc_notice irc_msg)] );

$poe_kernel->run();
