/*
    $Id: cdpath.c,v 1.3 2001/03/01 02:10:08 hvr Exp $

    Copyright (C) 2000 Herbert Valerio Riedel <hvr@gnu.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

static const char _rcsid[] = "$Id: cdpath.c,v 1.3 2001/03/01 02:10:08 hvr Exp $";

#include "cdpath.h"

#include <string.h>

#include <glib.h>

static GSList *cdpath = NULL;

typedef struct {
  gchar *pathname;
  gint32 uid;
  gint32 gid;
  gint32 perm;
} cdpath_data_t;

void
cdpath_reset (void)
{
  g_assert (cdpath == NULL);
  /* fixme */
}

static cdpath_data_t *
_lookup (const gchar pathname[])
{
  GSList *p;

  g_return_val_if_fail (pathname != NULL, NULL);

  for (p = cdpath; p; p = g_slist_next (p))
    {
      cdpath_data_t *d = p->data;

      if (!strcmp (pathname, d->pathname))
	return d;
    }
  
  return NULL;
}

void 
cdpath_set (const gchar pathname[], gint32 uid, gint32 gid, guint32 perm)
{
  cdpath_data_t *data;

  data = _lookup (pathname);

  /* g_printerr ("debug set: %s %d %d %x\n", pathname, uid, gid, perm); */

  if (data) /* overwrite */
    g_printerr ("cdpath_set: warning, overwritting entry for '%s'\n", 
		data->pathname);
  else
    {
      data = g_new0 (cdpath_data_t, 1);

      data->pathname = g_strdup (pathname);

      cdpath = g_slist_prepend (cdpath, data);
    }

  data->uid = uid;
  data->gid = gid;
  data->perm = perm;

}

gboolean
cdpath_get (const gchar pathname[], gint32 *uid, gint32 *gid, guint32 *perm)
{
  cdpath_data_t *data;

  if ((data = _lookup (pathname))) 
    {
      if (uid)
	*uid = data->uid;
      
      if (gid)
	*gid = data->gid;

      if (perm)
	*perm = data->perm;
      
      return TRUE;
    }

  return FALSE;
}

const gchar *
cdpath_match (const gchar pathname[])
{
  GSList *p;
  gchar *result = NULL;

  g_return_val_if_fail (pathname != NULL, NULL);

  for (p = cdpath; p; p = g_slist_next (p))
    {
      cdpath_data_t *d = p->data;

      /* match against entry in cdpath db */
      if (!strncmp (pathname, d->pathname, strlen (d->pathname)))
	{
	  /* replace last match, if this is longer */
	  if (result == NULL || strlen (d->pathname) > strlen (result))
	    result = d->pathname;
	}
    }

  /* g_printerr ("debug match: %s %s\n", pathname, result); */
  
  return result;
}


