static const char _rcsid[] = "$Id: permfile.c,v 1.3.2.1 2002/01/03 15:36:18 soulshock Exp $";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/stat.h>

#include <glib.h>

#include "pathutil.h"
#include "permfile.h"
#include "cdpath.h"
#include "site_vfs.h"
#include "strlcpy.h"
#include "strlcat.h"

/* fucking globals... let's get rid of them asap */
extern gchar *permission_file;
extern gchar *readme_file;
extern gchar *hidden_files;

/* fixme -- replace this one... with something better */
static gint
read_cdpath (const gchar *dir, gint32 *uid, gint32 *gid, gint32 *perm)
{
  gchar *_path;

  /* g_printerr ("read_cdpath: %s %s\n", cwd, dir); */

  if ((_path = path_add (vfs_getcwd (), dir)))
    {
      const gchar *tmp = cdpath_match (_path);
      g_free (_path);

      if (tmp)
	return cdpath_get (tmp, uid, gid, perm) ? 0 : -1;
    }

  return -1;
}

gint32 
dirperm_to_fileperm (gint32 perm)
{
  gint i;
  gint32 p = 0x01000000;

  for (i = 0; i <= 2; i++)
    {
      if ((perm >> ((2 - i) * 8)) & PE_READ)
	p |= FPE_READ << (2 - i) * 8;
      if ((perm >> ((2 - i) * 8)) & PE_OVERWRITE)
	p |= FPE_WRITE << (2 - i) * 8;
      if (((perm >> ((2 - i) * 8)) & PE_CHMOD) && !i)
	p |= FPE_CHMOD << (2 - i) * 8;
    }
  return p;
}

int
write_permfile (char *filename, char **buffer, int n)
{
  FILE *file;

  if ((file = fopen (filename, "wb")) == NULL)
    {
      g_free (*buffer);
      return -1;
    }
  n = (n * sizeof (struct file_s)) + sizeof (struct perm_s);
  if (fwrite (*buffer, 1, n, file) != n)
    {
      g_free (*buffer);
      fclose (file);
      return -1;
    }
  g_free (*buffer);
  fclose (file);
  return 0;
}

struct file_s *
find_name (struct file_s *buffer, char *name, int n, int *found)
{
  int i = 0;
  int j;
  int k = n - 1;

  while (i <= k)
    {
      j = (i + k) >> 1;
      n = strcmp (name, (buffer + j)->name);

      if (n < 0)
	k = j - 1;
      else if (n > 0)
	i = j + 1;
      else
	{
	  *found = 1;
	  return buffer + j;
	}
    }

  *found = 0;

  return buffer + i;
}

static FILE *
file_open_nocheck (const gchar *dir, gchar * filename, const gchar * mode)
{
  char wd[256];

  if (add_dir (vfs_getcwd (), dir, wd, sizeof (wd)) == -1)
    return NULL;

  if (strlen (wd) + strlen (filename) + 1 >= sizeof (wd))
    {
      errno = ENAMETOOLONG;
      return NULL;
    }

  if (strcmp (wd, "/"))
    strlcat(wd, "/", sizeof(wd));

  strlcat(wd, filename, sizeof(wd));

  return fopen (wd + 1, mode);
}


int
read_permfile (char *filename, char **buffer, int *n, guint32 umask_root)
{
  FILE *file;
  struct perm_s *pe;
  struct stat st;

  if (stat (filename + 1, &st) == -1)
    st.st_size = 0;
  if (st.st_size < sizeof (struct perm_s))
    {
      if ((*buffer = g_malloc (sizeof (struct perm_s) +
			       sizeof (struct file_s))) == NULL)
	return -1;
      pe = (struct perm_s *) *buffer;
      if ((file = fopen (filename + 1, "rt")) != NULL)
	{
	  if (fscanf (file, SI32 " " SI32 " " SI32, &pe->uid, &pe->gid, &pe->perm) != 3)
	    {
	      g_free (*buffer);
	      fclose (file);
	      return -1;
	    }
	  fclose (file);
	}
      else
	{
	  pe->uid = pe->gid = -1;
	  pe->perm = umask_root;
	}
      pe->dlc = pe->dlct = 0;
      memset (&pe->reserved, 0, sizeof (pe->reserved));
      *n = 0;
    }
  else
    {
      if ((*buffer = g_malloc (st.st_size + sizeof (struct file_s))) == NULL)
	{
	  return -1;
	}
      if ((file = fopen (filename + 1, "rb")) == NULL)
	{
	  g_free (*buffer);
	  return -1;
	}
      if (fread (*buffer, 1, st.st_size, file) != st.st_size)
	{
	  g_free (*buffer);
	  fclose (file);
	  return -1;
	}
      fclose (file);
      *n = (st.st_size - sizeof (struct perm_s)) / sizeof (struct file_s);
    }
  return 0;
}

int
read_permission_file (const gchar *dir, gint32 *uid, gint32 *gid, gint32 *perm,
		      gint32 *dlc, gchar **buffer, gint *n)
{
  FILE *file;
  char line[32];
  struct perm_s p;
  struct stat st;

  if (dlc)
    *dlc = 1;
  if (buffer)
    *buffer = NULL;
  if (n)
    *n = 0;
  if (read_cdpath (dir, uid, gid, perm) == 0)
    return 0;
  if ((file = file_open_nocheck (dir, permission_file, "rb")) == NULL)
    return -1;
  if (fstat (fileno (file), &st) == -1)
    {
      fclose (file);
      return -1;
    }
  if (st.st_size < sizeof (struct perm_s))
    {
      if (buffer && n)
	{
	  *n = 0;
	  if ((*buffer = g_malloc (sizeof (struct file_s))) == NULL)
	    {
	      fclose (file);
	      return -1;
	    }
	}
      if (fread (line, 1, st.st_size, file) != st.st_size)
	{
	  if (buffer)
	    g_free (*buffer);
	  fclose (file);
	  return -1;
	}
      if (sscanf (line, SI32 " " SI32 " " SI32, uid, gid, perm) != 3)
	{
	  if (buffer)
	    g_free (*buffer);
	  fclose (file);
	  return -1;
	}
      if (dlc)
	*dlc = 0;
    }
  else
    {
      if (!fread (&p, sizeof (p), 1, file))
	{
	  fclose (file);
	  return -1;
	}
      *uid = p.uid;
      *gid = p.gid;
      *perm = p.perm;
      if (dlc)
	*dlc = p.dlc;
      if (buffer && n)
	{
	  st.st_size -= sizeof (p);
	  if ((*buffer = g_malloc (st.st_size + sizeof (struct file_s))) ==
	      NULL)
	    {
	      fclose (file);
	      return -1;
	    }
	  if (fread (*buffer, 1, st.st_size, file) != st.st_size)
	    {
	      g_free (*buffer);
	      fclose (file);
	      return -1;
	    }
	  *n = st.st_size / sizeof (struct file_s);
	}
    }
  fclose (file);
  return 0;
}

int
read_file_permission (const char *dir, char *name, gint32 * uid, gint32 * gid,
		      gint32 * perm)
{
  char *buffer;
  struct file_s *f;
  int found;
  int n;

  if (read_permission_file (dir, uid, gid, perm, NULL, &buffer, &n) == -1)
    {
      errno = EACCES;
      return -1;
    }
  f = find_name ((struct file_s *) buffer, name, n, &found);
  if (found)
    {
      *uid = f->uid;
      *gid = f->gid;
      if (*perm)
	*perm = f->perm;
      else
	*perm = dirperm_to_fileperm (*perm);
    }
  else
    *perm = dirperm_to_fileperm (*perm);
  g_free (buffer);
  return 0;
}

int
remove_permission_file (char *dir)
{
  char wd[256];
  char *hfile;
  char str[256];

  if (strlen (dir) + strlen (permission_file) >= sizeof (wd))
    {
      errno = ENAMETOOLONG;
      return -1;
    }
  if (strcmp (dir, "/"))
    g_snprintf (wd, sizeof (wd), "%s/%s", dir + 1, permission_file);
  else
    strlcpy (wd, permission_file, sizeof (wd));
  unlink (wd);

  if (readme_file)
    {
      if (strcmp (dir, "/"))
	g_snprintf (wd, sizeof (wd), "%s/%s", dir + 1, readme_file);
      else
	strlcpy (wd, readme_file, sizeof (wd));
      unlink (wd);
    }

  /* added by primemover */
  strlcpy (str, hidden_files, sizeof (str));
  hfile = strtok (str, " \t");
  while (hfile != NULL)
    {
      if (strcmp (dir, "/"))
	g_snprintf (wd, sizeof (wd), "%s/%s", dir + 1, hfile);
      else
	strlcpy (wd, hfile, sizeof (wd));
      unlink (wd);

      hfile = strtok (NULL, " \t");
    }
  return 0;
}

int
write_permission_file (char *dir, gint32 uid, gint32 gid, gint32 perm)
{
  int e;
  FILE *file;
  struct perm_s p;
  struct stat st;

  if (path_in_cdpath (dir))
    return 0;
  if ((file = file_open_nocheck (dir, permission_file, "r+b")) == NULL)
    if ((file = file_open_nocheck (dir, permission_file, "w+b")) == NULL)
      return -1;
  if (fstat (fileno (file), &st) == -1)
    st.st_size = 0;
  p.uid = uid;
  p.gid = gid;
  p.perm = perm;
  if (st.st_size < sizeof (p))
    {
      p.dlc = p.dlct = 0;
      memset (&p.reserved, 0, sizeof (p.reserved));
    }
  if (!fwrite
      (&p, st.st_size < sizeof (p) ? sizeof (p) : sizeof (gint32) * 3, 1,
       file))
    {
      e = errno;
      fclose (file);
      errno = e;
      return -1;
    }
  fclose (file);
  return 0;
}
