/* ---------------------------------------------------------------------------
 *
 *   This file is part of OpenFTPD (based on FTP4ALL)
 *   (C) 1996-2000 by Crescent, (C) 2000 by the OpenFTPD Team
 *
 * ---------------------------------------------------------------------------
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 * ---------------------------------------------------------------------------
 *   HEADER: definesd.h - important structs and defines
 * ---------------------------------------------------------------------------
 */

#ifndef __DEFINESD_H_
#define __DEFINESD_H_

#include <glib.h>
#include <time.h>
#include <unistd.h>
#include <stdio.h>
#include "commondef.h"

struct user_v3012_s
{ gint32     uid;            /* user ID                              */
  gint32     gid;            /* group ID                             */
  gint32     gids[8];        /* additional groups                    */
  gint32     class;          /* user class                           */
  gint64     rights;         /* user rights                          */
  gint32     limit;          /* limit (ratio)                        */
  gint64     credit;         /* credit in bytes                      */
  gint32     dl_speed;       /* min and max download speed in kb/s   */
  gint32     ul_speed;       /* min and max upload speed   in kb/s   */
  gint64     dl_week;        /* download current week                */
  gint64     ul_week;        /* upload current week                  */
  gint64     dl_month;       /* download current month               */
  gint64     ul_month;       /* upload current month                 */
  gint64     dl_year;        /* download current year                */
  gint64     ul_year;        /* upload current year                  */
  gint64     dl_total;       /* download total                       */
  gint64     ul_total;       /* upload total                         */
  gint64     dl_tm_week;     /* download time (s) current week       */
  gint64     ul_tm_week;     /* upload time (s) current week         */
  gint64     dl_tm_month;    /* download time (s) current month      */
  gint64     ul_tm_month;    /* upload time (s) current month        */
  gint64     dl_tm_year;     /* download time (s) current year       */
  gint64     ul_tm_year;     /* upload time (s) current year         */
  gint64     dl_tm_total;    /* download time (s) total              */
  gint64     ul_tm_total;    /* upload time (s) total                */
  time_t    last_login;     /* time of last login                   */
  char      name[10];       /* username                             */
  char      password[40];   /* encrypted password                   */
  char      info[32];       /* info                                 */
  char      idents[10][20]; /* 12bytes for ident; 8byte for iprange */
} ;

struct user_beta29_s
{ 
   gint32     uid;            /* user ID                              */
   gint32     gid;            /* group ID                             */
   gint32     gids[8];        /* additional groups                    */
   gint32     class;          /* user class                           */
   gint64     rights;         /* user rights                          */
   gint32     limit;          /* limit (ratio)                        */
   gint64     credit;         /* credit in bytes                      */
   gint32     dl_speed;       /* min and max download speed in kb/s   */
   gint32     ul_speed;       /* min and max upload speed   in kb/s   */
   gint64     dl_week;        /* download current week                */
   gint64     ul_week;        /* upload current week                  */
   gint64     dl_month;       /* download current month               */
   gint64     ul_month;       /* upload current month                 */
   gint64     dl_year;        /* download current year                */
   gint64     ul_year;        /* upload current year                  */
   gint64     dl_total;       /* download total                       */
   gint64     ul_total;       /* upload total                         */
   gint64     dl_tm_week;     /* download time (s) current week       */
   gint64     ul_tm_week;     /* upload time (s) current week         */
   gint64     dl_tm_month;    /* download time (s) current month      */
   gint64     ul_tm_month;    /* upload time (s) current month        */  
   gint64     dl_tm_year;     /* download time (s) current year       */
   gint64     ul_tm_year;     /* upload time (s) current year         */
   gint64     dl_tm_total;    /* download time (s) total              */
   gint64     ul_tm_total;    /* upload time (s) total                */
   time_t    last_login;     /* time of last login                   */
   char      name[10];       /* username                             */
   char      password[40];   /* encrypted password                   */
   char      info[32];       /* info                                 */
   char      idents[10][20]; /* 12bytes for ident; 8byte for iprange */
   gint64     dl_day;         /* bytes downloaded today               */
   gint64     ul_day;         /* bytes uploaded today                 */
   gint64     dl_tm_day;      /* download time (s) today              */
   gint64     ul_tm_day;      /* upload time (s) today                */
   gint32     ul_f_day;       /* files uploaded today                 */
   gint32     dl_f_day;       /* files downloaded today               */ 
   gint32     ul_f_total;     /* files uploaded total                 */
   gint32     dl_f_total;     /* files downloaded total               */
   gint32     ul_f_week;      /* files uploaded current week          */
   gint32     dl_f_week;      /* files downloaded current week        */
   gint32     ul_f_month;     /* files uploaded current month         */
   gint32     dl_f_month;     /* files downloaded current month       */
   gint32     ul_f_year;      /* files uploaded current year          */
   gint32     dl_f_year;      /* files downloaded current year        */
   gint32     f_nuked_total;  /* files nuked total                    */
   gint64     b_nuked_total;  /* bytes nuked total                    */
   gint64     c_nuked_total;  /* credits nuked total                  */
   gint32     t_nuked_total;  /* times nuked total                    */
   gint32     f_nuked_week;   /* files nuked current week             */
   gint64     b_nuked_week;   /* bytes nuked current week             */
   gint64     c_nuked_week;   /* credits nuked current week           */
   gint32     t_nuked_week;   /* times nuked current week             */
   gint32     f_nuked_month;  /* files nuked current month            */
   gint64     b_nuked_month;  /* bytes nuked current month            */
   gint64     c_nuked_month;  /* credits nuked current month          */
   gint32     t_nuked_month;  /* times nuked current month            */ 
   gint32     f_nuked_year;   /* files nuked current year             */
   gint64     b_nuked_year;   /* bytes nuked current year             */
   gint64     c_nuked_year;   /* credits nuked current year           */
   gint32     t_nuked_year;   /* times nuked current year             */
   gint32     f_nuked_day;    /* files nuked today                    */
   gint64     b_nuked_day;    /* bytes nuked today                    */
   gint64     c_nuked_day;    /* credits nuked today                  */
   gint32     t_nuked_day;    /* times nuked today                    */
   time_t    creation_time;  /* creattion time of the account        */
   char      creator[10];    /* the creator of the account           */
} ; 

struct user_s
{ 
   gint32     uid;            /* user ID                              */
   gint32     gid;            /* group ID                             */
   gint32     gids[8];        /* additional groups                    */
   gint32     class;          /* user class                           */
   gint64     rights;         /* user rights                          */
   gint32     limit;          /* limit (ratio)                        */
   gint64     credit;         /* credit in bytes                      */
   gint32     dl_speed;       /* min and max download speed in kb/s   */
   gint32     ul_speed;       /* min and max upload speed   in kb/s   */
   gint64     dl_week;        /* download current week                */
   gint64     ul_week;        /* upload current week                  */
   gint64     dl_month;       /* download current month               */
   gint64     ul_month;       /* upload current month                 */
   gint64     dl_year;        /* download current year                */
   gint64     ul_year;        /* upload current year                  */
   gint64     dl_total;       /* download total                       */
   gint64     ul_total;       /* upload total                         */
   gint64     dl_tm_week;     /* download time (s) current week       */
   gint64     ul_tm_week;     /* upload time (s) current week         */
   gint64     dl_tm_month;    /* download time (s) current month      */
   gint64     ul_tm_month;    /* upload time (s) current month        */  
   gint64     dl_tm_year;     /* download time (s) current year       */
   gint64     ul_tm_year;     /* upload time (s) current year         */
   gint64     dl_tm_total;    /* download time (s) total              */
   gint64     ul_tm_total;    /* upload time (s) total                */
   time_t    last_login;     /* time of last login                   */
   char      name[10];       /* username                             */
   char      password[40];   /* encrypted password                   */
   char      info[32];       /* info                                 */
   char      idents[10][20]; /* 12bytes for ident; 8byte for iprange */
   gint64     dl_day;         /* bytes downloaded today               */
   gint64     ul_day;         /* bytes uploaded today                 */
   gint64     dl_tm_day;      /* download time (s) today              */
   gint64     ul_tm_day;      /* upload time (s) today                */
   gint32     ul_f_day;       /* files uploaded today                 */
   gint32     dl_f_day;       /* files downloaded today               */ 
   gint32     ul_f_total;     /* files uploaded total                 */
   gint32     dl_f_total;     /* files downloaded total               */
   gint32     ul_f_week;      /* files uploaded current week          */
   gint32     dl_f_week;      /* files downloaded current week        */
   gint32     ul_f_month;     /* files uploaded current month         */
   gint32     dl_f_month;     /* files downloaded current month       */
   gint32     ul_f_year;      /* files uploaded current year          */
   gint32     dl_f_year;      /* files downloaded current year        */
   gint32     f_nuked_total;  /* files nuked total                    */
   gint64     b_nuked_total;  /* bytes nuked total                    */
   gint64     c_nuked_total;  /* credits nuked total                  */
   gint32     t_nuked_total;  /* times nuked total                    */
   gint32     f_nuked_week;   /* files nuked current week             */
   gint64     b_nuked_week;   /* bytes nuked current week             */
   gint64     c_nuked_week;   /* credits nuked current week           */
   gint32     t_nuked_week;   /* times nuked current week             */
   gint32     f_nuked_month;  /* files nuked current month            */
   gint64     b_nuked_month;  /* bytes nuked current month            */
   gint64     c_nuked_month;  /* credits nuked current month          */
   gint32     t_nuked_month;  /* times nuked current month            */ 
   gint32     f_nuked_year;   /* files nuked current year             */
   gint64     b_nuked_year;   /* bytes nuked current year             */
   gint64     c_nuked_year;   /* credits nuked current year           */
   gint32     t_nuked_year;   /* times nuked current year             */
   gint32     f_nuked_day;    /* files nuked today                    */
   gint64     b_nuked_day;    /* bytes nuked today                    */
   gint64     c_nuked_day;    /* credits nuked today                  */
   gint32     t_nuked_day;    /* times nuked today                    */
   time_t    creation_time;  /* creattion time of the account        */
   char      creator[10];    /* the creator of the account           */
   gint64     wkly_alot;      /* weekly alottment to add credits      */
   /*gint64     temp[32];        temp buffer for future uses          */
   char      home[256];      /* this should work unless size(char) != 1 byte */ 
} ; 

struct user_v223_s
{ char   name[10];
  long   uid;
  long   gid;
  char   password[14];
  long   kb_downloaded;
  long   kb_uploaded;
  time_t last_login;
  char   idents[10][20]; /* 12bytes for ident; 8byte for iprange */
  int    limit;
  char   info[16];
  long   credit;
  long   gids[8];
  long   class;
  long   kb_down_tmp;
  long   kb_up_tmp;
} ;

struct group_v3009
{ gint32     gid;            /* group ID                             */
  gint32     limit;          /* limit (ratio)                        */
  gint64     credit;         /* credit in bytes                      */
  gint32     dl_speed;       /* min and max download speed in kb/s   */
  gint32     ul_speed;       /* min and max upload speed   in kb/s   */
  gint32     logins;         /* max number of logins                 */
  char      name[10];       /* group name                           */
  char      info[32];       /* group information                    */
  char      ident[20];      /* 12bytes for ident; 8byte for iprange */
} ;

struct group_v3012
{ gint32     gid;            /* group ID                             */
  gint32     limit;          /* limit (ratio)                        */
  gint64     credit;         /* credit in bytes                      */
  gint32     dl_speed;       /* min and max download speed in kb/s   */
  gint32     ul_speed;       /* min and max upload speed   in kb/s   */
  gint32     logins;         /* max number of logins                 */
  char      name[10];       /* group name                           */
  char      info[32];       /* group information                    */
  char      ident[20];      /* 12bytes for ident; 8byte for iprange */
  int       umask_group;    /* group based umask                    */
} ;

struct group_v1
{ gint32     gid;            /* group ID                              */
  gint32     limit;          /* limit (ratio)                         */
  gint64     credit;         /* credit in bytes                       */
  gint32     dl_speed;       /* min and max download speed in kb/s    */
  gint32     ul_speed;       /* min and max upload speed   in kb/s    */
  gint32     logins;         /* max number of logins                  */
  char      name[10];       /* group name                            */
  char      info[32];       /* group information                     */
  char      ident[20];      /* 12bytes for ident; 8byte for iprange  */
  int	    umask_group;    /* group based umask                     */
  gint32     max_logins;     /* max # of grp usrs logged at same time */
  gint32     max_slots;      /* max number of users a group can have  */
} ;

struct group_s
{ gint32     gid;            /* group ID                               */
  gint32     limit;          /* limit (ratio)                          */
  gint64     credit;         /* credit in bytes                        */
  gint32     dl_speed;       /* min and max download speed in kb/s     */
  gint32     ul_speed;       /* min and max upload speed   in kb/s     */
  gint32     logins;         /* max number of logins                   */
  char      name[10];       /* group name                             */
  char      info[32];       /* group information                      */
  char      ident[20];      /* 12bytes for ident; 8byte for iprange   */
  int	    umask_group;    /* group based umask                      */
  gint32     max_logins;     /* max # of grp usrs logged at same time  */
  gint32     max_slots;      /* max number of users a group can have   */
  gint64     rights;         /* rights that will be copied to the user */
  gint32     temp;           /* not used for the moment                */
  gint32     temp2;          /*                ""                      */
  gint64     temp3;          /*                ""                      */
} ;

struct child_s
{ 
  time_t         act_time;         /* time of last activity   */
  char           activity[160];    /* last activity           */
  int            admin;            /* flag: admin port        */
  int            bouncer;          /* 0x01=bouncer 0x02=ident */
  char           buffer[1024];     /* command buffer          */
  int            buffer_pos;       /* position in buffer      */
  int            buffer_overrun;   /* flag: buffer overrun    */
  char           cwd[255];         /* current path            */
  gint64         dl;               /* bytes downloaded        */
  gint64         dl_tm;            /* download time (s)       */
  int            fd;               /* fd of server socket     */
  char           filename[MAXFILENAMELENGTH]; /* transferred file */
  int            hidden;           /* hidden activity: visible to superuser only */
  int            idle;             /* flag: idle              */
  unsigned long  ip;               /* logged in from this ip  */
  char		 ident[12];        /* ident                   */
  int            lim_bw;           /* limit: bandwidth        */
  int            lim_direction;    /* limit: direction        */
  int            lim_session;      /* limit: anon or user     */
  pid_t          pid;              /* process id              */
  FILE*          rdp;              /* read pipe               */
  gint64         speed;            /* current transfer speed  */
  int            sock;             /* socket number           */
  time_t         time;             /* login time              */
  gint64         ul;               /* bytes uploaded          */
  gint64         ul_tm;            /* upload time (s)         */
  struct user_s* user;             /* user record             */
  gint           xdupe;            /* x-dupe flag             */
  FILE*          wrp;              /* write pipe              */
} ;

struct shandler_s
{ int   fd;                      /* file descriptor to listen on */
  void* ptr;                     /* handler-specific pointer     */
  int   (*handler)(int,void*);   /* service handler              */
} ;

struct config_s
{ char  key[32];    /* key                     */
  int   type;       /* type (bool, int, char)  */
  void* value;      /* pointer to value        */
  gint64 min;        /* minimum value           */
  gint64 max;        /* maximum value           */
  void* dflvalue;   /* default value           */
} ;

struct limit_s
{ int sday;        /* start day of week                */
  int shour;       /* start hour                       */
  int smin;        /* start minute                     */
  int eday;        /* end day of week                  */
  int ehour;       /* end hour                         */
  int emin;        /* end minute                       */
  int max_anon;    /* number of anon                   */
  int max_anon_dom;/* number of anon per domain        */
  int max_anon_ip; /* number of anon per ip            */
  int max_user;    /* number of users                  */
  int max_sum;     /* flag: sum of anon & user         */
  int sp_dl;       /* download speed limit             */
  int sp_ul;       /* upload speed limit               */
  int sp_sum;      /* flag: sum of upload and download */
  int sp_dl_fact;  /* download speed factor            */
  int sp_ul_fact;  /* upload speed factor              */
  int sp_dl_avg;   /* average download speed           */
  int sp_dl_dbl;   /* double download speed            */
  int sp_dl_half;  /* half download speed              */
  int sp_ul_avg;   /* average upload speed             */
  int sp_ul_dbl;   /* double upload speed              */
  int sp_ul_half;  /* half upload speed                */
} ;

/* section related stuff */
struct sstat_v3012_601
{ gint32     uid;            /* user ID                              */
  gint64     credit;         /* credit in bytes                      */
  gint64     dl_week;        /* download current week                */
  gint64     ul_week;        /* upload current week                  */
  gint64     dl_month;       /* download current month               */
  gint64     ul_month;       /* upload current month                 */
  gint64     dl_year;        /* download current year                */
  gint64     ul_year;        /* upload current year                  */
  gint64     dl_total;       /* download total                       */
  gint64     ul_total;       /* upload total                         */
  gint64     dl_tm_week;     /* download time (s) current week       */
  gint64     ul_tm_week;     /* upload time (s) current week         */
  gint64     dl_tm_month;    /* download time (s) current month      */
  gint64     ul_tm_month;    /* upload time (s) current month        */
  gint64     dl_tm_year;     /* download time (s) current year       */
  gint64     ul_tm_year;     /* upload time (s) current year         */
  gint64     dl_tm_total;    /* download time (s) total              */
  gint64     ul_tm_total;    /* upload time (s) total                */
} ;

struct sstat_v1_beta29
{ gint32     uid;            /* user ID                              */
  gint64     credit;         /* credit in bytes                      */
  gint64     dl_week;        /* download current week                */
  gint64     ul_week;        /* upload current week                  */
  gint64     dl_month;       /* download current month               */
  gint64     ul_month;       /* upload current month                 */
  gint64     dl_year;        /* download current year                */
  gint64     ul_year;        /* upload current year                  */
  gint64     dl_total;       /* download total                       */
  gint64     ul_total;       /* upload total                         */
  gint64     dl_tm_week;     /* download time (s) current week       */
  gint64     ul_tm_week;     /* upload time (s) current week         */
  gint64     dl_tm_month;    /* download time (s) current month      */
  gint64     ul_tm_month;    /* upload time (s) current month        */
  gint64     dl_tm_year;     /* download time (s) current year       */
  gint64     ul_tm_year;     /* upload time (s) current year         */
  gint64     dl_tm_total;    /* download time (s) total              */
  gint64     ul_tm_total;    /* upload time (s) total                */
  gint32     ul_f_total;     /* files uploaded total                 */
  gint32     dl_f_total;     /* files downloaded total               */
  gint32     ul_f_week;      /* files uploaded current week          */
  gint32     dl_f_week;      /* files downloaded current week        */
  gint32     ul_f_month;     /* files uploaded current month         */
  gint32     dl_f_month;     /* files downloaded current month       */
  gint32     ul_f_year;      /* files uploaded current year          */
  gint32     dl_f_year;      /* files downloaded current year        */
  gint64     dl_day;         /* download current day                 */
  gint64     ul_day;         /* upload current day                   */
  gint64     dl_tm_day;      /* download time (s) current day        */
  gint64     ul_tm_day;      /* upload time (s) current day          */
  gint32     dl_f_day;       /* files downloaded today               */
  gint32     ul_f_day;       /* files uploaded today                 */
} ;

struct sstat_s
{ gint32     uid;            /* user ID                              */
  gint64     credit;         /* credit in bytes                      */
  gint64     dl_week;        /* download current week                */
  gint64     ul_week;        /* upload current week                  */
  gint64     dl_month;       /* download current month               */
  gint64     ul_month;       /* upload current month                 */
  gint64     dl_year;        /* download current year                */
  gint64     ul_year;        /* upload current year                  */
  gint64     dl_total;       /* download total                       */
  gint64     ul_total;       /* upload total                         */
  gint64     dl_tm_week;     /* download time (s) current week       */
  gint64     ul_tm_week;     /* upload time (s) current week         */
  gint64     dl_tm_month;    /* download time (s) current month      */
  gint64     ul_tm_month;    /* upload time (s) current month        */
  gint64     dl_tm_year;     /* download time (s) current year       */
  gint64     ul_tm_year;     /* upload time (s) current year         */
  gint64     dl_tm_total;    /* download time (s) total              */
  gint64     ul_tm_total;    /* upload time (s) total                */
  gint32     ul_f_total;     /* files uploaded total                 */
  gint32     dl_f_total;     /* files downloaded total               */
  gint32     ul_f_week;      /* files uploaded current week          */
  gint32     dl_f_week;      /* files downloaded current week        */
  gint32     ul_f_month;     /* files uploaded current month         */
  gint32     dl_f_month;     /* files downloaded current month       */
  gint32     ul_f_year;      /* files uploaded current year          */
  gint32     dl_f_year;      /* files downloaded current year        */
  gint64     dl_day;         /* download current day                 */
  gint64     ul_day;         /* upload current day                   */
  gint64     dl_tm_day;      /* download time (s) current day        */
  gint64     ul_tm_day;      /* upload time (s) current day          */
  gint32     dl_f_day;       /* files downloaded today               */
  gint32     ul_f_day;       /* files uploaded today                 */
  gint64     wkly_alot;      /* weekly alottment to add credits      */
} ;

struct dircfg_s
{ char *key;
  int  type;      /* type (bool, int, char)  */
  void *value;   
} ;

struct section_s
{ 
   char            section[16];
   char            config_file[64];
   char            stat_file[64];
   char            path[256];

   struct dircfg_s *config;
   int             config_size;
   int             config_count;

   struct sstat_s  *stat;
   int             stat_size;
   int             stat_count;
} ;



struct dupe_s
{
   char          filename[OPENFTPD_PATH_MAX];
   unsigned long crc32;   /* checksum of the first 4KB block */
   gint32         ul_uid;  /* UID of the uploader             */
   gint32         unduped; /* flag for unduped files          */
   gint32         deleted; /* flag for deleted files          */
   time_t        ul_time; /* upload time                     */
} ;


struct ddupe_s
{
   char          directory[OPENFTPD_PATH_MAX];
   gint32         ul_uid;  /* UID of the owner           */
   gint32         unduped; /* flag for unduped directory */
   gint32         deleted; /* flag for deleted directory */
   time_t        ul_time; /* upload time                */
} ;


struct ip_s
{ char ip[8];
} ;


#define IP_ALLOW    0
#define IP_DENY     1

#define sig_size 16

#define STATE_MASK               0x0F
#define STATE_CLOSE              0x01
#define STATE_OPEN               0x00
#define STATE_SHUTDOWN           0x02
#define STATE_FLAG_MASK          0xF0
#define STATE_FLAG_NORMAL        0x00
#define STATE_FLAG_TRANSACTIONAL 0x10
#define STATE_FLAG_IMMEDIATE     0x20

#if (INTSIZE >= 8)

/* constants for 64 bit processors */

#define R_STATE             0x00000001
#define R_PORTS             0x00000002
#define R_GUI               0x00000004
#define R_MONITOR           0x00000008
#define R_FILEACCESS        0x00000010
#define R_DUMP_FILES        0x00000020
#define R_REGISTRY          0x00000040
#define R_MESSAGES          0x00000080
#define R_CONFIG            0x00000100
#define R_NUKE              0x00000200 
#define R_UNDUPE            0x00000400  

#define R_GROUP_ALL         0x000F0000
#define R_GROUP_ADD         0x00010000
#define R_GROUP_EDIT        0x00020000
#define R_GROUP_DELETE      0x00040000
#define R_GROUP_LIST        0x00080000

#define R_USER_ALL          		0x7FFFFF700000000
#define R_USER_DISABLE      		0x000000100000000
#define R_USER_KICK         		0x000000200000000
#define R_USER_PASSWD       		0x000000400000000
#define R_USER_LOCK                     0x000000800000000
#define R_USER_ADD          		0x000001000000000
#define R_USER_EDIT         		0x000002000000000
#define R_USER_DELETE       		0x000004000000000
#define R_USER_READ         		0x000008000000000
#define R_USER_ADD_SO       		0x000010000000000
#define R_USER_EDIT_SO      		0x000020000000000
#define R_USER_DELETE_SO    		0x000040000000000
#define R_USER_READ_SO      		0x000080000000000
#define R_USER_ADD_GO       		0x000100000000000
#define R_USER_EDIT_GO      		0x000200000000000
#define R_USER_DELETE_GO    		0x000400000000000
#define R_USER_READ_GO      		0x000800000000000
#define R_USER_ADD_ANON     		0x001000000000000
#define R_USER_ADD_TMP      		0x002000000000000
#define R_USER_EDIT_IP      		0x004000000000000
#define R_USER_EDIT_GRPS    		0x008000000000000
#define R_USER_EDIT_CREDIT 	 	0x010000000000000
#define R_USER_EDIT_STATS   		0x020000000000000
#define R_USER_EDIT_RIGHTS	  	0x040000000000000
#define R_USER_EDIT_INFO	    	0x080000000000000
#define R_USER_EDIT_UID  	   	0x100000000000000
#define R_USER_EDIT_LOGINS  		0x200000000000000
#define R_USER_EDIT_SPEEDS  		0x400000000000000
#define R_ALWAYS_LOGIN			0x800000000000000
#define R_ALL              		0x7FFFFFFFFFFFFFF

#else

/* same constants for 32 bit processors */

#define R_STATE             0x00000001LL
#define R_PORTS             0x00000002LL
#define R_GUI               0x00000004LL
#define R_MONITOR           0x00000008LL
#define R_FILEACCESS        0x00000010LL
#define R_DUMP_FILES        0x00000020LL
#define R_REGISTRY          0x00000040LL
#define R_MESSAGES          0x00000080LL
#define R_CONFIG            0x00000100LL
#define R_NUKE              0x00000200LL 
#define R_UNDUPE            0x00000400LL  
  
#define R_GROUP_ALL         0x000F0000LL
#define R_GROUP_ADD         0x00010000LL
#define R_GROUP_EDIT        0x00020000LL
#define R_GROUP_DELETE      0x00040000LL
#define R_GROUP_LIST        0x00080000LL

#define R_USER_ALL          		0x7FFFFF700000000LL
#define R_USER_DISABLE      		0x000000100000000LL
#define R_USER_KICK         		0x000000200000000LL
#define R_USER_PASSWD       		0x000000400000000LL
#define R_USER_LOCK                     0x000000800000000LL
#define R_USER_ADD          		0x000001000000000LL
#define R_USER_EDIT         		0x000002000000000LL
#define R_USER_DELETE       		0x000004000000000LL
#define R_USER_READ         		0x000008000000000LL
#define R_USER_ADD_SO       		0x000010000000000LL
#define R_USER_EDIT_SO      		0x000020000000000LL
#define R_USER_DELETE_SO    		0x000040000000000LL
#define R_USER_READ_SO      		0x000080000000000LL
#define R_USER_ADD_GO       		0x000100000000000LL
#define R_USER_EDIT_GO      		0x000200000000000LL
#define R_USER_DELETE_GO    		0x000400000000000LL
#define R_USER_READ_GO      		0x000800000000000LL
#define R_USER_ADD_ANON     		0x001000000000000LL
#define R_USER_ADD_TMP      		0x002000000000000LL
#define R_USER_EDIT_IP      		0x004000000000000LL
#define R_USER_EDIT_GRPS    		0x008000000000000LL
#define R_USER_EDIT_CREDIT  		0x010000000000000LL
#define R_USER_EDIT_STATS   		0x020000000000000LL
#define R_USER_EDIT_RIGHTS  		0x040000000000000LL
#define R_USER_EDIT_INFO		0x080000000000000LL
#define R_USER_EDIT_UID			0x100000000000000LL
#define R_USER_EDIT_LOGINS 		0x200000000000000LL
#define R_USER_EDIT_SPEEDS 	 	0x400000000000000LL
#define R_ALWAYS_LOGIN			0x800000000000000LL
#define R_ALL               		0x7FFFFFFFFFFFFFFLL

#endif

#endif /* __DEFINESD_H_ */

/* EOF */
