/*   f4adp_admind.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996-2000 by Crescent
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *                       
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "ac_config.h"

#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifndef HAVE_INET_ATON
#include "inet_aton.c"
#endif
#include "externsd.h"
#include "ftpd.h"
#include "serverd.h"
#include "user.h"
#include "utils.h"
#include "f4adp_admin.h"

int             not_count   = 0;       /* notification counter       */
struct watch_s* watch       = NULL;    /* list of watchers           */
int             watch_count = 0;       /* current number of watchers */
int             watch_size  = 0;       /* max. number of watchers    */
int             watch_sock  = -1;      /* watch UDP socket           */

/* Format of a watch notification:

   Event   Flag Parameters
   -----------------------
   Activity  +A <not_cnt> <pid> <download> <upload> <idle> <action>
   Login     +L <not_cnt> <pid> <user> <ip> <time> <download> <upload> <idle> <activity>
   Logout    +O <not_cnt> <pid>
   ---
   Error Log +E <log entry>
   Shutdown  +S
*/

/* sendto with retries */
/* this is because with the java gui running locally, sendto() often fails
   with ECONNREFUSED
   (which I dont understand, so I hope one of the tries finally succeeds)
*/
void sendto_rt(int sock, char* buf, int len, int flags, struct sockaddr* addr,
  int addrlen)
{ int tries;

  for (tries=0; tries<10; tries++)
    if (sendto(sock,buf,len,flags,addr,addrlen) > 0) break;
}

/* add a watcher (called by client) */
int f4adp_watch_add(struct child_s* c, char** p, int n)
{ struct sockaddr_in addr;
  int                i;
  int                new_size;
  struct watch_s*    new_watch;
  int                port;
  int                interval; /* seconds to wait till next notification */
  char               str[256];
  int                su;
  time_t             t;

  if (!inet_aton(p[0],&addr.sin_addr) || !sscanf(p[1],"%i",&port) ||
      !sscanf(p[2], "%i", &interval))
  { server_reply(c,"-Syntax error");
    return 0;
  }
  if (interval < 5) {
     server_reply(c, "-Interval must be bigger than 5 seconds");
     return 0;
  }
  if (watch_count >= watch_size)
  { new_size = watch_size + 4;
    new_watch = g_realloc(watch,new_size*sizeof(struct watch_s));
    if (new_watch == NULL)
    { server_reply(c,"-No memory for watch");
      return 0;
    }
    watch = new_watch;
    watch_size = new_size;
  }

  if (watch_sock < 0)
  { watch_sock = get_socket(SOCK_DGRAM,0,-1,&addr);
    if (watch_sock < 0)
    { server_reply(c,"-Unable to get socket");
      return 0;
    }
  }

  watch[watch_count].addr.sin_family = AF_INET;
  inet_aton(p[0],&watch[watch_count].addr.sin_addr);
  watch[watch_count].addr.sin_port = htons((u_short)port);
  watch[watch_count].child = c;
  watch[watch_count].interval = interval;
  watch[watch_count].notification = 0;

  server_reply(c,"+Success");
 
  t = time(NULL); /* get current time */
  su = superuser(c->user);
  for (i=0; i<children; i++) if (child[i].user)
  { addr.sin_addr.s_addr = htonl(child[i].ip);
    g_snprintf(str, sizeof(str), "+L\t%i\t%ld\t%s\t%s\t"SITIME"\t"SI64"\t"SI64"\t%i\t%s",
      watch_count,
      (long)child[i].pid,
      child[i].user->name,
      inet_ntoa(addr.sin_addr),
      t - child[i].time,
      child[i].dl,
      child[i].ul,
      child[i].idle,
      su || !child[i].hidden ? child[i].activity : "LOGIN");
    sendto_rt(watch_sock,str,strlen(str),0,
      (struct sockaddr*)&watch[watch_count].addr,
      sizeof(struct sockaddr_in));
  }

  watch_count++;

  return 0;
}

/* delete all watchers of this client (called internal) */
int watch_del(struct child_s* c)
{ int i;
  int n;

  for (i=n=0; i<watch_count; ) if (watch[i].child == c)
  { if (i < --watch_count) memmove(watch+i,watch+i+1,
      (watch_count-i)*sizeof(struct watch_s));
    n++;
  } else i++;
  return n;
}

/* delete all watchers of this client (called by client) */
int f4adp_watch_del(struct child_s* c, char** p, int n)
{ n = watch_del(c);
  server_reply(c,n ? "+Success" : "-Not found");
  return 0;
}

/* updates pointers when child records have been moved */
void update_children_watch(struct child_s* c_old, struct child_s* c_new)
{ int i;

  if (c_old != c_new)
    for (i=0; i<watch_count; i++) if (watch[i].child >= c_old)
      watch[i].child = c_new + (watch[i].child - c_old);
}

/* notifies all watchers */
void notify_watchers(char* action, int hidden, struct child_s* c)
{ int i;
  time_t t;

  t = time(NULL); /* get current time */
  for (i=0; i<watch_count; i++) if (!hidden ||
    (c->user == watch[i].child->user) || superuser(watch[i].child->user)) {
     if (watch[i].notification + watch[i].interval < t) {
        sendto_rt(watch_sock,action,strlen(action),0,
                 (struct sockaddr*)&watch[i].addr,sizeof(struct sockaddr_in));
        watch[i].notification = t;
     }
  }
  if (++not_count < 0) not_count = 0;
}

/* notify watchers of user activity */
void notify_watchers_activity(struct child_s* c)
{ char str[256];

  g_snprintf(str, sizeof(str), "+A\t%i\t%ld\t"SI64"\t"SI64"\t%i\t%s",
    not_count,(long)c->pid,c->dl,c->ul,c->idle,c->activity);
  notify_watchers(str,c->hidden,c);
}

/* notify watchers of logged in user */
void notify_watchers_login(struct child_s* c)
{ struct sockaddr_in addr;
  char               str[256];
  time_t             t;

  t = time(NULL);
  addr.sin_addr.s_addr = htonl(c->ip);
  g_snprintf(str, sizeof(str), "+L\t%i\t%ld\t%s\t%s\t"SITIME"\t"SI64"\t"SI64"\t%i\t%s",
    not_count,
    (long)c->pid,
    c->user->name,
    inet_ntoa(addr.sin_addr),
    t - c->time,
    c->dl,
    c->ul,
    c->idle,
    c->activity);
  notify_watchers(str,0,c);
}

/* notify watchers of logged out user */
void notify_watchers_logout(struct child_s* c)
{ char str[256];

  g_snprintf(str, sizeof(str), "+O\t%i\t%ld",not_count,(long)c->pid);
  notify_watchers(str,0,c);
}

/* notify watchers of server shutdown */
void notify_watchers_shutdown(struct child_s* c)
{ notify_watchers("+S",0,c);
}

/* EOF */
