/*   f4adp_config.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996-1999 by Crescent
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *                       
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "externsd.h"
#include "ftpd.h"
#include "limits.h"
#include "log.h"
#include "serverd.h"
#include "utils.h"
#include "sections.h"
#include "strlcpy.h"
#include "strlcat.h"

int get_cfg_filename(char* filename, char** p)
{ strcpy(filename,server_dir);
  strcat(filename,DFL_CFGDIR);
  if (p[0])
  { strcat(filename,p[0]);
    strcat(filename,".cfg");
  }
  return 0;
}

char* cfg_files[] =
  { "cdpath", "check", "checkdelete", "checkdirdupe", "checkdupe", "customcmd", "dirshortcut", "limits", "sections", NULL };

/* get available config files */
int f4adp_cfg_avl(struct child_s* c, char** p, int n)
{ char        filename[256];
  int         i;
  char*       name;
  int         se;
  struct stat st;
  char        str[512] = "";

  p[0] = NULL;
  se = get_cfg_filename(filename,p);
  if (se)
  { server_reply(c,"-Syntax error");
    return 0;
  }
  name = filename + strlen(filename);
  *str = '\0';
  for (i=0; cfg_files[i]; i++)
  {
    strlcat(str, *str ? " " : "+", sizeof(str));
    strlcpy(name, cfg_files[i], sizeof(filename) - strlen(filename));

    /* name is now pointing to server_dir/CFG_DIR/filename in buffer filename */
    strncat(name, ".cfg", sizeof(filename) - strlen(filename) - 1);
    strlcat(name, ".cfg", sizeof(name));
    se = stat(filename,&st);
    strlcat(str, se < 0 ? "-" : "*", sizeof(str));
    strlcat(str, cfg_files[i], sizeof(str));
  }
  server_reply(c,str);
  return 0;
}

/* get config file */
int f4adp_cfg_get(struct child_s* c, char** p, int n)
{ FILE*       file = NULL;
  char        filename[256];
  int         se;
  struct stat st;
  char        str[256];

  se = get_cfg_filename(filename,p);
  if (se)
  { server_reply(c,"-Syntax error");
    return 0;
  }
  if (stat(filename,&st) == -1)
  { se = 1;
    g_snprintf(str, sizeof(str), "+ -1");
  } else
  { file = fopen(filename,"rt");
    if (!file) st.st_size = 0;
    g_snprintf(str, sizeof(str), "+ %u",(unsigned int)st.st_size);
    se = (st.st_size == 0);
  }
  server_reply(c,str);
  if (se) return 0;
  while (st.st_size > 0)
  { if (!fgets(str,sizeof(str),file))
    { se = (st.st_size >= sizeof(str)) ? sizeof(str)-1 : st.st_size;
      memset(str,'x',se-1);
      str[se++] = '\n';
      str[se] = 0;
    }
    st.st_size -= strlen(str);
    fprintf(c->wrp,"%s",str);
  }
  fflush(c->wrp);
  fclose(file);
  return 0;
}

/* remove config file */
int f4adp_cfg_rm(struct child_s* c, char** p, int n) {
  char filename[256];
  int  se;
  char str[256];

  if((se = get_cfg_filename(filename,p))) {
    server_reply(c,"-Syntax error");
    return 0;
  }
  if (unlink(filename) < 0)
    strlcpy(str, "-Remove config file failed", sizeof(str));
  else
    strlcpy(str, "+Config file removed", sizeof(str));

  if (!strcmp(p[0],"limits"))
    lim_read_config();

  server_reply(c,str);
  return 0;
}

/* set config file */
int f4adp_cfg_set(struct child_s* c, char** p, int n)
{ FILE*       file;
  char        filename[256] = "";
  int         se;
  int         size;
  char        str[256];

  se = get_cfg_filename(filename,p);
  if (!sscanf(p[1],"%i",&size)) se = 1;
  if (se)
  { server_reply(c,"-Syntax error");
    return 0;
  }
  strlcat(filename, ".tmp", sizeof(filename));
  file = fopen(filename,"wb");
  if (!file)
  { server_reply(c,"-Cannot open file");
    return 0;
  }
  server_reply(c,"+Ok");
  while (size > 0)
  { if ((se = read_to(fileno(c->rdp),str,size > sizeof(str) ?
      sizeof(str) : size,5)) <= 0) break;
    size -= se;
    fwrite(str,1,se,file);
  }
  fclose(file);
  if (size > 0) unlink(filename); else
  { strlcpy(str, filename, sizeof(str));
    str[strlen(str)-4] = '\0';
    if (rename(filename,str) < 0)
      errlog("rename(%s,%s): %s (%i)",filename,str,strerror(errno),errno);
    if (!strcmp(p[0],"limits")) lim_read_config();
    if (!strcmp(p[0],"sections")) load_global_section_config();
  }
  return 0;
}

/* EOF */
