/* ---------------------------------------------------------------------------
 *
 *   This file is part of OpenFTPD (based on FTP4ALL)
 *   (C) 1996-2000 by Crescent, (C) 2000 by the OpenFTPD Team
 *
 * ---------------------------------------------------------------------------
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 * ---------------------------------------------------------------------------
 *   MODULE: ftpd.c - main ftpd module
 * ---------------------------------------------------------------------------
 */

#include "ac_config.h"

#include <sys/types.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/wait.h>
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#include <glib.h>

#ifdef DES_ENCRYPTION
#include "des.h"
#endif
#include "globalsd.h"
#include "f4adp_admin.h"
#include "group.h"
#include "limits.h"
#include "log.h"
#include "registry.h"
#include "reject.h"
#include "serverd.h"
#include "stat.h"
#include "user.h"
#include "utils.h"
#include "sections.h"
#include "f4adp_ipban.h"
#include "dupecheck.h"
#include "strlcpy.h"
#include "strlcat.h"

void shandler_dump()
{ 
   int i;
   for (i = 0; i < shandler_count; i++) { 
      printf("%i %i %lX\n", i, shandler[i].fd, (long)shandler[i].ptr);
   }
}

/* add service handler for file descriptor fd */
int shandler_add(int fd, void* ptr, int (*handler)(int,void*))
{ 
   int                i,j,k;
   int                new_size;
   struct shandler_s* new_shandler;

   if (!shandler) FD_ZERO(&shandler_fdset);
   if (shandler_count >= shandler_size) { 
      new_size = shandler_size + 32;
      new_shandler = g_realloc(shandler, new_size * sizeof(struct shandler_s));
      if (new_shandler == NULL) { 
         errlog("no memory for service handler");
         return 0;
      }
      shandler = new_shandler;
      shandler_size = new_size;
   }
   k = shandler_count;
   i = 0;
   k = shandler_count - 1;
   while (i <= k) { 
      j = (i + k) >> 1;
      if (fd > shandler[j].fd) i = j + 1; 
      else if (fd < shandler[j].fd) k = j - 1; 
      else { 
         shandler[j].ptr     = ptr;
         shandler[j].handler = handler;
         return 1;
      }
   }
   if (++k < shandler_count) {
      memmove(shandler+k+1, shandler+k, (shandler_count-k) * 
         sizeof(struct shandler_s));
   }
 #ifdef DEBUG
   fprintf(debugf, "k=%i, fd=%i, handlers=%i\n", k, fd, shandler_count + 1);
 #endif
   shandler[k].fd      = fd;
   shandler[k].ptr     = ptr;
   shandler[k].handler = handler;
   shandler_count++;
   FD_SET(fd, &shandler_fdset);
   FD_ZERO(&fdset);
   /*printf("handlers\n");
   for (i=0; i<shandler_count; i++) printf("%i %i\n",i,shandler[i].fd);
   for (i=0; i<100; i++) if (FD_ISSET(i,&shandler_fdset)) printf("%i ",i);
   printf("\nchildren\n");
   for (i=0; i<children; i++) printf("%i %i %i %i\n",child[i].pid,
   fileno(child[i].rdp),fileno(child[i].wrp),child[i].sock);*/
   return 1;
}


/* delete service handler for file descriptor fd*/
void shandler_del(int fd)
{ 
   int i;
   int n = -1;

   /*printf("shandler_del %i\n",fd);*/
   for (i = 0; i < shandler_count; i++) {
      if (shandler[i].fd == fd) { 
	 n = i;
	 i = shandler_count;
      }
   }
   if (n == -1) return;
   FD_CLR(fd, &shandler_fdset);
   FD_ZERO(&fdset);
   shandler_count--;
   if (n < shandler_count) {
      memmove(shandler+n,shandler + n + 1, (shandler_count-n) * 
      sizeof(struct shandler_s));
   }
   /*printf("handlers\n");
   for (i=0; i<shandler_count; i++) printf("%i %i\n",i,shandler[i].fd);
   for (i=0; i<100; i++) if (FD_ISSET(i,&shandler_fdset)) printf("%i ",i);
   printf("\n");*/
}


/* log currently logged in users to who.log */
int dump_users_logged_in()
{
   int             i;
   struct child_s  *u;
   struct group_s  *g;
   char            str[256];
   FILE            *fp;
   char            ontime[16];
   char            actime[16];
/*   char            host[64];*/
   time_t          t;
   

   mk_filename(str, DFL_WHO_LOG, NULL);
   fp = fopen(str, "wt");
   if (!fp) {
      errlog("fopen(%s): %s (%i)", str, strerror(errno), errno);
      return -1;
   }
   u = child;
   for (i = 0; i < children; i++, u++) {
      if (!u->admin && u->user && !invisible(u->user) && !u->hidden) {
         t = time(NULL) - u->time;
         strftime(ontime, sizeof(ontime), "%H:%M", gmtime(&t));
         t = time(NULL) - u->act_time;
         strftime(actime, sizeof(actime), "%H:%M", gmtime(&t));
/*
         strncpy(host, u->ident, sizeof(host) - 1);
         host[strlen(u->ident)] = '@';
         host[strlen(u->ident) + 1] = '\0';
         get_hostname(u->ip, host + strlen(host));
*/
	 g = get_group_by_gid(u->user->gid);
	 fprintf(fp, "%s\t%s\t%s\t%s\t"SI64"\t"SI64"\t%s\t"SI64"\t%s"
	    "\t%s\t%ld\n", u->user->name, g ? g->name : "unknown", 
	    ontime, actime, u->ul, u->dl, u->activity, u->speed, u->cwd, 
	    u->user->info, (long)u->pid); 
      }
   }
   fclose(fp);
   return 0;
}


/* updates pointers when child records have been moved */
void update_children_shandler(struct child_s* c_old, struct child_s* c_new)
{ 
   int i;

   if (c_old != c_new) for (i=0; i<shandler_count; i++) {
      if (((void*)shandler[i].handler == (void*)child_communication) &&
         ((struct child_s*)shandler[i].ptr >= c_old)) {
         shandler[i].ptr = (struct child_s*)
           (c_new + ((struct child_s*)shandler[i].ptr - c_old));
      }
   }
}


/* send FTP reply to client */
int send_reply(int s, int rno, char* buffer)
{ 
   char msg[8];

   g_snprintf(msg, sizeof(msg), "%d ", rno);
   if (send(s, msg, strlen(msg), 0) == -1) return -1;
   if (send(s, buffer, strlen(buffer),0) == -1) return -1;
   if (*(buffer + strlen(buffer) - 1) != '\n') {
      if (send(s,"\n",1,0) == -1) return -1;
   }
   return 0;
}


void signal_handler(int sig)
{ 
   if (((sig_write==sig_size-1) && !sig_read) || (sig_write+1==sig_read)) { 
      if (errlogf) fprintf(errlogf,"array full, signal %i lost\n",sig);
      return;
   }
   sig_array[sig_write] = sig;
   if (++sig_write >= sig_size) sig_write = 0;
}


void write_pidfile(void)
{ 
   FILE* file;
   char  str[256];

   mk_filename(str,pid_file,NULL);
   file = fopen(str,"wt");
   if (file) { 
      fprintf(file,"%lu\n",(unsigned long)getpid());
      fclose(file);
   }
}


void remove_pidfile(void)
{ 
   char str[256];
   mk_filename(str,pid_file,NULL);
   unlink(str);
}


void print_stats_iprange_part(char* s, char l, char h, char c)
{ 
   if (l == h) sprintf(s,"%i",(int)(unsigned char)l); 
   else if ((l == 0) && ((unsigned char)h == 0xFF)) sprintf(s,"*"); 
   else sprintf(s,"%i-%i",(int)(unsigned char)l,(int)(unsigned char)h);
   sprintf(s+strlen(s),"%c",c);
}
    
    
void print_stats_iprange(char* s, char* iprange)
{ 
   print_stats_iprange_part(s,*(iprange+12),*(iprange+13),'.');
   print_stats_iprange_part(s+strlen(s),*(iprange+14),*(iprange+15),'.');
   print_stats_iprange_part(s+strlen(s),*(iprange+16),*(iprange+17),'.');
   print_stats_iprange_part(s+strlen(s),*(iprange+18),*(iprange+19),'\0');
}

 
void print_stats_ident(char* s, char* ident, int s_size) 
{ 
   if (*ident!='\0') sprintf(s,"%s@",ident); 
   else memset(s, 0, s_size);
   print_stats_iprange(s+strlen(s),ident);
}

  
void write_stat_file(void)
{ 
   FILE*          file;
   int            first;
   int            i;
   int            j;
   char           filename[256];
   struct user_s* u = user;
   char           iprange[50];
   mk_filename(filename,stat_file,DFL_STATFILE);
   if ((file = fopen(filename,"wt")) == NULL) return;
   /* printing some basic info */
   for (i=0; i<user_count; i++, u++) {
      fprintf(file,"%s:"SI32":"SI32":%s:"SI64":"SI64":"SITIME":"
	 SI32":%s:0x%04X:"SI64":"SI64":"SI64":"SI64":"
	 SI64":"SI32":"SI32":"SI64":"SI64":"
	 SI64":"SI64":"SI64":"SI64":"SI64":"
	 SI64":"SI64":"SI64":"SI64":"SI64":"SI32":"
	 SI32":"SI32":"SI32":"SI32":"SI32":"
	 SI32":"SI32":"SI32":"SI32":"SITIME":"
	 "%s:"SI32":"SI64":"SI64":"
	 SI32":"SI32":"SI64":"SI64":"
	 SI32":"SI32":"SI64":"SI64":"
	 SI32":"SI32":"SI64":"SI64":"
	 SI32":"SI32":"SI64":"SI64":"
	 SI32":"SI64,
	 u->name,u->uid,u->gid,u->password,u->dl_total,u->ul_total,u->last_login,
	 u->limit,u->info,u->class,u->credit,u->ul_week,u->dl_week,u->ul_month,
	 u->dl_month,u->ul_speed,u->dl_speed,u->ul_tm_week,u->dl_tm_week,
	 u->ul_tm_month,u->dl_tm_month,u->ul_tm_year,u->dl_tm_year,u->ul_tm_total,
	 u->dl_tm_total,u->ul_day,u->dl_day,u->ul_tm_day,u->dl_tm_day,u->ul_f_day,
	 u->dl_f_day,u->ul_f_total,u->dl_f_total,u->ul_f_week,u->dl_f_week,
	 u->ul_f_month,u->dl_f_month,u->ul_f_year,u->dl_f_year,u->creation_time,
	 u->creator,u->f_nuked_total,u->b_nuked_total,u->c_nuked_total,
	 u->t_nuked_total,u->f_nuked_day,u->b_nuked_day,u->c_nuked_day,
	 u->t_nuked_day,u->f_nuked_week,u->b_nuked_week,u->c_nuked_week,
	 u->t_nuked_week,u->f_nuked_month,u->b_nuked_month,u->c_nuked_month,
	 u->t_nuked_month,u->f_nuked_year,u->b_nuked_year,u->c_nuked_year,
	 u->t_nuked_year,u->wkly_alot);

       /* printing the IP's */
      first=1;
      for (j = 0; j < 10; j++) {
	 if (memcmp(u->idents[j]+12, "\0\0\0\0\0\0\0\0", 8)) { 
	    print_stats_ident(iprange,u->idents[j], sizeof(iprange));
	    fprintf(file,"%c%s",first ? ':' : ',' ,iprange);
	    first = 0;
	 }
      }
      first = 1;
      for (j = 0; j < 8; j++) { 
         fprintf(file, "%c"SI32, first ? ':' : ',' , u->gids[j]);
 	 first = 0;
      }
      fprintf(file,"\n");
   }
   fclose(file);
}


void write_gstat_file(void)
{
   FILE*            file;
   int              i;
   struct group_s*  g = group;
   char             filename[256];

   mk_filename(filename, gstat_file, DFL_GSTATFILE);
   if ((file = fopen(filename, "wt")) == NULL) return;
   for (i = 0; i < group_count; i++, g++) {
     fprintf(file,"%s:"SI32":"SI32":"SI64":"SI32":"SI32":"SI32":%s:"SI32":"SI32"\n",
     g->name,g->gid,g->limit,g->credit,g->dl_speed,g->ul_speed,g->logins,
     g->info[0] ? g->info : "-",g->max_logins,g->max_slots);
   }
   fprintf(file,"\n");
   fclose(file);        
}


void handle_signals(void)
{ 
   struct child_s* c;
   int             e;
   int             i;
   pid_t           pid;

   while (sig_read != sig_write) {
      #ifdef DEBUG
      fprintf(debugf,"processing signal %i\n",sig_array[sig_read]);
      #endif
      switch (sig_array[sig_read]) { 
         case SIGTERM :
            for (i=0, c=child; i<children; i++, c++) {
               if (c->pid != -1) kill(c->pid,SIGTERM);
            }
	    if (user_dirty) save_user_file(user_file);
            if (group_dirty) save_group_file(group_file);
            if (registry_dirty) save_registry(NULL);
            if (section_stat_dirty) save_section_stat();
            if (dupe_dirty) save_dupe_file();
            if (ipbans_dirty) {
	       if (save_ip_file(IP_ALLOW) != -1 && save_ip_file(IP_DENY) 
	          != -1) {
        	  ipbans_dirty = 0;
               }
            }
            errlog("terminated on SIGTERM");
            notify_watchers_shutdown(NULL);
            if (log_program_f) fclose(log_program_f);
            while (children) { 
	       if (child->pid != -1) kill(child->pid,SIGKILL);
               child_logout(child,CL_FORCE);
            }
            if (ss != -1) close(ss);
            if (ss2 != -1) close(ss2);
            if (admin_socket != -1) close(admin_socket);
            remove_pidfile();
            exit(0);
            break;
	 case SIGHUP :
            if (user_dirty) if (!save_user_file(user_file)) user_dirty = 0;
            if (group_dirty) if (!save_group_file(group_file)) group_dirty = 0;
            if (registry_dirty) if (!save_registry(NULL)) registry_dirty = 0;
            if (section_stat_dirty) {
	       if (!save_section_stat()) section_stat_dirty = 0;
            }	   
            if (ipbans_dirty) {
	       if (save_ip_file(IP_ALLOW) != -1 && save_ip_file(IP_DENY) != -1) {
        	  ipbans_dirty = 0;
               }
            }
            if (dupe_dirty) {
	       if (!save_dupe_file()) dupe_dirty = 0;
	    }
	    if (logf) fflush(logf);
            if (errlogf) fflush(errlogf);
            break;
	 case SIGUSR1 :
            write_stats(time(NULL));
            break;
	 case SIGUSR2 :
	    dump_users_logged_in();
	    write_stat_file();
            write_gstat_file();
	    break;
	 case SIGPIPE :
            errlog("SIGPIPE received, %i children left",children);
            break;
	 case SIGCHLD :
            do { 
	       pid = waitpid(-1,&e,WNOHANG);
               if (pid > 0) { 
	          if (pid == log_pid) { 
		     fclose(log_program_f);
        	     log_pid = 0;
        	     log_program_f = NULL;
        	  } 
		  else {
		     for (i=0, c=child; i<children; i++, c++) {
			if (c->pid == pid) { 
			   child_logout(c,WEXITSTATUS(e));
        		   i = children;
        		}
		     }
		  }   
               }
            } while (pid > 0);
            break;
      }
      if (++sig_read >= sig_size) sig_read = 0;
   }
}


int in_ipranges(unsigned long ip)
{ 
   int            i;
   int            j;
   struct user_s* u = user;

   if (!flag_get(flags,FLAG_PRIVATEMODE)) return 1;
   for (i=0; i<user_count; i++, u++) {
      for (j=0; j<10; j++) {
         if (in_iprange(ip,(unsigned char*)u->idents[j]+12)) return 1;
      }
   }
   return 0;
}


int in_idents(unsigned long ip, char* ident)
{ 
   int            i;
   int            j;
   struct user_s* u = user;

   if (!flag_get(flags,FLAG_PRIVATEMODE)) return 1;
   if (!flag_get(flags,FLAG_IDENTMODE)) return in_ipranges(ip);
   for (i=0; i<user_count; i++, u++) {
      for (j=0; j<10; j++) {
         if (in_ident(ip,ident,u->idents[j])) return 1;
      }
   }
   return 0;
}

/* change user id to root */
void switch_max_rights(void)
{ 
   uid_t euid = geteuid();
   uid_t ruid = getuid();
   gid_t egid = getegid();
   gid_t rgid = getgid();
   /* if real user id == 0 (root) and effective user id is a normal user */
   if (euid && !ruid) {
#ifdef HAVE_SETEUID
      seteuid(ruid);
      setegid(rgid);
#else
#warning seteuid() missing
#endif
//      setuid(ruid);
//      setgid(rgid);
   }
}

/* change user id to normal user */
void switch_min_rights(void)
{ 
   uid_t euid = geteuid();
   uid_t ruid = getuid();
   if (set_uid && !ruid) {
#ifdef HAVE_SETEUID
//      seteuid(set_uid);
//      setegid(set_gid);
#else
#warning seteuid() missing
#endif
//    setuid(set_uid);
//    setgid(set_gid);
   }
}


void run_alternative(int cs)
{ 
   char  buf[256];
   pid_t pid;

   if (alternative_srv) { 
      if (logf) fflush(logf);
      if (errlogf) fflush(errlogf);
      pid = fork();
      switch (pid) { 
         case -1 :
            errlog("fork: %s (%i)",strerror(errno),errno);
            send_reply(cs,421,"Server error, service not available.");
            break;
	 case 0 :
            close(0);
            close(1);
            close(2);
            if (logf) fclose(logf);
            dup2(cs,0);
            dup2(cs,1);
            dup2(cs,2);
            close(cs);
            switch_max_rights();
            mk_filename(buf,alternative_srv,NULL);
            execl(buf,alternative_srv,NULL);
            errlog("execl: %s (%i)",strerror(errno),errno);
            send_reply(1,421,"Server error, service not available (exec).");
            exit(-1);
	 default :
            break;
      }
   }
   close(cs);
}


int get_socket(int type, int port, int queue, struct sockaddr_in* addr)
{ 
   int           i = 1;
   struct linger li = { 0, 0 };
   int           sock;

   sock = socket(AF_INET,type,0);
   if (sock < 0) { 
      errlog("socket: %s (%i)",strerror(errno),errno);
      return -1;
   }
   if (fcntl(sock,F_SETFD,1) < 0) { 
      errlog("fcntl: %s (%i)",strerror(errno),errno);
      close(sock);
      return -1;
   }
   if (setsockopt(sock,SOL_SOCKET,SO_LINGER,(void*)&li,sizeof(li)) < 0) {
      errlog("setsockopt(SO_LINGER): %s (%i)",strerror(errno),errno);
   }
   if (queue >= 0) if (setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,(void*)&i,
      sizeof(i)) < 0) {
      errlog("setsockopt(SO_REUSEADDR): %s (%i)",strerror(errno),errno);
   }
   if (port >= 0) { 
      memset(addr,0,sizeof(struct sockaddr_in));
      addr->sin_family      = AF_INET;
      addr->sin_addr.s_addr = other_ip ? inet_addr(other_ip) : htonl(INADDR_ANY);
      addr->sin_port        = htons((u_short)port);
      if (port && (port < 1024)) switch_max_rights();
      if (bind(sock,(struct sockaddr*)addr,sizeof(struct sockaddr)) == -1) { 
         errlog("bind(%s:%i): %s (%i)",inet_ntoa(addr->sin_addr),
            (int)ntohs(addr->sin_port),strerror(errno),errno);
	 if (port && (port < 1024)) switch_min_rights();
	 close(sock);
	 return -1;
      }
      if (port && (port < 1024)) switch_min_rights();
      if (queue >= 0) {
	 if (listen(sock,queue) == -1) { 
	    errlog("listen(%i): %s (%i)", queue, strerror(errno), errno);
	    close(sock);
	    return -1;
	 }
      }
   }
   return sock;
}


/* check if given ip addr is an addr specified in bouncer string */
int is_bouncer(struct sockaddr_in* addr)
{ 
   char* caddr;
   int   l;
   char* ptr;

   if (bouncer) { 
      caddr = inet_ntoa(addr->sin_addr);
      l = strlen(caddr);
      ptr = strstr(bouncer,caddr);
      while (ptr) { 
         if (((ptr == bouncer) || isspace(*(ptr-1))) &&
            (!ptr[l] || isspace(ptr[l]))) return 1;
         ptr = strstr(ptr+l,caddr);
      }
   }
   return 0;
}


int srv_accept(int fd, void* ptr)
{ 
   struct child_s*    child_new;
   int                cs;
   struct hostent*    hp;
   int                i;
   struct linger      li = { 0, 0 };
   pid_t              pid;
   int                pipefd1[2];
   #ifndef HAVE_SOCKETPAIR
   int                pipefd2[2];
   #endif
   struct sockaddr_in rem_addr;
   char               str[64];

   #ifdef DEBUG
   fprintf(debugf,"creating new child (fd=%i)\n",fd);
   #endif
   i = sizeof(rem_addr);
   cs = accept(fd,(struct sockaddr*)&rem_addr,&i);

   if (cs >= 0) { 
      if (setsockopt(cs, SOL_SOCKET, SO_LINGER, (void*)&li, sizeof(li)) < 0) {
         errlog("setsockopt(SO_LINGER): %s (%i)",strerror(errno),errno);
      }
      /* set ident/bouncer flags */
      i = flag_get(flags,FLAG_PRIVATEMODE) ? 0x00 : 0x04;
      if (is_bouncer(&rem_addr)) i |= 0x05;
      if (flag_get(flags,FLAG_IDENTMODE)) i |= 0x02;

      /* check for rejected address */
      if (!(i & 0x01) && 
         reject_is_rejected_ip(ntohl(rem_addr.sin_addr.s_addr), 1)) { 
	 close(cs);
      } 
      else { 
         if (fd == admin_socket) { 
	    g_snprintf(str, sizeof(str), "BOUNCER=%i",i);
            putenv(str);
            g_snprintf(str, sizeof(str), "BOUNCERTIMEOUT=%i",bouncertimeout);
            putenv(str);
            g_snprintf(str, sizeof(str), "SOCK=%i",cs);
            putenv(str);
	 }
	 if ((i & 0x04) || in_ipranges(ntohl(rem_addr.sin_addr.s_addr))) { 
	    if (children >= child_size) { 
	      child_size += 16;
              child_new = (struct child_s*)g_realloc(child,
        	 child_size*sizeof(struct child_s));
              if (child_new) { 
	         update_children_shandler(child,child_new);
        	 update_children_watch(child,child_new);
              }
            }
	    else child_new = child;
            if (child_new) { 
	       child = child_new;
               child_new = child+children;
               child_new->admin = (fd == admin_socket);
               child_new->sock = cs;
               child_new->fd = fd;
               child_new->user = NULL;
               child_new->ip = ntohl(rem_addr.sin_addr.s_addr);
               child_new->ident[0] = '\0';
               strlcpy(child_new->activity,child_new->admin ? "ADMIN" : "LOGIN", sizeof(child_new->activity));
               time(&child_new->act_time);
               strlcpy(child_new->cwd, "/", sizeof(child_new->cwd));
	       child_new->idle = 1;
               child_new->hidden = 0;
               child_new->dl = child_new->ul = child_new->dl_tm = child_new->ul_tm = 0;
               child_new->lim_bw = child_new->lim_direction =
        	 child_new->lim_session = 0;
               child_new->bouncer = i;
               child_new->buffer_pos = child_new->buffer_overrun = 0;
	       child_new->speed = 0;
               child_new->xdupe = 0;

               #ifdef HAVE_SOCKETPAIR
               if ((i = socketpair(AF_UNIX,SOCK_STREAM,0,pipefd1)) < 0) { 
	          errlog("socketpair(): %s (%i)",strerror(errno),errno);
               #else
               i = pipe(pipefd1);
               if (i || pipe(pipefd2)) { 
	          errlog("pipe(): %s (%i)",strerror(errno),errno);
               #endif
        	  send_reply(cs, 421, "Server error, service not available.");
        	  close(cs);
        	  cs = -1;
        	  if (!i) { 
		     close(pipefd1[0]);
        	     close(pipefd1[1]);
        	  }
        	  errlog("pipe error, i have %i children, list follows:",children);
        	  child_new = child;
        	  for (i = 0; i < children; i++, child_new++) { 
		      errlog("child %2i, user=%-9s, socket=%3i, readp=%3i, "
		         "writep=%3i\n", i, child_new->user ? 
			 child_new->user->name : "UNKNOWN", child_new->sock,
			 fileno(child_new->rdp), fileno(child_new->wrp));
        	  }
               }
            } 
	    else { 
	       child_size -= 16;
               errlog("no memory for child");
               send_reply(cs,421,"Server error, service not available.");
               close(cs);
               cs = -1;
            }
            if (cs != -1) { 
	       if (logf) fflush(logf);
               if (errlogf) fflush(errlogf);
               pid = fork();
               switch (pid) { 
	          case -1 :
        	     errlog("fork: %s (%i)",strerror(errno),errno);
        	     send_reply(cs, 421, "Server error, service not available.");
        	     close(pipefd1[0]);
        	     close(pipefd1[1]);
                     #ifndef HAVE_SOCKETPAIR
        	     close(pipefd2[0]);
        	     close(pipefd2[1]);
                     #endif
        	     break;

        	  case 0 :
        	     close(0);
        	     close(1);
                     #ifdef HAVE_SOCKETPAIR
        	     dup2(pipefd1[1],1);
        	     dup2(pipefd1[1],0);
        	     close(pipefd1[0]);
        	     close(pipefd1[1]);
                     #else
        	     close(pipefd1[0]);
        	     close(pipefd2[1]);
        	     dup2(pipefd1[1],1);
        	     dup2(pipefd2[0],0);
        	     close(pipefd1[1]);
        	     close(pipefd2[0]);
                     #endif
        	     if (fd == admin_socket) { 
		        execl(admin_program, "ftpa", NULL);
                        errlog2("execl(%s): %s (%i)", admin_program,
 			   strerror(errno), errno);
        	     } 
		     else { 
		        execl(server_program,program_name ? program_name : 
			   server_program, NULL);
                        errlog2("execl(%s): %s (%i)",server_program,
		           strerror(errno),errno);
                        send_reply(cs,421,"Server error, service not "
			   "available (exec).");
        	     }
        	     exit(0);
        	     break;

        	  default :
        	     child_new->pid  = pid;
        	     close(pipefd1[1]);
        	     if (fcntl(pipefd1[0],F_SETFD,1) < 0) {
                        errlog("fcntl(F_SETFD): %s (%i)", strerror(errno), errno);
        	     }
		     if ((i = fcntl(pipefd1[0],F_GETFL)) < 0) {
                        errlog("fcntl(F_GETFL): %s (%i)",strerror(errno),errno); 
		     }
		     else { 
		        if (fcntl(pipefd1[0],F_SETFL,i | O_NONBLOCK) < 0) {
                	   errlog("fcntl(F_SETFL): %s (%i)", strerror(errno), errno);
        	        }
		     }
                     #ifdef HAVE_SOCKETPAIR
        	     child_new->rdp = fdopen(pipefd1[0],"rt");
        	     if (child_new->rdp) child_new->wrp = fdopen(pipefd1[0],"wt");
                     #else
        	     close(pipefd2[0]);
        	     if (fcntl(pipefd2[1],F_SETFD,1) < 0)
                       errlog("fcntl(F_SETFD): %s (%i)",strerror(errno),errno);
        	     child_new->rdp  = fdopen(pipefd1[0],"rt");
        	     child_new->wrp  = fdopen(pipefd2[1],"wt");
                     #endif
        	     if (child_new->rdp && child_new->wrp)
        	     { children++;
                       shandler_add(fileno(child_new->rdp),child_new,child_communication);
                       /*if (bounced_user && !strncmp(bounced_idnt,"AUTH",4))
                	 send_reply(cs,220,"Authorization OK, proceed login with USER");*/
        	     } else
        	     { errlog("fdopen(): %s (%i)",strerror(errno),errno);
                       close(pipefd1[0]);
                       #ifndef HAVE_SOCKETPAIR
                       close(pipefd2[1]);
                       #endif
        	     }
        	     break;
               } /* switch */
               close(cs);
            }
	 } 
	 else { /* not in ipranges */
            hp = gethostbyaddr((char*)&rem_addr.sin_addr,sizeof(rem_addr.sin_addr),
                 AF_INET);
            server_log_local(NULL,LG_SECURITY, "UNAUTHORIZED CONNECTION from %s (%s)",
                inet_ntoa(rem_addr.sin_addr),hp ? hp->h_name : "unknown host");
            run_alternative(cs);
         }
      } /* else */
   } /* if */
   else errlog("accept: %s (%i)",strerror(errno),errno);
   return 0;
}


int server_main(char* prg, int flags)
{ 
   struct sockaddr_in addr;
   int                i;
   char               name[32];
   char*              p;
   pid_t              pid;
   #ifndef HAVE_SIGSET
   struct sigaction   sig_act;
   #endif
   time_t             t;
   time_t             t_last;
   time_t             t_last_u;
   struct timeval     to;

#ifndef HAVE_SIGSET
   sigfillset(&sig_act.sa_mask);
   sig_act.sa_flags = SA_RESTART;
#define sigset(x,y) sig_act.sa_handler = y; sigaction(x,&sig_act,NULL)
#endif

#ifdef DEBUG
   debugf = fopen("debug.log", "wt");
   if (!debugf) { 
     printf("cannot open debug file\n");
     return -1;
   }
   setvbuf(debugf,NULL,_IOLBF,0);
#endif

   /* set startup time to calc uptime */
   startup_time = time(NULL);
   
   /* cd to server directory */
   if (getcwd(server_dir, sizeof(server_dir)) == NULL) { 
     perror("getcwd");
     return -1;
   }
   p = strrchr(prg,'/');
   if (p != NULL) { 
      *p = '\0';
      if (chdir(prg) == -1) { 
         *p = '/';
         perror("chdir");
         return -1;
      }
      *p = '/';
   }
   chdir("..");
   if (getcwd(server_dir,sizeof(server_dir)) == NULL) { 
      perror("getcwd");
      return -1;
   }
   if (strcmp(server_dir, "/"))
     strlcat(server_dir, "/", sizeof(server_dir));
   init_stats();

   /* load registry, group and user file */
   if (load_config(DFL_REGFILE,0,1) == -1) return -1;

   /* create initial superuser */
   if (flags & 0x02) restore_initial_superuser();
   if (!user_count) create_initial_superuser();

   /* load  section file. Return -1 if default.cfg is not found */
   if (load_global_section_config() == -1) return -1;

   /* load ipban files */
   if ((load_ip_file(IP_ALLOW) == -1) || (load_ip_file(IP_DENY) == -1)) {
      errlog("unable to load ipban files.");
      return -1;
   }

   /* g_reallocate ports */
   if (admin_port) { 
      admin_socket = get_socket(SOCK_STREAM,admin_port,4,&addr);
      if (admin_socket == -1) return errno;
      shandler_add(admin_socket,NULL,srv_accept);
   }
   if (!port && !port2) { 
      errlog("port or port2 must be set !");
      return -1;
   }
   if (port) { 
      ss = get_socket(SOCK_STREAM,port,16,&addr);
      if (ss == -1) return errno;
      shandler_add(ss,NULL,srv_accept);
   }
   if (port2) { 
      ss2 = get_socket(SOCK_STREAM,port2,16,&addr);
      if (ss2 == -1) return errno;
      shandler_add(ss2,NULL,srv_accept);
   }

   /* go daemon */
   if (flags & 0x05) pid = 1; 
   else { 
      if (logf) fflush(logf);
      if (errlogf) fflush(errlogf);
      pid = fork();
      if (pid < 0) { 
         errlog("fork: %s (%i)",strerror(errno),errno);
         return errno;
      }
   }
   if (flags & 0x08) { 
      if (pid)
	{ 
		if (flags & 0x0F && !(flags & 0x10))
		{
			freopen("/dev/tty","wt",stdin);
			gethostname(name,sizeof(name)-1);
			fprintf(stdin,"\e[32m|\e[0m Server on host %s is ready and listening"
			  " on %s:", name,other_ip ? other_ip : "*");
			if (port)
			{ 
	 			fprintf(stdin,"%i",port);
				if (port2) fprintf(stdin, ",%i", port2);
			}
	 		else fprintf(stdin, "%i", port2);
			fprintf(stdin, "\n");
			fprintf(stdin, "\e[32m|---------------------------------------------"
			   "-------------------------- ---  -\e[0m\n");
			print_configuration(stdin);
		}
	if (!(flags & 0x05)) exit(0);
      }
      /* release tty */
      freopen("/dev/null","rt",stdin);
      if (!(flags & 0x01)) setsid();
   }
   write_pidfile();

   /* set sig handlers */
   sigset(SIGHUP,  signal_handler);
   sigset(SIGINT,  signal_handler);
   sigset(SIGQUIT, signal_handler);
   sigset(SIGCHLD,  signal_handler);
   sigset(SIGPIPE, signal_handler);
   sigset(SIGTERM, signal_handler);
   sigset(SIGUSR1, signal_handler);
   sigset(SIGUSR2, signal_handler);
   errlog("ftp server started");
#ifdef DEBUG
   fprintf(debugf,"admin_fd=%i, srv1_fd=%i, srv2_fd=%i\n",admin_socket,ss,ss2);
#endif

   lim_read_config();
   write_stats(time(NULL));
   t = t_last_u = time(&t_last);

   /* daemon service loop */
   for (;;) { 
      handle_signals();
   #ifdef DEBUG
      fprintf(debugf,"calling select ... ");
   #endif
      fdset = shandler_fdset;
      to.tv_sec = 60;
      to.tv_usec = 0;
      i = select(FD_SETSIZE, &fdset, 0, 0, &to);
   #ifdef DEBUG
      fprintf(debugf,"select returned %i\n",i);
   #endif
      if (i < 0) { 
         if (errno == EINTR) handle_signals(); 
	 else { 
	    errlog("select: %s (%i)", strerror(errno), errno);
            raise(SIGTERM);
         }
      } 
      else { 
         if (i) { 
	    for (i=0; i<shandler_count; i++) {
	       if FD_ISSET(shandler[i].fd, &fdset) {
                  shandler[i].handler(shandler[i].fd,shandler[i].ptr);
               }
	    }
	 }
	 time(&t);
	 if (t-t_last >= 120) { 
            /* do every two minutes */
            if (registry_dirty) {
	       if (!save_registry(NULL)) registry_dirty = 0;
            }
	    if (group_dirty) {
	       if (!save_group_file(group_file)) group_dirty = 0;
            }
	    write_stats(t);
            t_last = t;
	 }
	 if (t-t_last_u >= 900) { 
            /* do every 15 minutes */
            if (user_dirty) {
	       if (!save_user_file(user_file)) user_dirty = 0;
            }
	    if (section_stat_dirty) {
	       if (!save_section_stat()) section_stat_dirty = 0;
            }
	    if (ipbans_dirty) {
	       if (save_ip_file(IP_ALLOW) != -1 && save_ip_file(IP_DENY) != -1) {
        	  ipbans_dirty = 0;
               }
            }
            t_last_u = t;
         }
      }
   }
}


int respawner(int flags)
{ 
   pid_t  pid;
   time_t t;
   time_t t_fork;
   pid_t  wpid;

   /* go daemon */
   pid = fork();
   if (pid < 0) exit(errno);
   if (pid) exit(0);

   /* fork off initial child */
   pid = fork();
   if (pid < 0) exit(errno);
   if (!pid) return flags | 0x08;

   /* release terminal */
   freopen("/dev/null", "rt", stdin);
   freopen("/dev/null", "wt", stdout);
   freopen("/dev/null", "wt", stderr);
   setsid();

   while (pid) { 
      time(&t_fork);
      /* wait for child to die */
      wpid = wait(&flags);
      if (wpid < 0) exit(0);
      if (pid == wpid) { 
         /* child died -> fork off new child (max. 1 fork per minute) */
	 time(&t);
	 if (t-t_fork < 60) sleep(t_fork+60-t);
	 pid = fork();
	 if (pid < 0) exit(errno);
      }
   }
   /* new forked-off child returns */
   return 0x04;
}


int syntax(char* prg)
{ 
   printf(
      "syntax: %s [-f] [-h] [-q] [-s] [-v] [-w]\n"
      "        -f = run in foreground\n"
      "        -h = print this help message\n"
      "        -q = quiet startup (even without errormsg)\n"
      "        -s = restore initial superuser\n"
      "        -v = print version information and exit\n"
      "        -w = watcher daemon restarts ftpd as soon as it crashes\n"
      ,prg);
   return -1;
}


int main(int argc, char** argv, char** env)
{ 
   int flags = 0;
   int i;

   while ((i = getopt(argc,argv,"fhqsvw")) > 0) {
      switch (i) { 
         case 'f' :
  	    flags |= 0x01;
	    break;
	 case 'h' :
	    syntax(argv[0]);
	    return 0;
	 case 'q' :
	    flags |= 0x10;
	    break;
	 case 's' :
	    flags |= 0x02;
	    break;
	 case 'v' :
	    g_print ("version is "VERSION"\n"
		     "Copyright (C) 2001 OpenFTPD Team\n");
	    return 0;
	 case 'w' :
	    flags |= 0x04;
	    break;
	 default  :
	    return syntax(argv[0]);
      }
   }
   if ((optind < argc) || ((flags & 0x05) == 0x05)) return syntax(argv[0]);
 #ifdef DES_ENCRYPTION
   /* read key */
   strlcpy(server_dir, getpass("DES key: "), sizeof(server_dir));
   string_to_key(server_dir, &user_key);
 #endif
   if (!(flags & 0x10)) printf(
	"\n"
	"\e[32m.----------------------------------------------------------------------- ---  -\e[0m\n"
	"\e[32m| \e[0mOpenFTPD " VERSION " [" HOST_ARCH "] (based on FTP4ALL by Crescent)\e[0m\n"
	"\e[32m| \e[0mThis program is FREE SOFTWARE and distributed under GNU PUBLIC LICENSE\e[0m\n"
	"\e[32m|----------------------------------------------------------------------- ---  -\e[0m\n"
	);
   if (flags & 0x04) flags = respawner(flags); 
   else flags |= 0x08;
   i = server_main(argv[0], flags);
   if (!(flags & 0x10))
   {
	freopen("/dev/tty", "wt", stdout);
	printf("ftpd[%ld] ERROR: Startup of server failed. See error log for "
	"details.\n", (long)getpid());
   }
   return i;
}

/* EOF */
