/*   group.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996-1999 by Crescent
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#ifdef DES_ENCRYPTION
  #include "des.h"
#endif
#include "externsd.h"
#include "log.h"
#include "utils.h"
#include "strlcpy.h"
#include "strlcat.h"

int group_dirty = 0;
#ifdef DES_ENCRYPTION
   extern C_Block user_key;
   void   des_set_key(C_Block*,Key_schedule*);
   int    des_cbc_encrypt(C_Block*,C_Block*,int,Key_schedule*,C_Block*,int);
#endif
  
gint32 get_gid(char* name)
{ 
   struct group_s* g = group;
   gint32           i;
   for (i = 0; i < group_count; i++) { 
      if (!strcasecmp(g->name,name)) return g->gid;
      g++;
   }
   if (sscanf(name,SI32,&i)) return i; 
   else return -1;
}


struct group_s* get_group_by_gid(gint32 gid)
{ 
   struct group_s* g = group;
   int             i;
   for (i=0; i<group_count; i++, g++) if (g->gid == gid) return g;
   return NULL;
}


struct group_s* get_group_by_name(char* name)
{ 
   struct group_s* g = group;
   /* gint32           gid; */
   int             i;
   for (i=0; i<group_count; i++, g++) {
      if (!strcasecmp(g->name, name)) return g;
   }
   /* if (sscanf(name, SI32, &gid)) return get_group_by_gid(gid); */
   return NULL;
}


int get_groupname(gint32 gid, char* name, size_t namelen)
{ 
   struct group_s* g = group;
   int             i;

   if (gid == -1) { 
      strlcpy(name, "nogroup", namelen);
      return 1;
   }

   for (i=0; i < group_count; i++, g++) {
      if (g->gid == gid) { 
         strlcpy(name, g->name, namelen);
         return 1;
      }
  }

  g_snprintf(name, namelen, SI32, gid);
  return 0;
}


int get_grp_umask(gint32 gid)
{ 
   struct group_s* g = group;
   int             i;
   for (i=0; i<group_count; i++, g++) {
      if (g->gid == gid) return g->umask_group;
   }
   return 0;
}


int get_group(char* p)
{ 
   char            c;
   char*           q;
   struct group_s* grp;

   if (group_count >= group_size) { 
      group_size += 4;
      grp = (struct group_s*)g_realloc(group,group_size*sizeof(struct group_s));
      if (!grp) { 
         group_size -= 4;
         errlog("get_group(): %s",strerror(errno));
         return -1;
      }
   } else grp = group;
   group = grp;
   grp = group+group_count;
   skipspace(&p);
   q = p;
   if (!skiptocolon(&q)) return -1;
   if (q-p<1 || q-p>=sizeof(grp->name)) return -1;
   strncpy(grp->name,p,q-p);
   *((grp->name)+(q-p)) = '\0';
   if (!skipspctocolon(&q)) return -1;
   p = q;
   skipnspace(&q);
   c  = *q;
   *q = '\0';
   if (!sscanf(p,SI32,&grp->gid)) return -1;
   *q = c;
   if (skipspace(&q)) return -1;
   /* new fields */
   grp->credit = 0;
   grp->limit = 0;
   grp->dl_speed = 0;
   grp->ul_speed = 0;
   grp->logins = 0;
   grp->umask_group = 0;
   grp->info[0] = 0;
   memset(grp->ident,0,sizeof(grp->ident));
   grp->max_logins = 0;
   grp->max_slots = 0; 
   grp->rights = 0;
   grp->temp3 = grp->temp = grp->temp2 = 0;
   /**************/
   group_count++;
   return 0;
}


int load_group_file_v223(char* filename, int force)
{ 
   FILE* file;
   char  line[256];
   int   lineno = 0;
   char* p;

   mk_filename(line,filename,DFL_GROUPFILE);
   file = fopen(line,"rt");
   if (file == NULL) { 
      if (!force) return 0;
      errlog("cannot open group file");
      return -1;
   }
   while (readline(line,sizeof(line),file)) { 
      lineno++;
      p = line;
      skipspace(&p);
      if (*p && (*p!='#')) if (get_group(p) == -1) { 
         errlog("load_group_file_v223(): syntax error on line %i (%s)", lineno, line);
         fclose(file);
         return -1;
      }
   }
   fclose(file);
   group_dirty = 0;
   return 0;
}


struct gf_ident_s
{ 
   char id[32];
   int  version;
   int  count;
   int  encrypted;
   int  temp1;
} g_id = { "OpenFTPD group database", 101, 0, 0, 0 };
/* Version set to 100 to avoid futures versions of official FTP4ALL */

int load_group_file_v3009(char* filename, int force)
{ 
   FILE*               file;
   char                fn_tmp[256];
   struct gf_ident_s   id_test;
   long                n;
   struct group_v3009* old_group = NULL;
   struct group_v3009* og = NULL;
   struct group_s*     g = NULL;
   gint32	       i;
 #ifdef DES_ENCRYPTION
   char*               buffer;
   C_Block             c;
   C_Block             out;
   Key_schedule        s;
 #endif

   mk_filename(fn_tmp,filename,DFL_GROUPFILE);
   file = fopen(fn_tmp,"rb");
   if (file == NULL) { 
      if (!force) return 0;
      errlog("fopen(%s): %s (%i)",fn_tmp,strerror(errno),errno);
      return -1;
   }
   if (!fread(&id_test,sizeof(id_test),1,file)) id_test.id[0] = 0;
   n = id_test.count;
   if (n > group_size) { 
      old_group = (struct group_v3009*) g_realloc(NULL,n*sizeof(struct group_v3009));
      if (!old_group) { 
         errlog("out of memory (group file)");
         fclose(file);
         return -1;
      }
      group_size = n;
   }
   if (fread(old_group,sizeof(struct group_v3009),n,file) != n) { 
      errlog("fread(group file(v3.009)): %s (%i)",strerror(errno),errno);
      group_count = 0;
      fclose(file);
      return -1;
   }
   group_count = n;
   fclose(file);
   if (id_test.encrypted) {
 #ifdef DES_ENCRYPTION
      c = user_key;
      des_set_key(&c,&s);
      buffer = (char*)old_group;
      n = group_count * sizeof(struct group_v3009);
      while (n >= 8) { 
         des_cbc_encrypt((C_Block*)buffer,&out,8,&s,&c,DES_DECRYPT);
         memmove(buffer,&out,8);
         buffer += 8;
         n -= 8;
      }
 #else
      group_count = 0;
      errlog("cannot read crypted group file");
      return -1;
 #endif
   }
   group = (struct group_s*)g_realloc(group,group_count*sizeof(struct group_s));
   memset(group,0,group_count*sizeof(struct group_s));
   g = group;
   og = old_group;
   for (i=0; i<group_count; i++) {
      memcpy(g++,old_group++,sizeof(struct group_v3009));
   }
   g_free(og);
   group_dirty = 1;
   return 0;
}

int load_group_file_v3012(char* filename, int force)
{
   FILE*               file;
   char                fn_tmp[256];
   struct gf_ident_s   id_test;
   long                n;
   struct group_v3012* old_group = NULL;
   struct group_v3012* og = NULL;
   struct group_s*     g;
   gint32               i;
 #ifdef DES_ENCRYPTION
   char*               buffer;
   C_Block             c;
   C_Block             out;
   Key_schedule        s;
 #endif

   mk_filename(fn_tmp,filename,DFL_GROUPFILE);
   file = fopen(fn_tmp,"rb");
   if (file == NULL) {
      if (!force) return 0;
      errlog("fopen(%s): %s (%i)",fn_tmp,strerror(errno),errno);
      return -1;
   }
   if (!fread(&id_test,sizeof(id_test),1,file)) id_test.id[0] = 0;
   n = id_test.count;
   if (n > group_size) {
      old_group = (struct group_v3012*) g_realloc(NULL,n*sizeof(struct group_v3012));
      if (!old_group) {
         errlog("out of memory (group file)");
         fclose(file);
         return -1;
      }
      group_size = n;
   }
   if (fread(old_group,sizeof(struct group_v3012),n,file) != n) {
      errlog("fread(group file(v3.012)): %s (%i)",strerror(errno),errno);
      group_count = 0;
      fclose(file);
      return -1;
   }
   group_count = n;
   fclose(file);
   if (id_test.encrypted) { 
 #ifdef DES_ENCRYPTION
      c = user_key;
      des_set_key(&c,&s);
      buffer = (char*)old_group;
      n = group_count * sizeof(struct group_v3012);
      while (n >= 8) {
         des_cbc_encrypt((C_Block*)buffer,&out,8,&s,&c,DES_DECRYPT);
         memmove(buffer,&out,8);
         buffer += 8;
         n -= 8;
      }
 #else
      group_count = 0;
      errlog("cannot read crypted group file");
      return -1;
 #endif
   }
   group = (struct group_s*)g_realloc(group,group_count*sizeof(struct group_s));
   memset(group,0,group_count*sizeof(struct group_s));
   g = group;
   og = old_group;
   for (i=0; i<group_count; i++) { 
      memcpy(g++,old_group++,sizeof(struct group_v3012));
   }
   g_free(og);
   group_dirty = 1;
   return 0;
}

int load_group_file_v1(char* filename, int force)
{
   FILE*               file;
   char                fn_tmp[256];
   struct gf_ident_s   id_test;
   long                n;
   struct group_v1*    old_group = NULL;
   struct group_v1*    og;
   struct group_s*     g;
   gint32               i;
 #ifdef DES_ENCRYPTION
   char*               buffer;
   C_Block             c;
   C_Block             out;
   Key_schedule        s;
 #endif

   mk_filename(fn_tmp,filename,DFL_GROUPFILE);
   file = fopen(fn_tmp,"rb");
   if (file == NULL) {
      if (!force) return 0;
      errlog("fopen(%s): %s (%i)",fn_tmp,strerror(errno),errno);
      return -1;
   }
   if (!fread(&id_test,sizeof(id_test),1,file)) id_test.id[0] = 0;
   n = id_test.count;
   if (n > group_size) {
      old_group = (struct group_v1*) g_realloc(NULL,n*sizeof(struct group_v1));
      if (!old_group) {
         errlog("out of memory (group file)");
         fclose(file);
         return -1;
      }
      group_size = n;
   }
   if (fread(old_group,sizeof(struct group_v1),n,file) != n) {
      errlog("fread(group file(v1)): %s (%i)",strerror(errno),errno);
      group_count = 0;
      fclose(file);
      return -1;
   }
   group_count = n;
   fclose(file);
   if (id_test.encrypted) { 
 #ifdef DES_ENCRYPTION
      c = user_key;
      des_set_key(&c,&s);
      buffer = (char*)old_group;
      n = group_count * sizeof(struct group_v1);
      while (n >= 8) {
         des_cbc_encrypt((C_Block*)buffer,&out,8,&s,&c,DES_DECRYPT);
         memmove(buffer,&out,8);
         buffer += 8;
         n -= 8;
      }
 #else
      group_count = 0;
      errlog("cannot read crypted group file");
      return -1;
 #endif
   }
   group = (struct group_s*)g_realloc(group,group_count*sizeof(struct group_s));
   memset(group,0,group_count*sizeof(struct group_s));
   g = group;
   og = old_group;
   for (i=0; i<group_count; i++) { 
      memcpy(g++,old_group++,sizeof(struct group_v1));
   }
   g_free(og);
   group_dirty = 1;
   return 0;
}

int load_group_file(char* filename, int force)
{ FILE*             file;
  char              fn_tmp[256];
  struct gf_ident_s id_test;
  long              n;
  struct group_s*   new_group;
  int               new_size;
#ifdef DES_ENCRYPTION
  char*             buffer;
  C_Block           c;
  C_Block           out;
  Key_schedule      s;
#endif

  mk_filename(fn_tmp,filename,DFL_GROUPFILE);
  file = fopen(fn_tmp,"rb");
  if (file == NULL)
  { if (!force) return 0;
    errlog("fopen(%s): %s (%i)",fn_tmp,strerror(errno),errno);
    return -1;
  }
  if (!fread(&id_test,sizeof(id_test),1,file)) id_test.id[0] = 0;
  if (strcmp(id_test.id,g_id.id) && strcmp(id_test.id,"FTP4ALL group database"))
  { errlog("group file has bad ID, assuming pre 2.23 group file");
    fclose(file);
    return load_group_file_v223(filename,force);
  }
  if (id_test.version > g_id.version)
  { errlog("group file is newer version than daemon");
    fclose(file);
    return -1;
  }
  
  if (id_test.version < g_id.version)
  { 
    errlog("group file is older version than daemon");
    fclose(file);
    if (id_test.version == 0) {
        errlog("group file is 3.009 version");
	return load_group_file_v3009(filename,force);
    }
    else if (id_test.version == 1) {
        errlog("group file is 3.012 version");
        return load_group_file_v3012(filename,force);
    }
    else if (id_test.version == 100) {
        errlog("group file is OpenFTPD 1.0 version");
        return load_group_file_v1(filename,force);        
    } else {
    return -1;
    }
  }
  n = id_test.count;
  if (n >= group_size)
  { new_size = n + 8;
    new_group = (struct group_s*)g_realloc(group,new_size*sizeof(struct group_s));
    if (!new_group)
    { errlog("out of memory (group file)");
      fclose(file);
      return -1;
    }
    group_size = new_size;
    group = new_group;
  }
  group_count = 0;
  if (fread(group,sizeof(struct group_s),n,file) != n)
  { errlog("fread(group file): %s (%i)",strerror(errno),errno);
    fclose(file);
    return -1;
  }
  group_count = n;
  fclose(file);
  if (id_test.encrypted)
  {
#ifdef DES_ENCRYPTION
    c = user_key;
    des_set_key(&c,&s);
    buffer = (char*)group;
    n = group_count * sizeof(struct group_s);
    while (n >= 8)
    { des_cbc_encrypt((C_Block*)buffer,&out,8,&s,&c,DES_DECRYPT);
      memmove(buffer,&out,8);
      buffer += 8;
      n -= 8;
    }
#else
    group_count = 0;
    errlog("cannot read crypted group file");
    return -1;
#endif
  }
  group_dirty = 0;
  return 0;
}

int save_group_file(char* filename)
{ FILE*        file;
  char         fn_tmp[256];
  char         fn_tmp2[256];
  size_t       result;
#ifdef DES_ENCRYPTION
  char*        buffer;
  C_Block      c;
  long         n;
  C_Block      out;
  Key_schedule s;
#endif

  mk_filename(fn_tmp,filename,DFL_GROUPFILE);
  strlcpy(fn_tmp2, fn_tmp, sizeof(fn_tmp2));
  strlcat(fn_tmp, ".tmp", sizeof(fn_tmp));
  file = fopen(fn_tmp,"wb");
  if (file == NULL)
  { errlog("fopen(%s): %s (%i)",fn_tmp,strerror(errno),errno);
    return -1;
  }
  g_id.count = group_count;
#ifdef DES_ENCRYPTION
  g_id.encrypted = flag_get(flags,FLAG_DESENCRYPTION);
#endif
  if (!fwrite(&g_id,sizeof(g_id),1,file))
  { errlog("fwrite(%s): %s (%i)",fn_tmp,strerror(errno),errno);
    fclose(file);
    return -1;
  }
#ifdef DES_ENCRYPTION
  if (g_id.encrypted)
  { c = user_key;
    des_set_key(&c,&s);
    buffer = (char*)group;
    n = group_count * sizeof(struct group_s);
    while (n >= 8)
    { des_cbc_encrypt((C_Block*)buffer,&out,8,&s,&c,DES_ENCRYPT);
      memmove(buffer,&out,8);
      buffer += 8;
      n -= 8;
    }
  }
#endif
  result = fwrite(group,sizeof(struct group_s),group_count,file);
  if (result != group_count)
  { errlog("fwrite(%s): %s (%i)",fn_tmp,strerror(errno),errno);
    fclose(file);
  } else
  { fclose(file);
    if (rename(fn_tmp,fn_tmp2) < 0)
      errlog("rename(%s,%s): %s (%i)",fn_tmp,fn_tmp2,strerror(errno),errno);
  }
#ifdef DES_ENCRYPTION
  if (g_id.encrypted)
  { c = user_key;
    des_set_key(&c,&s);
    buffer = (char*)group;
    n = group_count * sizeof(struct group_s);
    while (n >= 8)
    { des_cbc_encrypt((C_Block*)buffer,&out,8,&s,&c,DES_DECRYPT);
      memmove(buffer,&out,8);
      buffer += 8;
      n -= 8;
    }
  }
#endif
  return result == group_count ? 0 : -1;
}

/* EOF */
