/*   reject.c
 *
 *   This file is part of FTP4ALL
 *
 *   Copyright (C) 1996-1999 by Crescent
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *                       
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include "externsd.h"
#include "limits.h"
#include "log.h"
#include "user.h"
#include "utils.h"

char* reject      = NULL;
int   reject_cnt  = 0;
int   reject_size = 0;

int reject_add(char* newrange)
{ char* ptr;
  int   new_size;

  if (reject_cnt >= reject_size)
  { new_size = reject_cnt + 16;
    ptr = g_realloc(reject,new_size*8);
    if (!ptr) return -1;
    reject = ptr;
    reject_size = new_size;
  }
  memmove(reject+(8*reject_cnt++),newrange,8);
  return 0;
}

int reject_parse_one(char* range)
{ int  i;
  char iprange[20];

  i = get_iprange(&range,iprange);
  if (!i) i = reject_add(iprange+12);
  return i;
}

int reject_parse(char* ranges)
{ int   i;
  char* p;
  char* q;
  char  range[32];

  p = ranges;
  reject_cnt = 0;
  if (p) while (*p)
  { skipspace(&p);
    q = p;
    skipnspace(&q);
    i = q - p;
    if (i && (i < sizeof(range)))
    { strncpy(range,p,i);
      range[i] = '\0';
      reject_parse_one(range);
    }
    p = q;
  }
  return 0;
}

int reject_is_rejected_ip(unsigned long ip, int log)
{ struct in_addr  addr;
  struct hostent* hp;
  int             i;
  char*           r;

  for (i=0, r=reject; i<reject_cnt; i++, r+=8) if (in_iprange(ip,(unsigned char*)r))
  { if (log)
    { addr.s_addr = htonl(ip);
      hp = gethostbyaddr((char*)&addr,sizeof(addr),AF_INET);
      server_log_local(NULL,LG_SECURITY, "REJECTED CONNECTION from %s (%s)",
        inet_ntoa(addr),hp ? hp->h_name : "unknown host");
    }
    return 1;
  }
  return 0;
}

/* EOF */
