/* ---------------------------------------------------------------------------
 *
 *   This file is part of OpenFTPD (based on FTP4ALL)
 *   (C) 1996-2000 by Crescent, (C) 2000 by the OpenFTPD Team
 *
 * ---------------------------------------------------------------------------
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 * ---------------------------------------------------------------------------
 *   MODULE: commands.c - basic ftp commands
 * ---------------------------------------------------------------------------
 */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <dirent.h>
#include "externs.h"
#include "common.h"
#include "log.h"
#include "perm.h"
#include "utils.h"
#include "list.h"
#include "cmd_protos.h"

#include "site_vfs.h"

CMD_UNIMPLEMENTED (cmd_smnt, SMNT);
CMD_UNIMPLEMENTED (cmd_rein, REIN);
CMD_UNIMPLEMENTED (cmd_allo, ALLO);
CMD_UNIMPLEMENTED (cmd_stou, STOU);

/* RNTO (rename to) command */
int cmd_rnto (gchar **p, gint n, gchar *buf, gint buflen) 
{
	return send_reply (503, "Bad sequence of commands, RNFR not issued.");
}


/* RNFR (raname from) command */
int cmd_rnfr(char** p, int n, char* str, int len)
{ 
	char*       dir;
	char*       name;
	char*       name_alt;
	char        new[256];
	int         ok;
	int         perm;
	struct stat st;
	char        wd[256];
	// vfs_stat_t  vst;    

	if (n != 1) return send_reply(501, "Syntax: RNFR <filename>");
	ok = add_dir (vfs_getcwd (), *p, wd, sizeof(wd));
	if (user_locked && *usr.home) 
	{
		if (strncmp(usr.home, wd, strlen(usr.home))) 
		{
			return send_reply(550, "!Bno permission.!0");
		}
	}

	if (ok != -1) 
	{
		split_dir(wd, &dir, &name, 1);
		ok = permission_ok(dir, PE_CHDIR);
		perm = ok < 0 ? 0 : ok;
		if (ok != -1) 
		{
			ok = file_permission_ok(dir, name, FPE_WRITE);
		}
		if (ok != -1) 
		{
			if (*name) 
			{
				*(name-1) = '/';
				name_alt = name;
			} 
			else 
			{
				name_alt = NULL;
			}
			ok = stat(wd+1, &st);
			if (ok != -1) 
			{
				if (send_reply(350, "File exists, ready for destination name.") == -1) 
				{
					return -1;
				}
				switch (read_ftp_command(str, len, p, 4, &n)) 
				{
					case -1 : return -1;
					case  0 : break;
					default : return 0;
				}
				if (strcasecmp(*p, "RNTO")) 
				{
					return send_reply(503, "Bad sequence of commands, RNTO expected.");
				}
				if (n != 2) 
				{
					return send_reply(501,"Syntax: RNTO <filename>");
				}
				ok = add_dir(vfs_getcwd (), *(p+1), new, sizeof(new));
			}
			if (ok != -1) 
			{
				split_dir(new, &dir, &name,1);
				if (S_ISDIR(st.st_mode)) 
				{       
					/* We dont check the parent dir */
					ok = permission_ok(wd, PE_WRITE); /* We check ext one */
				}
				else 
				{
					ok = permission_ok(dir,PE_WRITE);
				}
				if (ok > 0) perm = ok;
			}
			if (ok != -1) 
			{
				if (*name) *(name - 1) = '/'; 
				else if (name_alt) ok = add_dir(new, name_alt, new, sizeof(new));
				if (ok != -1) 
				{
					perm = directory_flag_set_recursive(wd, DP_HIDDEN);
					ok = rename(wd+1, new+1);
		    			if (ok < 0) 
					{
						log_action(perm, LG_FILE, "RENAME","%s to %s FAILED: %s",
					   		wd, new, strerror(errno));
					}
					else 
					{ 
						log_action(perm, LG_FILE, "RENAME", "%s to %s", wd, new);
					}
				}
			} 
		}
	}
	if (ok == -1) return send_error(553,"rename",errno);
   
	if (!path_in_cdpath(wd)) 
	{
		printf("FRN- %s\n",wd);
		fgets(wd, sizeof(wd), stdin);
		if (*wd == '+') 
		{
			printf("%s\n", new);
			fgets(wd,sizeof(wd),stdin);
		}
	}
	return send_reply(250, "RNTO command successful.");
}


/* ABOR command */
int cmd_abor (gchar **p, gint n, gchar *_buffer, gint _buflen)
{ 
	return send_reply(226, "ABOR command successful.");
}


int cmd_dele_check(char* path, char* name)
{
   struct check_s* ch = delcheck;
   int             i;
   char*           p[8];
   char            buffer[4096];

   /* Same size that the buffer send to cmd_stor_dcheck (transfer.c) */
   g_snprintf(buffer, sizeof(buffer), "%s/%s", path, name);
   for (i = 0; i < delcheck_count; i++, ch++) {
      if (wildcard_matching(ch->wildcard, name, 0)) {
        i = 0;
        p[i++] = ch->name;
         if (*ch->param) p[i++] = ch->param;
         p[i++] = buffer + 1;
         p[i++] = usr.name;
         if (run_program(ch->name, p, i, 0, 0, NULL, &i, buffer, sizeof(buffer),
             path ? path : "/") == -1) 
            return 0;
         if (WIFEXITED(i)) {
            if (WEXITSTATUS(i) && (WEXITSTATUS(i) != -1)) i = 550;
            else i = 257;
         }
         else i = 257;
         if ((i == 550) && !*buffer) strcpy(buffer,"Delete check failed."); 
         if (*buffer) if (send_message(i,i == 550,buffer,NULL) == -1) return -1;
         return i == 550 ? 1 : 0;
      }
   }
   return 0;
} 


int cmd_dele (gchar **p, gint n, gchar *_buffer, gint _buflen)
{ 
   char*       dir;
   char*       name;
   int         ok;
   gint32       perm;
   int         permok;
   int         section_id;
   gint32       rf_gid;
   gint32       rf_uid;
   struct stat st;
   char        wd[256];
   char        str[256];
   int         time;
   int         undupe = 0;

   if (n != 1) return send_reply(501,"Syntax: DELE <filename>");
   ok = add_dir(vfs_getcwd (), *p, wd, sizeof(wd));
   if (user_locked && *usr.home) {
      if (strncmp(usr.home, wd, strlen(usr.home)))
         return send_reply(550, "!Bno permission.!0");
   }
   printf("SGV_DIR internal_dupecheck %s\n", wd);
   if (!fgets(str, sizeof(str), stdin))
      undupe = 0;
   else if (*(str + 1) == '1')
      undupe = 1;

   if (undupe) {
      printf("SGV_DIR undupe_on_delete %s\n", wd);
      if (!fgets(str, sizeof(str), stdin) || *(str + 1) == '0') 
         undupe = 0;
      else 
         undupe = 1;
   }
   if (ok != -1) {
      split_dir(wd, &dir, &name, 0);
      if (cmd_dele_check(dir, name)) return 0;
      ok = permission_ok(dir, PE_CHDIR);
      permok = ok < 0 ? 0 : ok;
      if (ok != -1) ok = file_permission_ok(dir, name, FPE_WRITE);
      if (ok != -1) {
         *(name-1) = '/';
         if (stat(wd+1, &st) < 0) st.st_size = 0;
         permok = directory_flag_set_recursive(wd, DP_HIDDEN);
         ok = unlink(wd+1);
         if (ok < 0) {
	    log_action(permok,LG_FILE,"DELETE","%s FAILED: %s",
               wd,strerror(errno));
         }
	 else {
	    log_action(permok,LG_FILE,"DELETE","%s (%lu bytes)", wd, 
	       st.st_size);
           
	    /***********************************
	     * mark file as deleted in dupelog *
	     ***********************************/
	    printf("DUP_DEL \"%s\"\n", wd);
            if (!fgets(str, sizeof(str), stdin)) {
	       errlog("pipe error on DUP_DEL in cmd_dele()");
	    }	 
            if (undupe) {
               printf("DUP_UNN -x \"%s\"\n", wd);
               if (!fgets(str, sizeof(str), stdin)) {
	          errlog("pipe error on DUP_UND in cmd_dele()");
	       }	 
            }
	 }
      }
   }
   if (ok == -1) return send_error(553, "delete", errno);
   read_file_permission(dir ? dir : "/",name, &rf_uid, &rf_gid, &perm);
   if (!path_in_cdpath (wd)) 
     {
       printf("FRM- %s\n",wd);
       fgets(str,sizeof(str), stdin);
     }
   
    /* don't remove credits for superusers and siteops on delete if 
       if they are not owner of file! */
   if ((rf_gid != -2) && st.st_size && (!superuser() || (superuser() &&
      usr.uid == rf_uid)) && (!siteop() || (siteop() && usr.uid == rf_uid))) {

      printf("SG_ID %s\n", wd);
      if (!fgets(str, sizeof(str), stdin)) {
         send_message(550, 0, "!Bpipe error!0", NULL);
         return -1;
      }
      else if (!sscanf(str + 1, "%d", &section_id)) {
         section_id = -1; 
      }

      time = get_time(st.st_mtime);
      if (!time)
         printf("UUR %s %u 1 1 1 1 1 %d\n",usr.name, (unsigned int)st.st_size,
        	section_id);
      else if (time == 1)
         printf("UUR %s %u 0 1 1 1 1 %d\n",usr.name, (unsigned int)st.st_size,
        	section_id);
      else if (time == 2)
         printf("UUR %s %u 0 0 1 1 1 %d\n",usr.name, (unsigned int)st.st_size,
        	section_id);
      else if (time == 3)
         printf("UUR %s %u 0 0 0 1 1 %d\n",usr.name, (unsigned int)st.st_size,
        	section_id);
      else 
         printf("UUR %s %u 0 0 0 0 1 %d\n",usr.name, (unsigned int)st.st_size,
        	section_id);
      fgets(wd, sizeof(wd), stdin);
   }
   return send_message(250, 1, dele_msg.msg, *p);
}


/* SYST command */
int cmd_syst (gchar **p, gint n, gchar *_buffer, gint _buflen)
{ 
	if (n) return send_reply(501,"Syntax: SYST"); 
	else return send_reply(215,"UNIX Type: L8");
}


/******************************************************************************
 Handles the STAT command. If called with a parameter it lists the current 
 directory over the control connection (called stealth mode in pftp) without
 any parameter it shows some status information.
******************************************************************************/
int cmd_stat (gchar **p, gint n, gchar *_buffer, gint _buflen)
{ 
	const gchar *list_output;

	if (n == 0)
	{
		return send_message(211,1,
			"FTP status of server OpenFTPD "VERSION":\n"
			"Logged in as %s from %i (%h)\n"
			"There are %a anonymous (max. %A)\n"
			"and %u registered users (max. %U) logged in\n"
			"You uploaded %p and downloaded %d in the current session\n"
			"You uploaded %P and downloaded %D in all previous sessions\n"
			"Ratio of current session is %r%%, total is %R%%\n"
			"Your limit is %L, you have %c credit", usr.name);
	}
	else 
	{
		/*
		if (!perl_call_subf ("FTP::event", "si", "list_pre", 212))
		{
			return 0;
		}
		*/
  
		if (n)
		{
			list_output = (gchar*) perl_call_subf_str ("FTP::cmd_stat", "s", p[0]);
		}
		else
		{
			list_output = (gchar*) perl_call_subf_str ("FTP::cmd_stat", NULL);
		}
		send_reply_stat(213, (char*) list_output);
	
		/* perl_call_subf ("FTP::event", "si", "list_post", 226); */
		return 0;
	}
}


/* HELP command */
int cmd_help (gchar **p, gint n, gchar *_buffer, gint _buflen)
{ 
	return send_reply(214,
		"The following commands are recognized (* =>'s unimplemented).\n"
		"USER    PORT    STOR    MSAM*   RNTO    NLST    MKD     CDUP\n"
		"PASS    PASV    APPE    MRSQ*   ABOR    SITE    XMKD    XCUP*\n"
		"ACCT*   TYPE    MLFL*   MRCP*   DELE    SYST    RMD     STOU*\n"
		"SMNT*   STRU    MAIL*   ALLO    CWD     STAT    XRMD    SIZE\n"
		"REIN*   MODE    MSND*   REST    XCWD    HELP    PWD     MDTM\n"
		"QUIT    RETR    MSOM*   RNFR    LIST    NOOP    XPWD    IDLE*\n"
		"Direct comments to the site administrator.");
}


/* NOOP command */
int cmd_noop (gchar **p, gint n, gchar *_buffer, gint _buflen)
{ 
	return send_reply(200,"NOOP command successful.");
}

/* EOF */
