/* ---------------------------------------------------------------------------
 *
 *   This file is part of OpenFTPD (based on FTP4ALL)
 *   (C) 1996-2000 by Crescent, (C) 2000 by the OpenFTPD Team
 *
 * ---------------------------------------------------------------------------
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details..
 *
 * ---------------------------------------------------------------------------
 *   MODULE: dir.c - directory related functions
 * ---------------------------------------------------------------------------
 */

#include <sys/types.h>
#include <sys/param.h>
#include <dirent.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include "externs.h"
#include "common.h"
#include "list.h"
#include "log.h"
#include "perm.h"
#include "utils.h"

#include "site_vfs.h"
#include "perl_layer.h"
#include "strlcpy.h"
#include "strlcat.h"

int cmd_mkd_dirdcheck(char* dir, char* name, char* buffer, int bufsize)
{ 
   struct check_s* ch = dirdcheck;
   int             i;
   char*           p[8];
   char            str[1024];

   if (dir) g_snprintf(str, sizeof(str), "%s/%s", dir, name);
   else g_snprintf(str, sizeof(str), "/%s", name);
   for (i = 0; i < dirdcheck_count; i++, ch++) {
      if (wildcard_matching(ch->wildcard, name, 0)) { 
      	 i = 0;
      	 p[i++] = ch->name;
         if (*ch->param) p[i++] = ch->param;
         p[i++] = str;
         if (run_program(ch->name, p, i, 0, 0, NULL, &i, buffer, bufsize, dir ?
             dir : "/") == -1) return 0; 
	 if (WIFEXITED(i)) { 
            if (WEXITSTATUS(i) && (WEXITSTATUS(i) != -1)) return 1; 
         } 
      } 
   }
   return 0;
}

int show_readmefile(void)
{ FILE* file;
  char  line[256];

  if (readme_file) if ((file = file_open(vfs_getcwd (),readme_file,"rt",0)) != NULL)
  { while (fgets(line,sizeof(line),file))
    { if (send_message(250,0,line,vfs_getcwd ()) == -1)
      { fclose(file);
        return -1;
      }
    }
    fclose(file);
  }
  return 0;
}

#ifdef FUCK
void is_link(char *fullp, char *split, int size)
{ 
   char link[MAXPATHLEN];
   char *p;

   memset(split, 0, size);
   while(strcmp(fullp,"")) {
      memset(link, 0, sizeof(link));
      if (readlink(fullp, link, sizeof(link))!=-1) {
         if (link[0] != '/') {
            *(strrchr(fullp,'/')+1) = 0;
            strcat(fullp, link);
         } 
         else 
            strcpy(fullp,link);
         strcat(fullp, split);
         strcpy(split, "");
      } 
      else {
         strlcpy(link, strrchr(fullp,'/'), sizeof(link));
         strlcat(link, split, sizeof(link));
         strlcpy(split, link, sizeof(link));
         *strrchr(fullp,'/') = 0;
      }
   }
  /* split var should'n have now sym links
     Now strip .. or . from split (Usefull when someone made a link
     ../../.... out of sitedir path) */
   while ((p = strstr(split,"/..")) != NULL) {
      *p = 0;
      strlcpy(link, p+3, sizeof(link));
      *strrchr(split,'/') = 0;
      strcat(split,link);
   }
   while ((p = strstr(split,"/.")) != NULL) {
      *p=0;
      strcat(split,p+2);
   }
}
#endif

int find_matching_dir(char* path, char* wildcard)
{ struct dirent* d;
  DIR*           dir;
  char*          name;
  struct stat    st;

  dir = opendir(strcmp(path,"/") ? path+1 : ".");
  if (!dir) return 0;
  if (strcmp(path,"/")) strcat(path,"/");
  name = path + strlen(path);
  while ((d = readdir(dir)) != NULL)
  { strcpy(name,d->d_name);
    if (strcmp(name,".") && strcmp(name,"..") && !stat(path+1,&st) &&
      S_ISDIR(st.st_mode))
    { if (wildcard_matching(wildcard,name,0))
      { closedir(dir);
        return 1;
      }
    }
  }
  closedir(dir);
  return 0;
}

char* is_shortcut(char* shortcutpath)
{ int n;

  n = 0;
  while (n < dshortcut_len)
  { if (!strcasecmp(shortcutpath,dshortcut+n))
    { n += strlen(dshortcut+n) + 1;
      return dshortcut + n;
    }
    n += strlen(dshortcut+n) + 1;
    n += strlen(dshortcut+n) + 1;
  }
  return NULL;
}

/**** 
 * print directory stats is now handled by events.pl on chdir 
  
struct dirstat_s* alloc_dirstat(struct dirstat_s *dirstat, gint64 *dirstat_size)
{ 
   void *ptr;
   ptr = g_realloc(dirstat, (*dirstat_size + 1) * sizeof(struct dirstat_s));
   if (!ptr) return NULL;
   (*dirstat_size)++;
   return ptr;
}

void sort_dirstat(struct dirstat_s *dirstat, int dirstat_count)
{ 
   struct dirstat_s  dstmp;
   struct dirstat_s  *ds;
   int               n;
   int               sorted = 1;
  
   ds = &dstmp;
   while (sorted) {
      sorted = 0;
      for (n = 0; n < dirstat_count - 1; n++) {
         if ((dirstat+n)->size < (dirstat+n+1)->size) {
            memmove(ds,(dirstat+n), sizeof(struct dirstat_s));
            memmove((dirstat+n), (dirstat+n+1), sizeof(struct dirstat_s));
            memmove((dirstat+n+1), ds, sizeof(struct dirstat_s));
            sorted=1;
         }
      }
   } 
}

int print_directory_stats() 
{
    DIR              *dirp;
    struct dirent    *de;
    gint32           uid;
    gint32           gid;
    gint32           perm;
    char             buffer[256];
    char             fullpath[256];
    char             basedir[256];

    struct dirstat_s *dirstat = NULL;
    gint32           dirstat_count = 0;
    gint64           dirstat_size = 0;
    gint64           filesize = 0;
    struct stat      st;        
    gint             found = 0;  
    gint             i;
    gint64           totalsize = 0;
    gint32           totalfiles = 0;
    gchar            *username;
    gchar            *groupname;
    gfloat           percent = 0.0;
   
    getcwd(basedir, sizeof(basedir));
    sprintf(fullpath, "%s%s", basedir, vfs_getcwd ());

    if (NULL == (dirp = opendir(fullpath))) {
	return 0;
    }
    while (NULL != (de = readdir(dirp))) {      
	if (read_file_permission(vfs_getcwd (), de->d_name, &uid, &gid, &perm) != -1) {
	    if (uid >= 0 && gid >= 0) {
		sprintf(fullpath, "%s%s/%s", basedir, vfs_getcwd (), de->d_name);
		stat(fullpath, &st);
		if (S_ISREG(st.st_mode) && de->d_name[0] != '.') {
		    filesize = (gint64) st.st_size;
		    for (i = 0; i < dirstat_count; i++) {
			if ((dirstat+i)->uid == uid) {
			    found = 1;
			    break;
			}
		    }
		    if (found) {
			(dirstat+i)->uid = uid;
			(dirstat+i)->gid = gid;
			(dirstat+i)->files++;
			(dirstat+i)->size += filesize;
			totalsize += filesize;
			totalfiles++;
		    }
		    else {
			if (dirstat_count >= dirstat_size) { 
			    dirstat = alloc_dirstat(dirstat, &dirstat_size);
			    if (dirstat_count >= dirstat_size) { 
				errlog("out of memory (dirstats)");
				return 0;
			    }
			}
			(dirstat+dirstat_count)->uid = uid;
			(dirstat+dirstat_count)->gid = gid;		     
			(dirstat+dirstat_count)->files = 1;		     
			(dirstat+dirstat_count)->size = filesize;
			totalsize += filesize;
			totalfiles++;
			dirstat_count++;
		    }
		    found = 0;
		}    
	    }  
	}
    }
    closedir(dirp);
    sort_dirstat(dirstat, dirstat_count);
    if (totalfiles > 0) {
	send_message(250, 0, "!C.-------------------------------------------------------------------------.!0", NULL);         
	sprintf(buffer,      "!C|!3!H !BOpenFTPD !HRace Info: %-51.51s !0!C|!0", vfs_getcwd ());
	send_message(250, 0, buffer, NULL);      
	send_message(250, 0, "!C|-------------------------------------------------------------------------|!0", NULL);      
	for (i = 0; i < dirstat_count; i++) {
	    username = get_username((gint32)(dirstat+i)->uid);
	    groupname = get_groupname((gint32)(dirstat+i)->gid);
	    percent = totalsize > 0 ? (float)((dirstat+i)->size * 1.0 / totalsize * 1.0) * 100.0 : 0.0;

	    if (totalsize > 0x100000) {
		sprintf(buffer, "!C|!H [%02d] %-10.10s  %-10.10s                     %4dF  %5.0f MB  %5.1f%% !C|!0", 
			i + 1, username, groupname, (dirstat+i)->files, (float)(dirstat+i)->size / 0x100000 * 1.0,
			percent);
	    }
	    else {
		sprintf(buffer, "!C|!H [%02d] %-10.10s  %-10.10s                     %4dF  %5.0f kB  %5.1f%% !C|!0", 
			i + 1, username, groupname, (dirstat+i)->files, (float)(dirstat+i)->size / 1024.0,
			percent);
	    }
	    send_message(250, 0, buffer, NULL);      
	}
	send_message(250, 0, "!C|-------------------------------------------------------------------------|!0", NULL);      
	if (totalsize > 0x100000) {
	    sprintf(buffer, "!C|!H Total:                                          %4dF  %5.0f MB  100.0%% !C|!0", 
		    totalfiles, (float) totalsize / 0x100000 * 1.0);
	}
	else {
	    sprintf(buffer, "!C|!H Total:                                          %4dF  %5.0f kB  100.0%% !C|!0", 
		    totalfiles, (float) totalsize / 1024.0);
	}
	send_message(250, 0, buffer, NULL);      
	send_message(250, 0, "!C`-------------------------------------------------------------------------+!0", NULL);      
    }
    return 0;
}
*****/

int cmd_cwd(char** p, int n)
{
  const gchar *newpath = NULL;
  gboolean hidden = FALSE;
  gchar str[256];
  gchar wd[OPENFTPD_PATH_MAX];
  // gchar *dir, *name;  /* what is this good for soulshock ? */

  if (n != 1)
    return send_reply(501, "Syntax: CWD <directory>");

  // split_dir(p[0], &dir, &name, 0);
  
  if (!(newpath = is_shortcut (p[0]))) {
    newpath = p[0];
  }

  if (vfs_setcwd (newpath))
    return send_reply (550, vfs_strerror (vfs_errno));
  
  if (show_readmefile() == -1) 
    return -1;

  printf("SGV_DIR show_dirstat %s\n", vfs_getcwd ());
  if (!fgets(str, sizeof(str), stdin)) 
    strncpy(str, "-pipe error", sizeof(str) - 1); 
/**  else if (strncmp(str, "+0", 2) != 0)
    print_directory_stats();	**/

  hidden = directory_flag_set_recursive(vfs_getcwd (), DP_HIDDEN);  
  update_user_activity(hidden, ACT_IDLE, 0, vfs_getcwd ());
  
  {
    /* fixme */
    gchar str[256] = { 0, };
    set_enviroments(str, sizeof(str), NULL);
  }

  perl_call_subf ("FTP::event", "si", "chdir_post", 250);
  return 0;
}
 
gint cmd_cdup(char** p, int n)
{ 
  gboolean hidden = FALSE;

  if (n) 
    return send_reply(501,"Syntax: CDUP");

  if (vfs_setcwd (".."))
    return send_reply (550, vfs_strerror (vfs_errno));

  if (show_readmefile() == -1)
    return -1;

  hidden = directory_flag_set_recursive(vfs_getcwd (), DP_HIDDEN);  
  update_user_activity(hidden, ACT_IDLE, 0, vfs_getcwd ());

  {
     /* fixme */
     gchar str[256] = { 0, };
     set_enviroments(str, sizeof(str), NULL);
  }
  
  perl_call_subf ("FTP::event", "si", "chdir_post", 250);
  return 0;
}

int cmd_mkd(char** p, int n)
{ 
   char* dir;
   char* name;
   int   ok;
   int   perm;
   char  wd[258];
   char  buffer[4096] = "\0"; 

   if (n != 1) return send_reply(501, "Syntax: MKD <directory>");
   ok = add_dir(vfs_getcwd (), *p, wd, sizeof(wd)-2);
   if (ok >= 0) { 
      split_dir(wd, &dir, &name, 0);
      convert_filename(name, fconv_mode >> 4);
      ok = permission_ok(dir, PE_MKDIR);
      if (ok >= 0) { 
         /* directory dupecheck failed */
         if (cmd_mkd_dirdcheck(dir, name, buffer, sizeof(buffer))) { 
            if (strlen(buffer) == 0) {
	       strlcpy(buffer, "!BDupe dir check failed.!0", sizeof(buffer));
            }   
	    return send_message(550, 1, buffer, NULL);
         }
	 perm = directory_flag_set_recursive(wd, DP_HIDDEN); 
         *(name-1) = '/';
         if (user_locked && *usr.home) {
            if (strncmp(usr.home, wd, strlen(usr.home)))
               return send_reply(550, "!Bno permission.!0");
         }
         ok = mkdir(wd+1, 0777);
         if (ok < 0) log_action(perm, LG_DIRECTORY, "MKDIR", "%s FAILED: %s",
            wd, strerror(errno)); 
	 else { 
	    log_action(perm, LG_DIRECTORY, "MKDIR", "%s", wd);
            /* take out the hidden flag */
            if (perm & DP_HIDDEN)
               perm &= ~DP_HIDDEN;
            /* get umask value */
            printf("GETMASK %s\n", wd);
            if (!fgets(buffer, sizeof(buffer), stdin)) 
               umask_value = default_umask;
            if (sscanf(buffer + 1, "%ld", &umask_value) != 1) 
               umask_value = default_umask;
            ok = write_permission_file(wd, usr.uid, usr.gid,
               umask_value | (perm << 24));
            if (ok < 0) {
               if (errno == EACCES)
                  ok = 0;
               else 
                  rmdir(wd + 1);
            }
         }
      }
   }
   if (ok < 0) return send_error(550, "mkdir", errno);

   /* masked the directory name with "" for the script */
   memmove(wd+1, wd, sizeof(wd)-1);
   *wd = '\"';
   strlcat(wd, "\"", sizeof(wd));
   /* send_message(257, 0, buffer, NULL); */ /* mhh contains umask, for debug ? */
   return send_message(257, 1, mkdir_msg.msg, wd);
}


/*****************************************************************************
   Returns 0 if at least one normal file was found in the directory. 
   If directory is totally empty or only special/hidden files were
   found 1 is returned.

   Last change: primemover (2000-07-21)
*****************************************************************************/
int directory_empty(char* wd)
{ 
   DIR*           d;
   struct dirent* de;
   char*          hfile;
   char           str[256];
   int            i = 0;

   if (!strcmp(wd, "/")) { 
      errno = EACCES;
      return -1;
   }
   d = opendir(wd+1);
   if (!d) return -1;
   de = readdir(d);
   while (de) { 
      if (strcmp(de->d_name, ".") && strcmp(de->d_name, "..") && 
          strcmp(de->d_name, permission_file))
      { 
         if (!readme_file || strcmp(de->d_name,readme_file)) {
   	    strlcpy(str, hidden_files, sizeof(str));
	    hfile = strtok(str, " \t");
	    while (hfile != NULL) {
	       if (!strcmp(de->d_name, hfile)) i = 1; /* hidden file */
	       hfile = strtok(NULL, " \t");
	    }
            /* at least one normal file was found */
            if(i == 0) {
	       closedir(d);
               return 0;	    
            }
	    i = 0;  /* reset value for next file */
         }
      }
      de = readdir(d);
   }
   closedir(d);
   return 1;
}

int cmd_rmd(char** p, int n)
{ 
   int   ok;
   int   perm;
   char  wd[258];

   if (n != 1) return send_reply(501,"Syntax: RMD <directory>");
   ok = add_dir(vfs_getcwd (),*p,wd,sizeof(wd)-2);
   if (ok >= 0) { 
      if (!strcmp(wd,vfs_getcwd ())) {
         return send_reply(553,"!BCannot remove current directory.!0");
      }
      if (user_locked && *usr.home) {
         if (strncmp(usr.home, wd, strlen(usr.home)))
            return send_reply(550, "!Bno permission.!0");
      }
      ok = permission_ok(wd, PE_RMDIR);
      if (ok >= 0) { 
         perm = ok;
         ok = directory_empty(wd);
         if (ok != -1) { 
	    if (!ok) return send_reply(553,"!BDirectory is not empty.!0");
            perm = directory_flag_set_recursive(wd, DP_HIDDEN);
            ok = remove_permission_file(wd);
            if (ok != -1) ok = rmdir(wd+1);
            if (ok < 0) log_action(perm,LG_DIRECTORY,"RMDIR","%s FAILED: %s",
                wd,strerror(errno));
            else log_action(perm,LG_DIRECTORY,"RMDIR","%s",wd);
         }
      }
   }
   if (ok < 0) return send_error(553, "rmdir", errno);
   memmove(wd+1, wd, sizeof(wd)-1);
   *wd = '\"';
   strlcat(wd, "\"", sizeof(wd));
   return send_message(250, 1, rmdir_msg.msg, wd);
}

int cmd_pwd(char** p, int n)
{
  if (n)
    return send_reply(501,"Syntax: PWD");
  
  return send_message(257, 1 ,"!H\"%s\"!0 is current directory.", vfs_getcwd ());
}

/* EOF */
