/* ---------------------------------------------------------------------------
 *
 *   This file is part of OpenFTPD (based on FTP4ALL)
 *   (C) 1996-2000 by Crescent, (C) 2000 by the OpenFTPD Team
 *
 * ---------------------------------------------------------------------------
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 * ---------------------------------------------------------------------------
 *   MODULE: log.c (ftps) - fuctions for logging
 * ---------------------------------------------------------------------------
 */

#include "ac_config.h"

#include <sys/types.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#include <unistd.h>
#include "externs.h"


void errlog(char* fmt, ...)
{ 
   va_list ap;
   char    buf[32];
   time_t  t;

   if (stderr) { 
      time(&t);
      strftime(buf, sizeof(buf), "%a %d/%m/%Y %H:%M:%S", localtime(&t));
      fprintf(stderr, "%s [%ld] ", buf, (long)getpid());
      va_start(ap, fmt);
      vfprintf(stderr, fmt, ap);
      va_end(ap);
      fputc('\n', stderr);
   }
}


void log_action(int hidden, int lno, char* mnemonic, char* str, ...)
{ 
   va_list ap;
   char    buf[32];

   printf("#4#LOG_ACT %i %i %s ",hidden, lno, mnemonic);
   va_start(ap,str);
   vprintf(str,ap);
   va_end(ap);
   if (str[strlen(str)-1] != '\n') fputc('\n', stdout);
   fgets(buf,sizeof(buf),stdin);
}


void log_begin(int hidden, int lno, char* mnemonic, char* str, ...)
{ 
   va_list ap;
   char    buf[32];

   printf("#4#LOG_BEG %i %i %s ",hidden, lno, mnemonic);
   va_start(ap,str);
   vprintf(str,ap);
   va_end(ap);
   if (str[strlen(str)-1] != '\n') fputc('\n',stdout);
   fgets(buf,sizeof(buf),stdin);
}


void log_end(int hidden, int lno, char* mnemonic, char* str, ...)
{ 
   va_list ap;
   char    buf[32];

   printf("#4#LOG_END %i %i %s ",hidden, lno, mnemonic);
   va_start(ap,str);
   vprintf(str,ap);
   va_end(ap);
   if (str[strlen(str)-1] != '\n') fputc('\n',stdout);
   fgets(buf,sizeof(buf),stdin);
}

/* EOF */
