
/* ---------------------------------------------------------------------------
 *
 *   This file is part of OpenFTPD (based on FTP4ALL)
 *   (C) 1996-2000 by Crescent, (C) 2000 by the OpenFTPD Team
 *
 * ---------------------------------------------------------------------------
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 * ---------------------------------------------------------------------------
 *   MODULE: site.c - site commands
 * ---------------------------------------------------------------------------
 */

#include <sys/types.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/signal.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "externs.h"
#include "common.h"
#include "ftps.h"
#include "list.h"
#include "log.h"
#include "perm.h"
#include "utils.h"
#include "nuke.h"
#include "perl_layer.h"

#include "site_vfs.h"
#include "strlcpy.h"
#include "strlcat.h"

int cmd_site_adduser(char** p, int n)
{
   int  i;
   char str[128];

   if (n < 5 || n > 14) {
      return cat_help("adduser", 501);
   }
   printf("USR_ADQ 0");
   for (i = 0; i<n; i++) printf(" %s", p[i]);
   printf("\n");
   if (!fgets(str, sizeof(str), stdin)) return send_reply(550, "-!Bpipe error!0");
   if (*str == '-') return send_reply(550, str+1);
   if (showinfo) {
      printf("PR_UI %s\n", *p);
      return send_ftpd_reply("!HUser added.!0");
   }
   else return send_reply(211, "!HUser added.!0");
}


int cmd_site_gadduser(char** p, int n)
{
   int  i;
   char str[128];

   if (n < 3 || n > 11) {
      return cat_help("gadduser", 501);
   }
   /* if first arg = 1 then the 3rd and 5th are ignored *
    * because groupops should not change the group or the limit */
   printf("USR_ADQ 1 %s 0 %s 0", p[0], p[1]);
   for (i = 2; i<n; i++) printf(" %s", p[i]);
   printf("\n");
   if (!fgets(str, sizeof(str), stdin)) return send_reply(550, "-!Bpipe error!0");
   if (*str == '-') return send_reply(550, str+1);
   if (showinfo) {
      printf("PR_UI %s\n", *p);
      return send_ftpd_reply("!HUser added.!0");
   }
   else return send_reply(211, "!HUser added.!0");
}

int cmd_site_bind(char** p, int n)
{
   int  port;
   int  port2;
   int  port3;
   char str[128];

   if ((n && (n != 2)) || ((n == 2) && !sscanf(p[1], "%i", &port))) {
      return cat_help("bind",501);
   }
   if (n) {
      printf("PRT_SET %s %i\n", p[0], port);
      if (!fgets(str, sizeof(str), stdin)) strlcpy(str, "-!Bpipe error!0", sizeof(str));
   }
   else {
      printf("PRT_GET\n");
      if (!fgets(str, sizeof(str), stdin)) strlcpy(str, "-!Bpipe error!0", sizeof(str));
      else {
         sscanf(str+1, "%i %i %i", &port, &port2, &port3);
         g_snprintf(str, sizeof(str), "+!HPorts (0=not used): admin=!B%i!H, primary=!B%i!H, secondary=!B%i!0",
                 port,port2,port3);
      }
   }
   return send_reply(*str == '+' ? 200 : 550,str+1);
}


int cmd_site_chmod(char** p, int n)
{
  char           cmd = 0;
  DIR*           d;
  struct dirent* de;
  gint32          dperm;
  char           dir[4] = "";
  int            file = 0;
  gint32          gid;
  long           mode = 0;
  char*          name;
  int            notfile = 0;
  char*          path;
  gint32          perm;
  char*          q = NULL;
  int            se = 0;
  struct stat    st;
  gint32          uid;
  char           wd[256];
  char           buff[256];
  int            who = 0;
  char           wildcard[64];

  if (!flag_get(flags,FLAG_CHMODALLOWED) && !superuser())
    return send_reply(550,"!BYou are not allowed to CHMOD.!0");
  if (n == 1) p[1] = dir;
  if (n<1 || n>2) se = 1; else
  { q = *p;
    while (*q && *q!='+'&& *q!='-' && *q!='=')
    { if (tolower(*q) == 'd') if (who) se = 1; else who = 0x08; else
      if (tolower(*q) == 'u') if (who & 0x01) se = 1; else who |= 0x01; else
      if (tolower(*q) == 'g') if (who & 0x02) se = 1; else who |= 0x02; else
      if (tolower(*q) == 'o') if (who & 0x04) se = 1; else who |= 0x04; else
      if (tolower(*q) == '*') if (who) se = 1; else who |= 0x07; else se = 1;
      q++;
    }
    if (!who) se = 1;
  }
  if (!se)
  { cmd = *q;
    if (cmd!='+' && cmd!='-' && cmd!='=') se = 1;
  }
  if (!se)
  { q++;
    while (*q) switch (tolower(*(q++)))
    { case 'a' :
        notfile = 1;
        if (mode & PE_CHMOD) se = 1; else mode |= PE_CHMOD;
        break;
      case 'g' :
        notfile = 1;
        if (mode & PE_READ) se = 1; else mode |= PE_READ;
        break;
      case 'p' :
        notfile = 1;
        if (mode & PE_WRITE) se = 1; else mode |= PE_WRITE;
        break;
      case 'o' :
        notfile = 1;
        if (mode & PE_OVERWRITE) se = 1; else mode |= PE_OVERWRITE;
        break;
      case 'c' :
        notfile = 1;
        if (mode & PE_CHDIR) se = 1; else mode |= PE_CHDIR;
        break;
      case 'l' :
        notfile = 1;
        if (mode & PE_LIST) se = 1; else mode |= PE_LIST;
        break;
      case 'm' :
        notfile = 1;

        if (mode & PE_MKDIR) se = 1; else mode |= PE_MKDIR;
        break;
      case 'n' :
        notfile = 1;
        if (mode & PE_RMDIR) se = 1; else mode |= PE_RMDIR;
        break;
      case 'h' :
        notfile = 1;
        if (mode || (who != 0x08)) se = 1; else mode = DP_HIDDEN;
        break;
      case '*' :
        notfile = 1;
        if (mode) se = 1; else mode |= (PE_ALL & ~PE_CHMOD);
        break;
      case 'r' :
        file = 1;
        if (mode & FPE_READ) se = 1; else mode |= FPE_READ;
        break;
      case 'w' :
        file = 1;
        if (mode & FPE_WRITE) se = 1; else mode |= FPE_WRITE;
        break;
      case 'x' :
        file = 1;
        if (mode & FPE_CHMOD) se = 1; else mode |= FPE_CHMOD;
        break;
      default  :
        se = 1;
    }
    if (!mode || (file == notfile)) se = 1;
    if (file && (n != 2)) se = 1;
  }
  if (se) return cat_help("chmod",501);

  if (path_in_cdpath (wd))
    return send_reply(550,"!BRead-only path!0");

  if (add_dir(vfs_getcwd (),p[1],wd,sizeof(wd)) == -1)
    return send_error(550,"chmod",errno);
  log_action(directory_flag_set_recursive(wd,DP_HIDDEN), LG_FILE,"CHMOD","%s %s",
    p[0],wd);
  split_dir(wd,&path,&name,0);
  if (*name) strlcpy(wildcard, name, sizeof(wildcard)); else strlcpy(wildcard, ".", sizeof(wildcard));
  if (file)
  { if (permission_ok(path,PE_CHDIR) == -1)
    { errno = EACCES;
      return send_error(550,"chmod",errno);
    }
  }
  de = findfirst(&d,path ? path+1 : path,wildcard);
  while (de)
  { if (file)
    { se = read_file_permission(path ? path : "/",de->d_name,&uid,&gid,&perm);
      if ((se != -1) && !superuser() && ((uid != usr.uid) ||
        !((perm >> 16) & FPE_CHMOD)))
      { errno = EACCES;
        se = -1;
      }
      if (se != -1)
      { strcpy(name,de->d_name);
        *(name-1) = '/';
        if (stat(wd+1,&st) == -1) st.st_mode = S_IFDIR;
        *(name-1) = '\0';
        if (!S_ISDIR(st.st_mode))
        { printf("#1#FCM %i %c %ld %s/%s\n",who,cmd,mode,
            path ? path : "",de->d_name);
          fgets(*p, 8, stdin);
          if (**p == '+') send_message(250,0,"!H%s!0: success",de->d_name);
            else send_message(250,0,"!H%s!B: failed (server)!0",de->d_name);
        }
      }
    } else
    { strcpy(name,de->d_name);
      *(name-1) = '/';
      if (stat(wd+1,&st) == -1) st.st_mode = 0;
      if (S_ISDIR(st.st_mode))
      { se = read_permission_file(wd,&uid,&gid,&perm,NULL,NULL,NULL);
        if ((se == -1) && superuser())
        { uid = usr.uid;
          gid = usr.gid;
          add_dir(vfs_getcwd (), p[1], buff, sizeof(buff));
          printf("GETMASK %s\n", buff);
          if (!fgets(buff, sizeof(buff), stdin))
             perm = default_umask;
          if (sscanf(buff + 1, "%i", &perm) != 1)
             perm = default_umask;
          se = 0;
        }
        if ((se >= 0) && !superuser() && ((usr.uid != uid) ||
          !(perm & (PE_CHMOD << 16))))
        { errno = EACCES;
          se = -1;
        }
        if (se >= 0)
        { switch (cmd)
          { case '+' :
            case '-' :
              dperm = 0;
              if (who & 0x01) dperm |= mode<<16;
              if (who & 0x02) dperm |= mode<<8;
              if (who & 0x04) dperm |= mode;
              if (who & 0x08) dperm |= mode<<24;
              if (cmd == '+') perm |= dperm; else perm &= ~dperm;
              break;
            case '=' :
              if (who & 0x01) perm = (perm & 0xFF00FFFF) | (mode<<16);
              if (who & 0x02) perm = (perm & 0xFFFF00FF) | (mode<<8);
              if (who & 0x04) perm = (perm & 0xFFFFFF00) | mode;
              if (who & 0x08) perm = (perm & 0x00FFFFFF) | (mode<<24);
              break;
          }
          if (!superuser()) if (!(perm & (PE_CHMOD<<16))) perm |= (PE_CHMOD<<16);
          se = write_permission_file(wd,uid,gid,perm);
          if (se >= 0) send_message(250,0,"!H%s!0: success",name);
        }
      } else se = 0;
      *(name-1) = '\0';
    }
    if (se == -1)
      send_message(250,0,"!H%s!B: failed (permission denied)!0",de->d_name);
    de = findnext(d,wildcard);
  }
  endfind(d);
  return send_reply(250,"!HCHMOD command finished.!0");
}

int site_chmod_rec(char* path, int file, int who, char cmd, long mode)
{ DIR*           d;
  struct dirent* de;
  gint32          dperm;
  gint32          gid;
  char*          name;
  gint32          perm;
  int            se;
  struct stat    st;
  char           temp[8];
  gint32          uid;
  char           buff[256];

  de = findfirst(&d,strcmp(path,"/") ? path+1 : ".","*");
  if (strcmp(path,"/")) strcat(path,"/");
  name = path + strlen(path);
  while (de)
  { se = 0;
    if (file)
    { *name = '\0';
      se = read_file_permission(path,de->d_name,&uid,&gid,&perm);
      if ((se != -1) && !superuser() && ((uid != usr.uid) ||
        !((perm >> 16) & FPE_CHMOD)))
      { errno = EACCES;
        se = -1;
      }
      if (se != -1)
      { strcpy(name,de->d_name);
        if (stat(path+1,&st) == -1) st.st_mode = S_IFDIR;
        if (!S_ISDIR(st.st_mode))
        { printf("#1#FCM %i %c %ld %s\n",who,cmd,mode,path);
          fgets(temp, sizeof(temp), stdin);
          if (*temp == '+') send_message(250,0,"!H%s!0: success",path);
            else send_message(250,0,"!H%s!B: failed (server)!0",path);
        }
      }
    } else if (strcmp(de->d_name,".."))
    { strcpy(name,de->d_name);
      if (stat(path+1,&st) == -1) st.st_mode = 0;
      if (S_ISDIR(st.st_mode))
      { se = read_permission_file(path,&uid,&gid,&perm,NULL,NULL,NULL);
        if ((se == -1) && superuser())
        { uid = usr.uid;
          gid = usr.gid;
          printf("GETMASK %s\n", path);
          if (!fgets(buff, sizeof(buff), stdin))
             perm = default_umask;
          if (sscanf(buff + 1, "%i", &perm) != 1)
             perm = default_umask;
          se = 0;
        }
        if ((se >= 0) && !superuser() && ((usr.uid != uid) ||
          !(perm & (PE_CHMOD << 16))))
        { errno = EACCES;
          se = -1;
        }
        if (se >= 0)
        { switch (cmd)
          { case '+' :
            case '-' :
              dperm = 0;
              if (who & 0x01) dperm |= mode<<16;
              if (who & 0x02) dperm |= mode<<8;
              if (who & 0x04) dperm |= mode;
              if (who & 0x08) dperm |= mode<<24;
              if (cmd == '+') perm |= dperm; else perm &= ~dperm;
              break;
            case '=' :
              if (who & 0x01) perm = (perm & 0xFF00FFFF) | (mode<<16);
              if (who & 0x02) perm = (perm & 0xFFFF00FF) | (mode<<8);
              if (who & 0x04) perm = (perm & 0xFFFFFF00) | mode;
              if (who & 0x08) perm = (perm & 0x00FFFFFF) | (mode<<24);
              break;
          }
          if (!superuser()) if (!(perm & (PE_CHMOD<<16))) perm |= (PE_CHMOD<<16);
          se = write_permission_file(path,uid,gid,perm);
          if (se >= 0) send_message(250,0,"!H%s!0: success",path);
        }
      } else se = 0;
    }
    if (se == -1)
      send_message(250,0,"!H%s!B: failed (permission denied)!0",path);
    else if (S_ISDIR(st.st_mode) && strcmp(de->d_name,".") &&
      strcmp(de->d_name,"..")) site_chmod_rec(path,file,who,cmd,mode);
    de = findnext(d,"*");
  }
  endfind(d);
  return 0;
}

int cmd_site_chmodr(char** p, int n)
{
  char           cmd = 0;
  char           dir[4] = "";
  int            file = 0;
  long           mode = 0;
  int            notfile = 0;
  char*          q = NULL;
  int            se = 0;
  char           wd[256];
  int            who = 0;

  if (!flag_get(flags,FLAG_CHMODALLOWED) && !superuser())
    return send_reply(550,"!BYou are not allowed to CHMODR.!0");
  if (n == 1) p[1] = dir;
  if (n<1 || n>2) se = 1; else
  { q = *p;
    while (*q && *q!='+'&& *q!='-' && *q!='=')
    { if (tolower(*q) == 'u') if (who & 0x01) se = 1; else who |= 0x01; else
      if (tolower(*q) == 'g') if (who & 0x02) se = 1; else who |= 0x02; else
      if (tolower(*q) == 'o') if (who & 0x04) se = 1; else who |= 0x04; else
      if (tolower(*q) == 'd') if (who) se = 1; else who = 0x08; else
      if (tolower(*q) == '*') if (who) se = 1; else who |= 0x07; else se = 1;
      q++;
    }
    if (!who) se = 1;
  }
  if (!se)
  { cmd = *q;
    if (cmd!='+' && cmd!='-' && cmd!='=') se = 1;
  }
  if (!se)
  { q++;
    while (*q) switch (tolower(*(q++)))
    { case 'a' :
        notfile = 1;
        if (mode & PE_CHMOD) se = 1; else mode |= PE_CHMOD;
        break;
      case 'g' :
        notfile = 1;
        if (mode & PE_READ) se = 1; else mode |= PE_READ;
        break;
      case 'p' :
        notfile = 1;
        if (mode & PE_WRITE) se = 1; else mode |= PE_WRITE;
        break;
      case 'o' :
        notfile = 1;
        if (mode & PE_OVERWRITE) se = 1; else mode |= PE_OVERWRITE;
        break;
      case 'c' :
        notfile = 1;
        if (mode & PE_CHDIR) se = 1; else mode |= PE_CHDIR;
        break;
      case 'l' :
        notfile = 1;
        if (mode & PE_LIST) se = 1; else mode |= PE_LIST;
        break;
      case 'm' :
        notfile = 1;
        if (mode & PE_MKDIR) se = 1; else mode |= PE_MKDIR;
        break;
      case 'n' :
        notfile = 1;
        if (mode & PE_RMDIR) se = 1; else mode |= PE_RMDIR;
        break;
      case 'h' :
        notfile = 1;
        if (mode || (who != 0x08)) se = 1; else mode = DP_HIDDEN;
        break;
      case '*' :
        notfile = 1;
        if (mode) se = 1; else mode |= (PE_ALL & ~PE_CHMOD);
        break;
      case 'r' :
        file = 1;
        if (mode & FPE_READ) se = 1; else mode |= FPE_READ;
        break;
      case 'w' :
        file = 1;
        if (mode & FPE_WRITE) se = 1; else mode |= FPE_WRITE;
        break;
      case 'x' :
        file = 1;
        if (mode & FPE_CHMOD) se = 1; else mode |= FPE_CHMOD;
        break;
      default  :
        se = 1;
    }
    if (!mode || (file == notfile)) se = 1;
  }
  if (se) return cat_help("chmodr",501);
  if (add_dir(vfs_getcwd (),*(p+1),wd,sizeof(wd)) == -1)
    return send_error(550,"chmodr",errno);

  if (path_in_cdpath (wd))
    return send_reply(550,"!BRead-only path!0");

  log_action(directory_flag_set_recursive(wd,DP_HIDDEN),LG_FILE,"CHMODR","%s %s",
    p[0],wd);
  site_chmod_rec(wd,file,who,cmd,mode);
  return send_reply(250,"!HCHMODR command finished.!0");
}

/*
 * recursive chown dir/file
 * 01102001 DiZZY Fixed it to actually work, see code for changes
 *
 */
int site_chown_rec(char* path, int file, int new_uid, int new_gid)
{
   DIR*           d;
   struct dirent* de;
   gint32          gid;
   char*          name;
   int            se;
   struct stat    st;
   char           temp[256];
   char 	  temppath[MAXFILENAMELENGTH];
   gint32          uid;
   gint32          perm;

   de = findfirst(&d, strcmp(path,"/") ? path+1 : ".", "*");
   if (strcmp(path, "/")) strcat(path, "/");
   name = path + strlen(path);
   while (de) {
      se = 0;
      /* single file */
      if (file) {
         *name = '\0';
         se = read_file_permission(dirname(path), de->d_name, &uid, &gid, &perm);
         if ((se != -1) && !superuser() && ((uid != usr.uid) ||
            !((perm >> 16) & FPE_CHMOD))) {
	    errno = EACCES;
            se = -1;
         }
         if (se != -1) {
	    strcpy(name, de->d_name);
            if (stat(path+1, &st) == -1) st.st_mode = S_IFDIR;
            if (!S_ISDIR(st.st_mode)) {
	       if (new_uid == -1) new_uid = uid;
	       if (new_gid == -1) new_gid = gid;
	       /* call the F4ADP command to chown */
	       printf("FCO- "SI32" "SI32" %s\n", new_uid, new_gid, path);
               fgets(temp, sizeof(temp), stdin);
               if (*temp == '+') send_message(250,0,"!H%s!0: success",path);
               else send_message(250,0,"!H%s!B: failed (server)!0",path);
            }
         }
      }
      else if (strcmp(de->d_name,"..")) {
	 strcpy(name, de->d_name);
	 if (stat(path+1, &st) == -1) st.st_mode = 0;
	 if (S_ISDIR(st.st_mode)) {
	     // do this when its a dir,
  	     // cant just use this code when its a file...
             // since read_permission_file requires a DIR and not a file
            se = read_permission_file(path,&uid,&gid,&perm,NULL,NULL,NULL);
            if (new_uid == -1) new_uid = uid;
	    if (new_gid == -1) new_gid = gid;
            se = write_permission_file(path, new_uid, new_gid, perm);
            if (se >= 0) send_message(250,0,"!H%s!0: success",path);
         } else  {
	    // its a file, we use the same single file code as above
	    // ok i know this could be done nicer, but not now..
	    strlcpy(temppath, path, sizeof(temppath));

         *name = '\0';
         se = read_file_permission(dirname(temppath), de->d_name, &uid, &gid, &perm);
         if ((se != -1) && !superuser() && ((uid != usr.uid) ||
            !((perm >> 16) & FPE_CHMOD))) {
	    errno = EACCES;
            se = -1;
         }
         if (se != -1) {
	    strcpy(name, de->d_name);
            if (stat(path+1, &st) == -1) st.st_mode = S_IFDIR;
            if (!S_ISDIR(st.st_mode)) {
	       if (new_uid == -1) new_uid = uid;
	       if (new_gid == -1) new_gid = gid;
	       /* call the F4ADP command to chown */
	       printf("FCO- "SI32" "SI32" %s\n", new_uid, new_gid, path);
               fgets(temp, sizeof(temp), stdin);
               if (*temp == '+') send_message(250,0,"!H%s!0: success",path);
               else send_message(250,0,"!H%s!B: failed (server)!0",path);
            }
         }
	}
      }
      if (se == -1) {
	 if (S_ISDIR(st.st_mode))
         send_message(250,0,"!H%s!B: failed (permission denied)!0", path);
	 else
         send_message(250,0,"!H%s!B: failed (permission denied)!0", temppath);
      }
      else if (S_ISDIR(st.st_mode) && strcmp(de->d_name,".") &&
        strcmp(de->d_name,"..")) site_chown_rec(path, file, new_uid, new_gid);
      de = findnext(d,"*");
   }
   endfind(d);
   return 0;
}


int cmd_site_chown(char** p, int n)
{
   gint32 gid;
   gint32 new_gid = 0;
   gint32 new_uid = 0;
   gint32 perm;
   char* q;
   int   se = 0;
   gint32 uid;
   char  wd[256];

   if (!superuser()) return send_reply(550,"!BOnly superuser can do CHOWNR.!0");
   if (n < 1 || n > 2) se = 1;
   else {
      q = strchr(*p,':');
      if (q) *q = '\0';
      if (**p) {
         if ((new_uid = get_uid(*p)) == -1) {
	    return send_reply(501,"!BUser not found.!0");
         }
      } else new_uid = -1;
      if (q && *(q + 1)) {
         if ((new_gid = get_gid(q+1)) == -1) {
            return send_reply(501,"!BGroup not found.!0");
         }
      } else new_gid = -1;
      if (new_uid == new_gid == -1) se = 1;
   }
   if (se) return cat_help("chown",501);

   if (path_in_cdpath (n ==2 ? *(p + 1) : ""))
     return send_reply (550, "!BRead-only path!0");

   se = read_permission_file(n == 2 ? *(p+1) : "", &uid, &gid, &perm,
      NULL,NULL,NULL);
   if ((se == -1) && (errno == ENOTDIR)) {
      if (n != 2) return send_error(550, "chown", errno);
      add_dir(vfs_getcwd (), *(p+1), wd, sizeof(wd));
      printf("FCO- "SI32" "SI32" %s\n", new_uid, new_gid, wd);
      fgets(wd, sizeof(wd), stdin);
      return send_reply(*wd == '+' ? 250 : 550,wd+1);
   }
   if (se == -1) {
      uid = gid = 0;
      add_dir(vfs_getcwd (), *(p + 1), wd, sizeof(wd));
      printf("GETMASK %s\n", wd);
      if (!fgets(wd, sizeof(wd), stdin))
         perm = default_umask;
      if (sscanf(wd + 1, "%i", &perm) != 1)
         perm = default_umask;
   }
   if (new_uid != -1) uid = new_uid;
   if (new_gid != -1) gid = new_gid;
   if (n == 2) se = write_permission_file(*(p+1), uid, gid, perm);
   else se = write_permission_file("",uid,gid,perm);
   if (se == -1) return send_error(550,"chown",errno);
   else return send_reply(250,"!HCHOWN command successful.!0");
}


int cmd_site_chownr(char** p, int n)
{
   gint32 gid;
   gint32 new_gid = 0;
   gint32 new_uid = 0;
   gint32 perm;
   char* q;
   int   se = 0;
   gint32 uid;
   char  wd[256];

   if (!superuser()) return send_reply(550,"!BOnly superuser can do CHOWNR.!0");
   if (n < 1 || n > 2) se = 1;
   else {
      q = strchr(*p,':');
      if (q) *q = '\0';
      if (**p) {
         if ((new_uid = get_uid(*p)) == -1) {
	    return send_reply(501,"!BUser not found.!0");
         }
      } else new_uid = -1;
      if (q && *(q + 1)) {
         if ((new_gid = get_gid(q+1)) == -1) {
            return send_reply(501,"!BGroup not found.!0");
         }
      } else new_gid = -1;
      if (new_uid == new_gid == -1) se = 1;
   }
   if (se) return cat_help("chownr",501);

   if (path_in_cdpath (n == 2 ? *(p + 1) : ""))
     return send_reply (550, "!BRead-only path!0");

   // dont know why this was not active,
   // makes sense to me to also chown the argument
   // ??
   se = read_permission_file(n == 2 ? *(p+1) : "", &uid, &gid, &perm,
      NULL,NULL,NULL);
   if ((se == -1) && (errno == ENOTDIR)) {
      if (n != 2) return send_error(550, "chown", errno);
      add_dir(vfs_getcwd (), *(p+1), wd, sizeof(wd));
      printf("FCO- "SI32" "SI32" %s\n", new_uid, new_gid, wd);
      fgets(wd, sizeof(wd), stdin);
      return send_reply(*wd == '+' ? 250 : 550,wd+1);
   }

   if (se == -1) {
      /* prime whatchout with this as umask has changed */
      uid = gid = 0;
      perm = umask_value;
   }
   if (new_uid != -1) uid = new_uid;
   if (new_gid != -1) gid = new_gid;
   if (n == 2) se = write_permission_file(*(p+1), uid, gid, perm);
   else se = write_permission_file("",uid,gid,perm);
   if (add_dir(vfs_getcwd (), *(p+1), wd, sizeof(wd)) == -1) {
      return send_error(550,"chmod",errno);
   }
   site_chown_rec(wd, 0, new_uid, new_gid);
   if (se == -1) return send_error(550,"chown",errno);
   else return send_reply(250,"!HCHOWN command successful.!0");
}

int cmd_site_close(char** p, int n)
{ char str[128];

  if (n > 1) return cat_help("close",501);
  if (n)
  { if (!strcasecmp(p[0],"normal")) g_snprintf(str, sizeof(str), "cn"); else
      if (!strcasecmp(p[0],"transactional")) g_snprintf(str, sizeof(str), "ct"); else
      if (!strcasecmp(p[0],"immediate")) g_snprintf(str, sizeof(str), "ci"); else
      return cat_help("close",501);
  } else g_snprintf(str, sizeof(str), "ci");
  printf("SRV_SST %s\n",str);
  if (!fgets(str,sizeof(str),stdin)) strlcpy(str, "-!Bpipe error!0", sizeof(str));
  return send_reply(*str == '+' ? 200 : 550,str+1);
}


int cmd_site_color(char** p, int n)
{ char str[128];

  if (n) return cat_help("color",501);
  /* change it in the current session */
  usr.class ^= UF_COLOR;
  /* change it in the users data */
  printf("COL\n");
  if (!fgets(str,sizeof(str),stdin)) strlcpy(str, "-!Bpipe error!0", sizeof(str));
  return send_reply(*str == '+' ? 200 : 500,str+1);
}

int cmd_site_count(char** p, int n)
{ char str[128];

  printf("CNT\n");
  if (!fgets(str,sizeof(str),stdin)) strlcpy(str, "-!Bpipe error!0", sizeof(str));
  return send_reply(*str == '+' ? 200 : 550,str+1);
}

int cmd_site_version(char** p, int n)
{
  char str[128];

  printf("VER\n");
  if (!fgets(str,sizeof(str),stdin)) strlcpy(str, "-!Bpipe error!0", sizeof(str));
  return send_reply(*str == '+' ? 200 : 550,str+1);
}


int cmd_site_uptime(char** p, int n)
{
  char str[128];

  printf("UPTIME\n");
  if (!fgets(str,sizeof(str),stdin)) strlcpy(str, "-!Bpipe error!0", sizeof(str));
   return send_reply(*str == '+' ? 200 : 550, str+1);
}


off_t del_files_r(char* path)
{
   struct dirent* d;
   DIR*           dir;
   char*          name;
   off_t          size = 0;
   struct stat    st;
   int            ok;
   char           str[128];
   int            hidden;

   dir = opendir(path);
   if (!dir) return 0;
   strcat(path, "/");
   name = path + strlen(path);
   while ((d = readdir(dir)) != NULL) {
      strcpy(name,d->d_name);
      if (strcmp(d->d_name, ".") && strcmp(d->d_name, "..") &&
         !stat(path, &st)) {

	 if (S_ISDIR(st.st_mode)) {
	    size += del_files_r(path);
            strcpy(name, d->d_name);
            rmdir(path);
         }
	 else {
            hidden = directory_flag_set_recursive(path, DP_HIDDEN);
            ok = unlink(path);
            if (ok < 0) {
	       log_action(hidden, LG_FILE, "DELETE", "%s FAILED: %s",
        	  path, strerror(errno));
	    }
	    else {
	       log_action(hidden, LG_FILE, "DELETE","%s %s (total %lu bytes)",
                	  usr.name, path, size);

	       /***********************************
		* mark file as deleted in dupelog *
		***********************************/
	       printf("DUP_DEL \"%s\"\n", path);
               if (!fgets(str, sizeof(str), stdin)) {
		  errlog("pipe error on DUP_DEL in cmd_dele()");
	       }
            }
            if (*d->d_name != '.') size += st.st_size;
         }
      }
   }
   closedir(dir);
   return size;
}

int cmd_site_del(char** p, int n)
{
   int            ok;
   off_t          size = 0;
   struct stat    st;
   char           wd[256];
   char           str[256];
   char           str2[256];
   int            x;
   int            hidden;

   if (!superuser()) return send_reply(550,"!BOnly superuser can do SITE DEL.!0");
   if (n < 1) return cat_help("del",501);

   memset(str, 0, sizeof(str));
   for (x = 0; x < n; x++) {
     g_snprintf(str2, sizeof(str), "%s ", p[x]);
     strlcat(str, str2, sizeof(str));
   }
   g_strchomp(str);

   ok = add_dir(vfs_getcwd (), str, wd, sizeof(wd));
   hidden = directory_flag_set_recursive(wd, DP_HIDDEN);
   if (ok != -1) ok = stat(wd+1, &st);
   if (ok != -1) {
      /* directory */
      if (S_ISDIR(st.st_mode)) {
         ok = remove_permission_file(wd);
	 if (ok != -1) size = del_files_r(wd+1);
	 if (ok != -1) ok = add_dir(vfs_getcwd (), str, wd, sizeof(wd));
	 if (ok != -1) rmdir(wd+1);
	 log_action(hidden, LG_MISC, "SITE DEL","%s %s (total %lu bytes)",
                    usr.name, wd, size);
      }
      /* regular file */
      else if(S_ISREG(st.st_mode)) {
	size = st.st_size;
	ok = unlink(wd+1);
	if (ok < 0) {
	  log_action(hidden, LG_FILE,"DELETE","%s FAILED: %s",
		     wd,strerror(errno));
	}
	else {
	  log_action(hidden, LG_FILE, "DELETE","%s %s (total %lu bytes)",
		     usr.name, wd, size);

            /***********************************
             * mark file as deleted in dupelog *
             ***********************************/
            printf("DUP_DEL \"%s\"\n", wd);
            if (!fgets(str, sizeof(str), stdin)) {
               errlog("pipe error on DUP_DEL in cmd_dele()");
            }
         }
      }
   }
   if (ok == -1) log_action(hidden, LG_MISC, "SITE DEL", "%s %s FAILED: %s",
      usr.name, wd, strerror(errno));                      /* new line */
   if (ok == -1) return send_error(550,"delete",errno);
   if (S_ISDIR(st.st_mode)) {
     /* directory */
      g_snprintf(wd, sizeof(wd), "Directory !H%s!0 deleted. (!H%ld!0 bytes total), (not unduped)", str, (long)size);
   }
   else if(S_ISREG(st.st_mode))
     /* filename */
     g_snprintf(wd, sizeof(wd), "Filename deleted. (!H%ld!0 bytes total), (not unduped)", (long)size);

   return send_reply(200, wd);
}


int cmd_site_disable(char** p, int n)
{
   char str[128];

   if (n != 1) return cat_help("disable",501);
   printf("ADI %s\n",*p);
   if (!fgets(str,sizeof(str),stdin)) return send_reply(550,"-!Bpipe error!0");
   if (*str=='-') return send_reply(550,str+1);
   if (showinfo) {
      printf("PR_UI %s\n", *p);
      return send_ftpd_reply("!HAccount disabled.!0");
   }
   else return send_reply(211, "!HAccount disabled.!0");
}


int cmd_site_dump(char** p, int n)
{
  int   i;
  int   l;
  char* ptr;
  char  str[256];

  if (n != 1) return cat_help("dump",501);
  if (!superuser()) return send_reply(550,"!BPermission denied!0");
  if (!strcasecmp(*p,"cdpath"))
    send_reply (200, "sorry, cdpath dumping currently not supported");
  else if (!strcasecmp(*p,"check"))
  { g_snprintf(str, sizeof(str), "check=!H0x%p!0, count=!H%i!0, size=!H%i!0",
      check,check_count,check_size);
    send_message(200,check_count ? 0 : 1,str,NULL);
    for (i=0; i<check_count; i++)
    { g_snprintf(str, sizeof(str), "!C%-8s !H%-48s !C%-8s!0",
        check[i].wildcard,check[i].name,check[i].param);
      send_message(200,i < check_count-1 ? 0 : 1,str,NULL);
    }
  } else if (!strcasecmp(*p,"config"))
  { g_snprintf(str, sizeof(str), "server_dir=!H%s!0", server_dir);
    send_message(200,0,str,NULL);
    g_snprintf(str, sizeof(str), "permission_file=!H%s!0", permission_file);
    send_message(200,0,str,NULL);
    g_snprintf(str, sizeof(str), "readme_file=!H%s!0", readme_file);
    send_message(200,0,str,NULL);
    /* added by primemover */
    g_snprintf(str, sizeof(str), "hidden_files=!H%s!0", hidden_files);
    send_message(200,0,str,NULL);
    g_snprintf(str, sizeof(str), "dupechk=!H%s!0", dupechk);
    send_message(200,0,str,NULL);
    g_snprintf(str, sizeof(str), "dupe_path=!H%s!0", dupe_path);
    send_message(200,1,str,NULL);
  } else if (!strcasecmp(*p,"dcheck"))
  { g_snprintf(str, sizeof(str), "dcheck=!H0x%p!0, count=!H%i!D, size=!H%i!0",
      dcheck,dcheck_count,dcheck_size);
    send_message(200,dcheck_count ? 0 : 1,str,NULL);
    for (i=0; i<dcheck_count; i++)
    { g_snprintf(str, sizeof(str), "!C%-8s !H%-48s !C%-8s!0",
        dcheck[i].wildcard,dcheck[i].name,dcheck[i].param);
      send_message(200,i < dcheck_count-1 ? 0 : 1,str,NULL);
    }
  } else if (!strcasecmp(*p,"dirdcheck")) {
     g_snprintf(str, sizeof(str), "dirdcheck=!H0x%p!0, count=!H%i!0, size=!H%i!0",
     dirdcheck,dirdcheck_count,dirdcheck_size);
     send_message(200,dirdcheck_count ? 0 : 1,str,NULL);
     for (i=0; i<dirdcheck_count; i++)  {
     	g_snprintf(str, sizeof(str), "!C%-8s !H%-48s !C%-8s!0", dirdcheck[i].wildcard,dirdcheck[i].name,dirdcheck[i].param);
     	send_message(200,i < dirdcheck_count-1 ? 0 : 1,str,NULL);
     }
  } else if (!strcasecmp(*p,"delcheck")) {
     g_snprintf(str, sizeof(str), "delcheck=!H0x%p!0, count=!H%i!0, size=!H%i!0",
     delcheck,delcheck_count,delcheck_size);
     send_message(200,delcheck_count ? 0 : 1,str,NULL);
     for (i=0; i<delcheck_count; i++) {
       g_snprintf(str, sizeof(str), "!C%-8s !H%-48s !C%-8s!0", delcheck[i].wildcard,delcheck[i].name,delcheck[i].param);
       send_message(200,i < delcheck_count-1 ? 0 : 1,str,NULL);
      }
  } else if (!strcasecmp(*p,"dshortcut"))
  { g_snprintf(str, sizeof(str), "dshortcut=!H0x%p!0, len=!H%i!0, size=!H%i!0",
      dshortcut,dshortcut_len,dshortcut_size);
    send_message(200,dshortcut_len ? 0 : 1,str,NULL);
    for (i=0; i<dshortcut_len;)
    { ptr = dshortcut + i;
      l = strlen(ptr) + 1;
      g_snprintf(str, sizeof(str), "!C%-20s !H%-s!0",ptr,ptr+l);
      i += l + strlen(ptr+l) + 1;
      send_message(200,i < dshortcut_len ? 0 : 1,str,NULL);
    }
  } else return cat_help("dump",501);
  return 0;
}


int cmd_site_enable(char** p, int n)
{
   char str[128];

   if (n != 1) return cat_help("enable",501);
   printf("AEN %s\n",*p);
   if (!fgets(str,sizeof(str),stdin)) return send_reply(550,"-!Bpipe error!0");
   if (*str=='-') return send_reply(550,str+1);
   if (showinfo) {
      printf("PR_UI %s\n", *p);
      return send_ftpd_reply("!HAccount enabled.!0");
   }
   else return send_reply(211, "!HAccount enabled.!0");
}


int cmd_site_exec(char** p, int n)
{ char* c;
  int   err;
  FILE* file;
  pid_t pid;
  int   pipefd[2];
  char  str[128];

#ifdef DISABLE_EXEC
   return send_reply(550, "!BOpenFTPD was compiled without this command.!0");
#endif

  if (!n) return cat_help("exec",501);
  if (!superuser()) return send_reply(550,"!BYou are not superuser.!0");
  err = socketpair(AF_UNIX,SOCK_STREAM,0,pipefd);
  if (err == -1) return send_error(550,"pipe",0);
  pid = fork();
  if (pid == -1) return send_error(550,"fork",errno);
  if (pid)
  { close(pipefd[1]);
    file = fdopen(pipefd[0],"rt");
    if (!file) return send_reply(550,"fdopen");
    do
    { c = fgets(str,sizeof(str)-1,file);
      if (c && strlen(str))
        if (send_message(200,0,str,NULL) == -1) return -1;
    } while (c);
    fclose(file);
    close(pipefd[0]);
    pid = wait(&err);
    g_snprintf(str, sizeof(str), "EXEC finished with exitcode !H%i!0.",err);
    return send_reply(200,str);
  } else
  { fclose(sockf);
    if (strcmp(vfs_getcwd (),"/")) chdir(vfs_getcwd ()+1);
    close(pipefd[0]);
    err = dup2(pipefd[1],1);
    if (err != -1) err = dup2(pipefd[1],2);
    if (err != -1) err = dup2(pipefd[1],0);
    close(pipefd[1]);
    *(p+n) = NULL;
    if (err != -1) err = execvp(*p,p);
    exit(err);
  }
}


int cmd_site_group(char** p, int n)
{
	int    i;
	int    se = 0;
	char   str[256];

	char   change_str[256];
	char   change_groupname[9];
	int    change_printinfo=0;
	int    change_printrights=0;
	int    change_loopi=0;
	char   *notnull;

	if (!n)
		se = 1;
	else if (!strcasecmp(*p, "ADD"))
	{
		if ((n < 2) || (n > 3))
			se = 1;
		else
		{
			printf("GRP_ADD %s %s\n", p[1], n == 3 ? p[2] : "-0");
			if (!fgets(str, sizeof(str), stdin))
				return send_reply(550,"-!Bpipe error!0");
			if (*str == '-')
				return send_reply(550, str+1);
			if (showinfo)
			{
				printf("PR_GI %s\n", *(p+1));
				return send_ftpd_reply(str+1);
			}
			return send_reply(200, str+1);
		}
	}
	else if (!strcasecmp(*p, "CHANGE"))
	{
		if (n < 3)
			return cat_help("group.change",501);
		else
		{
			change_groupname[8] = '\0';
			strncpy(change_groupname,p[1],8);
			printf("GRP_CHG %s", p[1]);
			/* OLD CODE */
			/*
			for (i = 2; i < n; i++)
				printf(" '%s'", p[i]);
			*/
			/* NEW CODE (added by wector) -- fixes following:
			 * NAME=<new name> (will show group correctly)
			 * WHATEVER=.... RIGHTS+|-= (will show both group and rights info if both are changed)
			 * This is a bit ugly as usual .. to not feel bad I blame original design to not
			 * allow this stuff in a clean way.
			 */
			for (i = 2; i < n; i++)
			{
				printf(" '%s'", p[i]);
				if (!strncasecmp(p[i],"RIGHTS",6))
					change_printrights=1;
				else if (!strncasecmp(p[i],"NAME",4))
					{
					strncpy(change_groupname,(p[i]+5),8);
					change_printinfo=1;
					}
				else
					change_printinfo=1;
			}
			/* END NEW CODE */
			printf("\n");
			if (!fgets(str, sizeof(str), stdin))
				return send_reply(550,"-!Bpipe error!0");
			if(*str == '-')
				return send_reply(550,str+1);
			else if (showinfo)
			{
				/* NEW CODE
				 * We really should call this function recursivly to achieve this
				 * but because of how things work I am unsure what if that is very
				 * wise at this point. fixme for the future maybe .. */
				if (change_printinfo)
					printf("PR_GI %s\n", change_groupname);
				if (change_printrights)
					printf("PR_RGT g %s\n", change_groupname);
				for (notnull = fgets(change_str,sizeof(change_str), stdin) ; (notnull) ;
					notnull = fgets(change_str, sizeof(change_str), stdin))
				{
					errlog("got: %s", change_str);
					if (!strcasecmp(change_str,"+\n"))
					{
						change_loopi++;
						if (change_loopi == change_printinfo + change_printrights)
							break;
					}
					else
						if (send_message(211,0,change_str+1,NULL) == -1)
							return -1;
				}
				if (notnull)
					return send_reply(211, str+1);
				else
					return send_reply(550, "!Bpipe error!0");

				/* OLD CODE
				if (strncasecmp(*(p+2), "RIGHTS", 5))
				{
					printf("PR_GI %s\n", (p+1));
					return send_ftpd_reply(str+1);
				}
				else {
					printf("PR_RGT g %s\n", p[1]);
					if (!fgets(str,sizeof(str), stdin))
						strlcpy(str, "-!Bpipe error!0", sizeof(str));
					while (*str == '+' && strcmp(str, "+\n")) {
						if (send_message(211, 0, str+1, NULL) == -1)
							return -1;
						if (!fgets(str,sizeof(str),stdin))
							strlcpy(str, "-", sizeof(str));
					}
					if (*str == '+')
						return send_reply(211, "!H<end of list>!0");
					if (strcmp(str,"-"))
						return send_reply(550, str+1);
					else
						return send_reply(211, "!Bpipe error, aborting list!0");
				}
				*/
			}
			else
				return send_reply(200,str+1);
		}
	}
	else if (!strcasecmp(*p, "LIST")) {
		printf("PR_GL\n");
		return send_ftpd_reply(NULL);
	}
	else if (!strcasecmp(*p, "SHOW")) {
		if (n != 2) se = 1;
		else {
			printf("PR_GI %s\n", *(p+1));
			return send_ftpd_reply(NULL);
		}
	}
	else if (!strcasecmp(*p, "REMOVE")) {
		if (n != 2) se = 1;
		else printf("GRP_RM %s\n", p[1]);
	}
	else if (!strcasecmp(*p,"RIGHTS")) {
		if (n != 2) se = 1;
		else
		{
			printf("PR_RGT g %s\n", p[1]);
			if (!fgets(str,sizeof(str), stdin))
				strlcpy(str, "-!Bpipe error!0", sizeof(str));
			while (*str == '+' && strcmp(str, "+\n"))
			{
				if (send_message(211, 0, str+1, NULL) == -1)
					return -1;
				if (!fgets(str,sizeof(str),stdin))
					strlcpy(str, "-", sizeof(str));
			}
			if (*str == '+')
				return send_reply(211, "!H<end of list>!0");
			if (strcmp(str,"-"))
				return send_reply(550, str+1);
			else
				return send_reply(211, "!Bpipe error, aborting list!0");
		}
	}
	else se = 1;
	if (se)
		return cat_help("group",501);
	if (!fgets(str,sizeof(str),stdin))
		strlcpy(str, "-!Bpipe error!0", sizeof(str));
	return send_reply(*str == '+' ? 200 : 550,str+1);
}


int cmd_site_groups(char** p, int n)
{
   int   i;
   char* name;
   char  str[128];

   if (!n) {
      strlcpy(str, "!HPrimary group:!0 ", sizeof(str));
      name = get_groupname(usr.gid);
      if (name) strlcat(str, name, sizeof(str));
      else g_snprintf(str + strlen(str), sizeof(str) - strlen(str), SI32, usr.gid);
      if (send_message(200, 0, str, NULL) == -1) return -1;
      strlcpy(str, "!HSecondary groups:!0", sizeof(str));
      for (i = 0; i < 8; i++) {
	 if (usr.gids[i] != -1) {
	    strlcat(str, " ", sizeof(str));
	    name = get_groupname(usr.gids[i]);
	    if (name)
	      strlcat(str, name, sizeof(str));
	    else
	      g_snprintf(str + strlen(str), sizeof(str) - strlen(str), SI32, usr.gids[i]);
	 }
      }
      if (strlen(str) == 17 * sizeof(char))
	strlcat(str, " (none)", sizeof(str));

      return send_reply(200,str);
   }
   if (n == 1) {
      printf("GRP %s\n",*p);
      fgets(str, sizeof(str), stdin);

      if (*str == '-')
	return send_reply(550,str+1);
      if (send_message(200,0,str+1,NULL) == -1)
	return -1;

      fgets(str, sizeof(str), stdin);

      return send_reply(200,str + 1);
   }
   if (n == 2) {
      printf("GRP %s %s\n", *p, *(p+1));
      if (!fgets(str,sizeof(str),stdin))
	return send_reply(550,"!Bpipe error!0");
      if (*str=='-')
	return send_reply(550,str+1);
      if (showinfo) {
	 printf("PR_UI %s\n", *p);
	 return send_ftpd_reply("!HUser added to group.!0");
      }
      else
	return send_reply(211, "!HUser added to group.!0");
   }
   return cat_help("groups", 501);
}


int cmd_site_idle(char **p, int n)
{
   return send_reply(550,"!BCurrently not implemented.!0");
}


int cmd_site_info(char** p, int n)
{
   char str[128];
   char str2[128];
   int  i;

   if (!n)
     printf("UIN %s\n", usr.name);
   else {
     memset(str, 0, sizeof(str));
     for(i = 0; i < n; i++) {
       g_snprintf(str2, sizeof(str2), "%s ", p[i]);
       strlcat(str, str2, sizeof(str));
     }
     g_strchomp(str);

     printf("UIN %s %s\n", usr.name, str);
   }
   if (!fgets(str, sizeof(str), stdin))
     strlcpy(str, "-!Bpipe error!0", sizeof(str));

   return send_reply(*str == '+' ? 200 : 550, str+1);
}


int cmd_site_ipban(char** p, int n)
{
   char           str[256];
   char           str2[256];

   if (!superuser()) {
      return send_reply(550,"!BOnly superusers may use this command.!0");
   }

   if (n > 0 && !strcasecmp(p[0], "list")) {
      if (send_message(211,0," !C.------------------------------------------------------------------------.!0",NULL) == -1) return -1;
      if (send_message(211,0," !C|!3 !BOpenFTPD!H BANNED HOSTS                                                  !0!C|!0",NULL) == -1) return -1;
      if (send_message(211,0," !C|------------------------------------------------------------------------|!0",NULL) == -1) return -1;
      if (send_message(211,0," !C|!h BANNED HOPS!C                                                            |!0",NULL) == -1) return -1;
      if (send_message(211,0," !C|!c -----------!C                                                            |!0", NULL) == -1) return -1;

      g_snprintf(str, sizeof(str), " !C| !H%-70s !C|!0", rejected_hops);
      if (send_message(211, 0, str, NULL) == -1) return -1;
      g_snprintf(str, sizeof(str), " !C|%72s|!0", " ");
      if (send_message(211, 0, str, NULL) == -1) return -1;
      if (send_message(211, 0," !C|!h BANNED HOSTS!C                                                           |!0",NULL) == -1) return -1;
      if (send_message(211, 0," !C|!c ------------!C                                                           |!0", NULL) == -1) return -1;

      printf("IP_LS deny\n");
      if (!fgets(str,sizeof(str),stdin))
	strlcpy(str, "-!Bpipe error!0", sizeof(str));
      while (*str == '+' && strcmp(str, "+\n")) {
         if (str[strlen(str) - 1] == '\n') str[strlen(str) - 1] = '\0';
         g_snprintf(str2, sizeof(str2), " !C|!0 !H%-70s !C|!0", str+1);
         if (send_message(211, 0, str2, NULL) == -1) return -1;
	 if (!fgets(str,sizeof(str),stdin))
	   strlcpy(str, "-", sizeof(str));
      }
      if (*str == '-') {
	 if (strcmp(str, "-")) return send_reply(211, str+1);
	 else return send_reply(211,"!Bpipe error, aborting list!0");
      }

      g_snprintf(str, sizeof(str), " !C|%72s|!0", " ");
      if (send_message(211, 0, str, NULL) == -1) return -1;
      if (send_message(211, 0, " !C|!h ALLOWED HOSTS !C                                                         |!0",NULL) == -1) {
         return -1;
      }
      if (send_message(211,0," !C|!c -------------!C                                                          |!0", NULL) == -1) {
         return -1;
      }

      printf("IP_LS allow\n");
      if (!fgets(str,sizeof(str),stdin))
	strlcpy(str, "-!Bpipe error!0", sizeof(str));
      while (*str == '+' && strcmp(str, "+\n")) {
         if (str[strlen(str) - 1] == '\n') str[strlen(str) - 1] = '\0';
	 g_snprintf(str2, sizeof(str2), " !C|!0 !H%-70s !C|!0", str+1);
	 if (send_message(211, 0, str2, NULL) == -1)
	   return -1;
	 if (!fgets(str, sizeof(str),stdin))
	   strlcpy(str, "-", sizeof(str));
      }
      if (*str == '-') {
	 if (strcmp(str, "-")) return send_reply(211, str+1);
	 else return send_reply(211, "!Bpipe error, aborting list!0");
      }

      g_snprintf(str, sizeof(str), " !C|%72s|!0", " ");
      if (send_message(211, 0, str, NULL) == -1) return -1;
      if (send_message(211, 1," !C`------------------------------------------------------------------------+!0",NULL) == -1) {
         return -1;
      }
      return 0;
   }
   else if (n == 2 && !strcasecmp(*p, "add")) {
      printf("IP_ADD %s\n",p[1]);
      if (!fgets(str, sizeof(str), stdin)) {
	 return send_reply(550,"!Bpipe error!0");
      }
      if (*str=='-') return send_reply(550, str+1);
      return send_reply(200, str+1);
   }
   else if (n == 2 && !strcasecmp(*p, "remove")) {
      printf("IP_RM %s\n",p[1]);
      if (!fgets(str, sizeof(str), stdin)) {
	 return send_reply(550,"!Bpipe error!0");
      }
      if (*str=='-') return send_reply(550, str+1);
      return send_reply(200, str+1);
   }
   else if (n == 2 && !strcasecmp(*p, "reset")) {
      printf("IP_RST %s\n", p[1]);
      if (!fgets(str, sizeof(str), stdin)) {
	 return send_reply(550,"!Bpipe error!0");
      }
      if (*str=='-') return send_reply(550, str+1);
      return send_reply(200, str+1);
   }
   return cat_help("ipban", 501);
}


int cmd_site_setinfo(char** p, int n)
{
   char str[128];
   char str2[128];
   int  i;

   if (n < 1)
     printf("UIN\n");
   else if (n == 1)
     printf("UIN %s\n", p[0]);
   else {
     memset(str, 0, sizeof(str));
     for(i = 1; i < n; i++) {
       g_snprintf(str2, sizeof(str2), "%s ", p[i]);
       strlcat(str, str2, sizeof(str));
     }
     /* don't allow ':' because it's used as limiter in the stat files */
     for (i = 0; i < strlen(str); i++) {
       if(str[i] == ':')
	 str[i] = ' ';
     }
     printf("UIN %s %s\n", p[0], str);
   }
   if (!fgets(str, sizeof(str), stdin))
     strlcpy(str, "-!Bpipe error!0", sizeof(str));

   return send_reply(*str == '+' ? 200 : 550,str+1);
}

int cmd_site_ginfo(char** p, int n)
{
   char str[128];
   char str2[128];
   int  i;

   if (!n) printf("GIN %s\n", get_groupname(usr.gid));
   else {
     memset(str, 0, sizeof(str));
     for(i = 0; i < n; i++) {
       g_snprintf(str2, sizeof(str2), "%s ", p[i]);
       strlcat(str, str2, sizeof(str));
     }
     /* don't allow ':' because it's used as limiter in the stat files */
     for (i = 0; i < strlen(str); i++) {
       if (str[i] == ':') str[i] = ' ';
     }
     printf("GIN %s %s\n", get_groupname(usr.gid), str);
   }
   if (!fgets(str, sizeof(str), stdin)) strlcpy(str, "-!Bpipe error!0", sizeof(str));
   return send_reply(*str == '+' ? 200 : 550,str+1);
}

int cmd_site_setginfo(char** p, int n)
{
   char str[128];
   char str2[128];
   int  i;

   if (n < 1) printf("GIN\n");
   else if (n == 1) printf("GIN %s\n", p[0]);
   else {
     memset(str, 0, sizeof(str));
      for(i = 1; i < n; i++) {
	g_snprintf(str, sizeof(str), "%s ", p[i]);
	strlcat(str, str2, sizeof(str));
      }
      /* don't allow ':' because it's used as limiter in the stat files */
      for (i = 0; i < strlen(str); i++) {
	 if (str[i] == ':') str[i] = ' ';
      }
      printf("GIN %s %s\n", p[0], str);
   }
   if (!fgets(str, sizeof(str), stdin)) strlcpy(str, "-!Bpipe error!0", sizeof(str));
   return send_reply(*str == '+' ? 200 : 550,str+1);
}


int cmd_site_kick(char** p, int n)
{ int  pid;
  char str[128];

  if ((n < 1) || (n > 2)) return cat_help("kick",501);
  if (sscanf(p[0],"%i",&pid)) printf("USR_KIL %i %s\n",pid,n == 2 ? p[1] : "");
    else printf("USR_KIC %s %s\n",p[0],n == 2 ? p[1] : "");
  if (!fgets(str,sizeof(str),stdin)) strlcpy(str, "-!Bpipe error!0", sizeof(str));
  return send_reply(*str == '+' ? 200 : 550,str+1);
}

void cat_permissions(char* s, long p)
{
  if ((p & PE_ALL) == PE_ALL)
  { strcat(s," (!Hfull rights!0)");
    return;
  }
  if (!p)
  { strcat(s," (!Hno rights!0)");
    return;
  }
  if (p & PE_CHDIR) strcat(s," !Hcd!0");
  if (p & PE_LIST) strcat(s," !Hls!0");
  if (p & PE_MKDIR) strcat(s," !Hmkdir!0");
  if (p & PE_CHMOD) strcat(s," !Hchmod!0");
  if (p & PE_READ) strcat(s," !Hget!0");
  if (p & PE_WRITE) strcat(s," !Hput!0");
  if (p & PE_OVERWRITE) strcat(s," !Hoverwrite!0");
  if (p & PE_RMDIR) strcat(s, " !Hrmdir!0");
}

void cat_file_permissions(char* s, long p)
{ if (!p)
  { strcat(s," (!Hno rights!0)");
    return;
  }
  if (p & FPE_READ) strcat(s," !Hread!0");
  if (p & FPE_WRITE) strcat(s," !Hwrite!0");
  if (p & FPE_CHMOD) strcat(s," !Hchmod!0");
}

int cmd_site_mod(char** p, int n)
{ char*          dir;
  gint32          gid;
  char*          name;
  gint32          perm;
  int            se;
  char           str[128];
  gint32          uid;

  if (n > 1) return cat_help("mod",501);
  if (n == 1)
  { se = permission_ok(*p,PE_CHDIR);
    if ((se == -1) && (errno == ENOTDIR))
    { split_dir(*p,&dir,&name,0);
      se = permission_ok(dir,PE_CHDIR);
      if (se != -1)
      { se = read_file_permission(dir,name,&uid,&gid,&perm);
        if (se != -1)
        { strlcpy(str, "owner [", sizeof(str));
          name = get_username(uid);
          if (name) g_snprintf(str+strlen(str), sizeof(str) - strlen(str), "!H%-9s!0",name); else
            g_snprintf(str+strlen(str), sizeof(str) - strlen(str), "!H%-9"I32"!0",uid);
          strlcat(str, "] :", sizeof(str));
          cat_file_permissions(str,(perm >> 16) & 0xFF);
          if (send_message(250,0,str,NULL) == -1) return -1;
          strlcpy(str, "group [", sizeof(str));
          name = get_groupname(gid);
          if (name) g_snprintf(str+strlen(str), sizeof(str) - strlen(str), "!H%-9s!0",name); else
            g_snprintf(str+strlen(str), sizeof(str) - strlen(str), "!H%-9"I32"!0",gid);
          strlcat(str, "] :", sizeof(str));
          cat_file_permissions(str,(perm >> 8) & 0xFF);
          if (send_message(250,0,str,NULL) == -1) return -1;
          strlcpy(str, "others            :", sizeof(str));
          cat_file_permissions(str,perm & 0xFF);
          return send_reply(250,str);
        } else errno = EACCES;
      }
    }
    if (se >= 0) se = read_permission_file(*p,&uid,&gid,&perm,NULL,NULL,NULL);
  } else
  { se = permission_ok("",PE_CHDIR);
    if (se >= 0) se = read_permission_file("",&uid,&gid,&perm,NULL,NULL,NULL);
  }
  if (se >= 0)
  { strlcpy(str, "directory         :", sizeof(str));
    if ((perm >> 24) & DP_HIDDEN)
      strlcat(str, " !Hhidden!0", sizeof(str));
    else if (directory_flag_set_recursive(*p, DP_HIDDEN))
      strlcat(str," !Hhidden (by recursivity)!0", sizeof(str));
    if (!((perm >> 24) & 0xFF))
      strlcat(str, " (!Hno flags!0)", sizeof(str));
    if (send_message(250,0,str,NULL) == -1)
      return -1;
    strlcpy(str, "owner [", sizeof(str));
    name = get_username(uid);
    if (name)
      g_snprintf(str + strlen(str), sizeof(str) - strlen(str), "!H%-9s!0", name);
    else
      g_snprintf(str + strlen(str), sizeof(str) - strlen(str), "!H%-9"I32"!0", uid);
    strlcat(str, "] :", sizeof(str));
    cat_permissions(str,(perm >> 16) & 0xFF);
    if (send_message(250,0,str,NULL) == -1)
      return -1;
    strlcpy(str, "group [", sizeof(str));
    name = get_groupname(gid);
    if (name)
      g_snprintf(str+strlen(str), sizeof(str) - strlen(str), "!H%-9s!0",name);
    else
      g_snprintf(str+strlen(str), sizeof(str) - strlen(str), "!H%-9"I32"!0",gid);
    strlcat(str, "] :", sizeof(str));
    cat_permissions(str,(perm >> 8) & 0xFF);
    if (send_message(250,0,str,NULL) == -1)
      return -1;
    strlcpy(str, "others            :", sizeof(str));
    cat_permissions(str,perm & 0xFF);
    return send_reply(250,str);
  }
  return send_error(550,"mod",errno);
}

int cmd_site_open(char** p, int n)
{
   char str[128];

   if (n) return cat_help("open",501);
   fputs("SRV_SST on\n",stdout);
   if (!fgets(str,sizeof(str),stdin)) strlcpy(str, "-!Bpipe error!0", sizeof(str));
   return send_reply(*str == '+' ? 200 : 550,str+1);
}


int cmd_site_passwd(char ** p, int n)
{
   char str[128];

   if ((n < 1) || (n > 2)) return cat_help("passwd",501);
   printf("USR_PWD %s %s\n",n == 1 ? usr.name : p[0],
      strcmp(p[n-1],"-") ? p[n-1] : "");
   if (!fgets(str,sizeof(str),stdin)) return send_reply(550,"!Bpipe error!0");
   if (*str=='-') return send_reply(550,str+1);
   return send_reply(200, str+1);
}


int cmd_site_reg(char** p, int n)
{ int   i;
  char  in[256];
  int   se = 0;
  char  str[256];

  if (n < 1) return cat_help("reg", 501);
  if (!strcasecmp(p[0],"DEL") || !strcasecmp(p[0],"RM"))
  { if (n != 2) se = 1; else printf("REG_RM %s\n",p[1]);
  } else if (!strcasecmp(p[0],"GET"))
  { if (n != 2) se = 1; else printf("REG_GET %s\n",p[1]);
  } else if (!strcasecmp(p[0],"LIST"))
  { if (n != 1) se = 1; else
    { printf("REG_LS\n");
      if (!fgets(in,sizeof(in),stdin)) strlcpy(in, "-!Bpipe error!0", sizeof(in));
      while ((*in == '+') && strcmp(in,"+\n"))
      {
	if (send_reply_full (211, FALSE, FALSE, TRUE, in+1) == -1)
	  return -1;

        if (!fgets(in,sizeof(in),stdin))
	  strlcpy(in, "-", sizeof(in));
      }
      if (*in == '+') return send_reply(211,"!H<end of list>!0");
      if (strcmp(in,"-"))
	return send_reply_raw (550,in+1);
      else
	return send_reply (211, "!Bpipe error, aborting list!0");
    }
  } else if (!strcasecmp(*p,"SET"))
  { if (n < 3) se = 1; else
    { printf("#2#REG_SET %s",p[1]);
      for (i=2; i<n; i++) printf(" %s",p[i]);
      printf("\n");
    }
  } else se = 1;
  if (se) return cat_help("reg",501);
  if (!fgets(str,sizeof(str),stdin)) strlcpy(str, "-!Bpipe error!0", sizeof(str));
  return send_reply (*str == '+' ? 200 : 550, str+1);
}


int cmd_site_rights(char **p,int n)
{
   char str[128];
   if ((n != 1)) printf("PR_RGT u %s\n", usr.name); /* u = user, g = group */
   else printf("PR_RGT u %s\n", p[0]);

   if (!fgets(str,sizeof(str),stdin)) strlcpy(str, "-!Bpipe error!0", sizeof(str));
   while (*str == '+' && strcmp(str, "+\n")) {
      if (send_message(211,0,str+1,NULL) == -1) return -1;
      if (!fgets(str,sizeof(str),stdin)) strlcpy(str, "-", sizeof(str));
   }
   if (*str == '+') return send_reply(211,"!H<end of list>!0");
   if (strcmp(str,"-")) return send_reply(550,str+1);
   else return send_reply(211,"!Bpipe error, aborting list!0");
}


int cmd_site_run_script(char** p, int n, int display_exit)
{
   char* c;
   int   err;
   FILE* file;
   pid_t pid;
   int   pipefd[2];
   char  str[256];
   char  str2[256];

   if (!superuser() && display_exit)
      return send_reply(550, "!BOnly superusers may use this command.!0");
   if (!n)
      return cat_help("run",501);
   err = socketpair(AF_UNIX, SOCK_STREAM, 0, pipefd);
   if (err == -1)
      return send_error(550, "pipe", 0);

   set_enviroments(str, sizeof(str), vfs_getcwd());

   pid = fork();
   if (pid == -1)
      return send_error(550, "fork", errno);
   if (pid) {
      close(pipefd[1]);
      file = fdopen(pipefd[0], "rt");
      if (!file)
         return send_reply(550, "fdopen");
      if (display_exit)
         c = (char*) - 1;
      else {
         c = fgets(str, sizeof(str) - 1, file);
         if (!c)
            strlcpy(str, "!H(no output from program)!0", sizeof(str));
      }
      while (c) {
         c = fgets(str2, sizeof(str2) - 1, file);
         if (c && strlen(str2)) {
            if (display_exit) {
               if (send_message(200, 0, str2, NULL) == -1)
                  return -1;
            }
            else {
               if (send_message(200, 0, str, NULL) == -1)
                  return -1;
               memset(str, 0, sizeof(str));
               strlcpy(str, str2, sizeof(str));
            }
         }
      }
      fclose(file);
      close(pipefd[0]);
      pid = wait(&err);
      if (WIFEXITED(err)) {
         if (WEXITSTATUS(err) == 255)
            g_snprintf(str, sizeof(str), "!BScript does not exist.!0");
         else if (display_exit)
            g_snprintf(str, sizeof(str), "Script finished with exitcode !H%i!0.",
                     WEXITSTATUS(err));
      }
      else if (display_exit) {
         g_snprintf(str, sizeof(str), "Script terminated on signal !H%i!0.",
                  WTERMSIG(err));
      }
      return send_reply(200, str);
   }
   else {
      fclose(sockf);
      if (strcmp(vfs_getcwd (), "/"))
         chdir(vfs_getcwd () + 1);
      close(pipefd[0]);
      err = dup2(pipefd[1], 1);
      if (err != -1)
         err = dup2(pipefd[1], 2);
      if (err != -1) err = dup2(pipefd[1], 0);
      close(pipefd[1]);
      *(p+n) = NULL;
      if (err != -1) {
         memset(str, 0, sizeof(str));
         strlcpy(str, server_dir, sizeof(str));
         strlcat(str, DFL_BINDIR, sizeof(str));
         strlcat(str, *p, sizeof(str));
         execv(str,p);
         memset(str, 0, sizeof(str));
         strlcpy(str, server_dir, sizeof(str));
         strlcat(str, DFL_BINDIR, sizeof(str));
         c = get_groupname(usr.gid);
         if (c) {
            strlcat(str, c, sizeof(str));
            strlcat(str, "/", sizeof(str));
            strlcat(str, *p, sizeof(str));
            err = execv(str,p);
         }
      }
      exit(-1);
   }
}


int cmd_site_run(char** p, int n)
{
   return cmd_site_run_script(p,n,1);
}

int cmd_site_sec(char** p, int n)
{
   int   i;
   int   se = 0;
   char  str[256];

   if (n < 1) return cat_help("sec", 501);
   if (!strcasecmp(p[0],"LIST")) {
      printf("SEC_LS %s\n", n < 2 ? "": p[1]);
      return send_ftpd_reply(NULL);
   }
   else if (!strcasecmp(*p,"SET")) {
      if (n < 4)
         se = 1;
      else {
         printf("SEC_SET %s",p[1]);
         for (i=2; i<n; i++)
            printf(" %s",p[i]);
         printf("\n");
      }
   }
   else
      se = 1;
   if (se)
      return cat_help("sec",501);
   if (!fgets(str,sizeof(str),stdin)) strlcpy(str, "-!Bpipe error!0", sizeof(str));
   return send_reply_raw (*str == '+' ? 200 : 550,str+1);
}

int cmd_site_section(char** p, int n)
{
   int  se = 0;
   char str[1024];
   char *username;
   char *tmpp;

  if (!n) return cat_help("section",501);
  else if (!strcasecmp(*p,"CHANGE")) {
     if (n != 4) return cat_help("section.change",501);
     else {
        printf("S_CHG %s %s %s",p[1],p[2],p[3]);
        username=p[2];
        tmpp=p[3];
        n -= 4;
        p += 4;
        while (n--) {
           printf(" %s",*p);
           p++;
        }
        printf("\n");
        if (!fgets(str,sizeof(str),stdin)) return send_reply(550,"!Bpipe error!0");
        if (*str=='-') return send_reply(550,str+1);
        else {
           return send_reply(200,str+1);
        }
     }
  }
  else if (!strcasecmp(*p,"LIST")) {
     if (n != 3 || (strcasecmp(p[2],"day") && strcasecmp(p[2],"week") &&
        strcasecmp(p[2],"month") && strcasecmp(p[2],"year") &&
        strcasecmp(p[2], "total"))) se = 1;
     else {
        printf("S_LIST %s %s\n", p[1], p[2]);
        if (!fgets(str,sizeof(str),stdin)) strlcpy(str, "-!Bpipe error!0", sizeof(str));
        while (*str=='+' && strcmp(str,"+\n")) {
           if (send_message(211,0,str+1,NULL) == -1) return -1;
           if (!fgets(str,sizeof(str),stdin)) strlcpy(str, "-", sizeof(str));
        }
        if (*str == '+') return send_reply(211,"!H<end of list>!0");
        if (strcmp(str,"-")) return send_reply(550,str+1);
        else return send_reply(211,"!Bpipe error, aborting list!0");
     }
  }
  else if (!strcasecmp(*p,"RESET")) {
     if (n != 4) return cat_help("section.reset",501);
     else
        printf("S_RST %s %s %s\n", p[1], p[2], p[3]);
  }
  else se = 1;

  if (se) return cat_help("section",501);
  if (!fgets(str,sizeof(str),stdin)) strlcpy(str, "-!Bpipe error!0", sizeof(str));
  return send_reply(*str == '+' ? 200 : 550,str+1);
}

int cmd_site_shutdown(char** p, int n)
{ char str[128];

  if (n > 1) return cat_help("shutdown",501);
  if (n)
  { if (!strcasecmp(p[0],"normal")) g_snprintf(str, sizeof(str), "sn"); else
      if (!strcasecmp(p[0],"transactional")) g_snprintf(str, sizeof(str), "st"); else
      if (!strcasecmp(p[0],"immediate")) g_snprintf(str, sizeof(str), "si"); else
      return cat_help("shutdown",501);
  } else g_snprintf(str, sizeof(str), "si");
  printf("SRV_SST %s\n",str);
  if (!fgets(str,sizeof(str),stdin)) strlcpy(str, "-!Bpipe error!0", sizeof(str));
  if (*str == '-') return send_reply(550,str+1);
  quit(CL_SHUTDOWN);
  return send_reply(421,str+1);
}

int cmd_site_stats(char **p, int n)
{
   char str[256];

   if ((n != 1)) printf("ULS %s\n", usr.name);
   else printf("ULS %s\n",p[0]);

   if (!fgets(str,sizeof(str),stdin)) strlcpy(str, "-!Bpipe error!0", sizeof(str));
   while (*str == '+' && strcmp(str, "+\n")) {
      if (send_message(211,0,str+1,NULL) == -1) return -1;
      if (!fgets(str,sizeof(str),stdin)) strlcpy(str, "-", sizeof(str));
   }
   if (*str == '+') return send_reply(211,"!H<end of list>!0");
   if (strcmp(str,"-")) return send_reply(550,str+1);
   else return send_reply(211,"!Bpipe error, aborting list!0");
}


int cmd_site_w(char** p, int n)
{
   char  str[512];
   char  str2[512];
   char* strptr;
   int   hidden;

   hidden = directory_flag_set_recursive(vfs_getcwd (), DP_HIDDEN);
   update_user_activity(hidden, ACT_SITE_WHO, 0, vfs_getcwd ());
   if (n) return cat_help("w",501);
   printf("WHO\n");
   strptr = str;
   if (!fgets(strptr,sizeof(str),stdin)) {
      strlcpy(str,"-!Bpipe error, aborting list!0", sizeof(str));
   }
   if (*strptr == '-') return send_reply(550,strptr+1);
   n = 0;
   while (strcmp(strptr,"+\n")) {
      strptr = (strptr == str) ? str2 : str;

      if (!fgets(strptr, (strptr == str) ? sizeof(str) : sizeof(str2), stdin))
         strlcpy(strptr, "-!Bpipe error!0, aborting list", (strptr == str) ? sizeof(str) : sizeof(str2));

      if (send_message(211,!strcmp(strptr,"+\n"), strptr == str ? str2+1 : str+1,NULL) == -1)
	return -1;
      if (*strptr == '-')
	return send_reply(211,strptr+1);
      n++;
   }
   return n ? 0 : send_reply(211,"!H<empty list>!0");
}


int cmd_site_user(char** p, int n)
{
   int  se = 0;
   char str[1024];
   char *username;
   char *tmpp;

   if (!n) return cat_help("user", 501);
   else if (!strcasecmp(*p, "CHANGE")) {
      if (n < 3) return cat_help("user.change", 501);
      else {
         printf("USR_CHG %s '%s'",p[1], p[2]);
         username = p[1];
         tmpp = p[2];
         n -= 3;
         p += 3;
         while (n--) {
            printf(" %s", *p);
            p++;
         }
         printf("\n");
         if (!fgets(str,sizeof(str),stdin)) {
	    return send_reply(550, "!Bpipe error!0");
         }
	 if (*str=='-') return send_reply(550,str+1);
         else {
            if (showinfo) {
               if (!strncasecmp("rights", tmpp, 6)) {
                  return cmd_site_rights(&username, 1);
               }
               else if (!strncasecmp("name", tmpp, 4)) {
                  username = strchr(tmpp, '=');
                  if (username)
                     username++;
               }
               if (username) printf("PR_UI %s\n", username);
               return send_ftpd_reply("!HUser changed.!0");
            }
            return send_reply(200,str+1);
         }
      }
   }

   else if(!strcasecmp(*p, "LIST")) {
      if (n > 3) se = 1;
      else {
         if (n == 3) printf("PR_UL %s %s\n", *(p+1), *(p+2));
         else if (n == 2) printf("PR_UL %s\n", *(p+1));
	 else printf("PR_UL\n");
         return send_ftpd_reply(NULL);
      }
   }
   //for backwards compatibility, just call the site who function
   else if(!strcasecmp(*p, "LISTA")) {
     return cmd_site_w(p,0);
   }
   else if(!strcasecmp(*p, "SHOW")) {
      if (n > 3) se = 1;
      else {
         printf("PR_UI %s\n", *(p+1));
         return send_ftpd_reply(NULL);
      }
   }
   else if (!strcasecmp(*p, "REMOVE")) {
      if (n != 2) se = 1;
      else printf("USR_RM %s\n",p[1]);
   }

   else if (!strcasecmp(*p, "RIGHTS")) {
      if (n != 2) se = 1;
      else {
         printf("PR_RGT u %s\n", p[1]);
         if (!fgets(str,sizeof(str), stdin)) strlcpy(str, "-!Bpipe error!0", sizeof(str));
         while (*str == '+' && strcmp(str, "+\n")) {
            if (send_message(211, 0, str+1, NULL) == -1) return -1;
            if (!fgets(str,sizeof(str),stdin)) strlcpy(str, "-", sizeof(str));
         }
         if (*str == '+') return send_reply(211, "!H<end of list>!0");
         if (strcmp(str,"-")) return send_reply(550, str+1);
         else return send_reply(211, "!Bpipe error, aborting list!0");
      }
   }

   else if (!strcasecmp(*p,"ADD")) {
      if ((n < 2) || (n > 3)) return cat_help("user.add",501);
      else {
         printf("USR_ADD %s %s\n",p[1],n == 3 ? p[2] : "");
         if (!fgets(str,sizeof(str),stdin)) return send_reply(550,"!Bpipe error!0");
         if (*str=='-') return send_reply(550,str+1);
	 if (showinfo) {
	    printf("PR_UI %s\n", *(p+1));
	    return send_ftpd_reply("!HUser added.!0");
	 }
         return send_reply(200,str+1);
      }
   }
   else if (!strcasecmp(*p,"RESET")) {
      if ((n < 2) || (n > 3)) return cat_help("user.reset",501);
      else
         printf("USR_RST %s %s\n",p[1],n == 3 ? p[2] : "");
   }
   else se = 1;

   if (se) return cat_help("user",501);
   if (!fgets(str,sizeof(str),stdin)) strlcpy(str, "-!Bpipe error!0", sizeof(str));
   return send_reply(*str == '+' ? 200 : 550,str+1);
}


int cmd_site_deluser(char **p,int n)
{
   char str[128];
   if ( (n<1) || (n>1)) {
      return cat_help("deluser",501);
   }
   else {
      printf("USR_RM %s\n", p[0]);
   }
   if (!fgets(str, sizeof(str), stdin)) strlcpy(str, "-!Bpipe error!0", sizeof(str));
   return send_reply(*str == '+' ? 200 : 550, str + 1);
}


int cmd_site_luser(char **p,int n)
{
   if (n != 1) return cat_help("luser", 501);
   printf("PR_UI %s\n", p[0]);
   return send_ftpd_reply(NULL);
}


int cmd_site_lgroup(char **p,int n)
{
   if (n != 1) return cat_help("lgrp", 501);
   printf("PR_UL -G %s\n", p[0]);
   return send_ftpd_reply(NULL);
}


int cmd_site_users(char **p,int n)
{
   if (n > 2) return cat_help("users", 501);
   printf("PR_UL %s %s\n", n > 0 ? *p : "", n > 1 ? *(p+1) : "");
   return send_ftpd_reply(NULL);
}


/*-------------- SITE ADDIP     --------------*/
int cmd_site_addip(char **p,int n)
{
   char str[1024];
   int counter = 1;

   if ( (n < 2) || (n > 11) ) {
      return cat_help("addip",501);
   }
   else{
      for(counter = 1; counter < n; counter++) {
         printf("USR_CHG %s ip+=%s\n",p[0],p[counter]);
         if(!fgets(str, sizeof(str),stdin)) return send_reply(550,"!Bpipe error!0");
         if(*str=='-') return send_reply(550,str+1);
         }
   }

   if (showinfo) {
      printf("PR_UI %s\n", *p);
      return send_ftpd_reply("!HUser changed.!0");
   }
   return send_reply(200,str+1);
}


/*--------------   SITE CHGRP   -----------------*/
int cmd_site_chgrp(char **p,int n)
{
   char str[1024];

   if ( (n < 2) || (n > 3) ) return cat_help("chgrp",501);
   printf("USR_CHG %s GROUP=%s\n",p[0],p[1]);
   if (!fgets(str,sizeof(str),stdin)) return send_reply(550,"!Bpipe error!0");
   if (*str=='-') return send_reply(550,str+1);
   if (showinfo) {
      printf("PR_UI %s\n", *p);
      return send_ftpd_reply("!HUser changed.!0");
   }
   return send_reply(200,str+1);
}


/*-------------- SITE DELIP --------------*/
int cmd_site_delip(char **p, int n)
{
	char str[1024];
	int counter = 1;
	if ( (n < 2) || (n > 11) ) {
		return cat_help("delip",501);
	}
	else{
		for(counter = 1; counter < n; counter++) {
		    printf("USR_CHG %s ip-=%s\n",p[0],p[counter]);
		    if (!fgets(str,sizeof(str),stdin)) return send_reply(550,"!Bpipe error!0");
		    if (*str=='-') return send_reply(550,str+1);
		}
	}
	if (showinfo) {
	    printf("PR_UI %s\n", *p);
	    return send_ftpd_reply("!HUser changed.!0");
	}
   return send_reply(200,str+1);
}


int cmd_site_addgrp(char **p,int n)
{
   char str[128];

   if (n != 1) return cat_help("addgrp", 501);
   printf("GRP_ADD %s -0\n", p[0]);
   if (!fgets(str, sizeof(str), stdin)) return send_reply(550, "!Bpipe error!0");
   if(*str == '-') return send_reply(550, str+1);
   if (showinfo) {
      printf("PR_GI %s\n", *p);
      return send_ftpd_reply(str+1);
   }
   return send_reply(200, str+1);
}


int cmd_site_addnuker(char **p,int n)
{
   char str[128];

   if (n != 1) return cat_help("addnuker", 501);
   printf("USR_CHG %s RIGHTS+=R_NUKE\n",p[0]);
   if (!fgets(str,sizeof(str),stdin)) return send_reply(550,"!Bpipe error!0");
   if (*str=='-') return send_reply(550,str+1);
   if (showinfo) {
      printf("PR_UI %s\n", *p);
      return send_ftpd_reply("!HUser changed.!0");
   }
   return send_reply(200,str+1);
}


int cmd_site_delnuker(char **p,int n)
{
   char str[128];

   if ((n < 1) || (n >1)) return cat_help("delnuker",501);
   printf("USR_CHG %s RIGHTS-=R_NUKE\n",p[0]);
   if (!fgets(str,sizeof(str),stdin)) return send_reply(550,"!Bpipe error!0");
   if (*str=='-') return send_reply(550,str+1);
   if (showinfo) {
      printf("PR_UI %s\n", *p);
      return send_ftpd_reply("!HUser changed.!0");
   }
   return send_reply(200,str+1);
}


int cmd_site_addgroupop(char **p,int n)
{
   char str[128];

   if ( (n < 1) || (n >1) ) return cat_help("addgroupop",501);
   printf("GOP_ADD %s\n", p[0]);
   if (!fgets(str,sizeof(str),stdin)) return send_reply(550,"!Bpipe error!0");
   if (*str=='-') return send_reply(550,str+1);
   return send_reply(200,str+1);
}


int cmd_site_delgroupop(char **p,int n)
{
   char str[128];
   if ( (n < 1) || (n >1) ) return cat_help("delgroupop",501);
   printf("GOP_DEL %s\n", p[0]);
   if (!fgets(str,sizeof(str),stdin)) return send_reply(550,"!Bpipe error!0");
   if (*str=='-') return send_reply(550,str+1);
   return send_reply(200,str+1);
}


int cmd_site_addsiteop(char **p,int n)
{
   char str[128];
   if ( (n < 1) || (n >1) ) return cat_help("addsiteop",501);
   printf("OP_ADD %s\n", p[0]);
   if (!fgets(str,sizeof(str),stdin)) return send_reply(550,"!Bpipe error!0");
   if (*str=='-') return send_reply(550,str+1);
   return send_reply(200,str+1);
}


int cmd_site_delsiteop(char **p,int n)
{
   char str[128];
   if ( (n < 1) || (n >1) ) return cat_help("delsiteop",501);
   printf("OP_DEL %s\n", p[0]);
   if (!fgets(str,sizeof(str),stdin)) return send_reply(550,"!Bpipe error!0");
   if (*str=='-') return send_reply(550,str+1);
   return send_reply(200,str+1);
}


int cmd_site_addunduper(char **p,int n)
{
   char str[128];
   if ( (n < 1) || (n >1) ) return cat_help("addunduper",501);
   printf("USR_CHG %s RIGHTS+=R_UNDUPE\n",p[0]);
   if (!fgets(str,sizeof(str),stdin)) return send_reply(550,"!Bpipe error!0");
   if (*str=='-') return send_reply(550,str+1);
   if (showinfo) {
      printf("PR_UI %s\n", *p);
      return send_ftpd_reply("!HUser changed.!0");
   }
   return send_reply(200,str+1);
}


int cmd_site_delunduper(char **p,int n)
{
   char str[128];
   if ( (n < 1) || (n >1) ) return cat_help("delunduper",501);

   printf("USR_CHG %s RIGHTS-=R_UNDUPE\n",p[0]);
   if (!fgets(str,sizeof(str),stdin)) return send_reply(550,"!Bpipe error!0");
   if (*str=='-') return send_reply(550,str+1);
   if (showinfo) {
      printf("PR_UI %s\n", *p);
      return send_ftpd_reply("!HUser changed.!0");
   }
   return send_reply(200,str+1);
}


int cmd_site_delgrp(char **p,int n)
{
   char str[128];

   if ((n < 1) || (n > 1)) return cat_help("delgrp",501);
   else printf("GRP_RM %s\n", p[0]);
   if (!fgets(str, sizeof(str), stdin)) strlcpy(str, "-!Bpipe error!0", sizeof(str));
   return send_reply(*str == '+' ? 200 : 550, str+1);
}


int cmd_site_kill(char**p, int n)
{
   int pid;
   char str[128];

   if ((n < 1) || ( n > 1)) return cat_help("kill",501);
   else {
      sscanf(p[0],"%i",&pid);
      printf("USR_KIL %i\n",pid);
   }
   if (!fgets(str, sizeof(str), stdin)) strlcpy(str, "-!Bpipe error!0", sizeof(str));
   return send_reply(*str == '+' ? 200 : 550, str+1);
}


int cmd_site_undupe(char**p, int n)
{
   char str[128];
   int  number = 0;

   if ((n < 1) || ( n > 2)) return cat_help("undupe", 501);
   else {
      (n == 1) ? printf("DUP_UND \"%s\"\n", p[0]) :
                 printf("DUP_UND %s \"%s\"\n", p[0], p[1]);
   }
   if (!fgets(str, sizeof(str), stdin)) strlcpy(str, "-!Bpipe error!0", sizeof(str));
   if (*str != '+') {
      return send_reply(550, str + 1);
   }
   sscanf(str + 1, "%d", &number);
   g_snprintf(str, sizeof(str), "!H%d!C files unduped.!0\n", number);
   return send_reply(200, str);
}


int cmd_site_swho(char** p, int n)
{
   char  str[512];
   char  str2[512];
   char* strptr;
   int   hidden;

   hidden = directory_flag_set_recursive(vfs_getcwd (), DP_HIDDEN);
   update_user_activity(hidden, ACT_SITE_WHO, 0, vfs_getcwd ());
   if (n) return cat_help("w",501);
   printf("SWHO\n");
   strptr = str;
   if (!fgets(strptr,sizeof(str),stdin)) {
      strlcpy(str,"-!Bpipe error, aborting list!0", sizeof(str));
   }
   if (*strptr == '-') return send_reply(550, strptr + 1);
   n = 0;
   while (strcmp(strptr, "+\n")) {
      strptr = (strptr == str) ? str2 : str;
      if (!fgets(strptr,sizeof(str),stdin))
	strlcpy(strptr, "-!Bpipe error, aborting list!0", (strptr == str) ? sizeof(str) : sizeof(str2));

      if (send_message(211, !strcmp(strptr,"+\n"), strptr == str ? str2 + 1 : str + 1,NULL) == -1) return -1;
      if (*strptr == '-') return send_reply(211, strptr + 1);
      n++;
   }
   return n ? 0 : send_reply(211, "!H<empty list>!0");
}


int cmd_site_traffic(char **p,int n)
{
   char str[128];
   if  (n != 0) return cat_help("traffic",501);
   else {
      printf("TRF\n");
      if (!fgets(str,sizeof(str),stdin)) strlcpy(str, "-!Bpipe error!0", sizeof(str));
      while (*str == '+' && strcmp(str, "+\n")) {
	 if (send_message(211,0,str+1,NULL) == -1) return -1;
	 if (!fgets(str,sizeof(str),stdin)) strlcpy(str, "-", sizeof(str));
      }
      if (*str == '+') return send_reply(211, "!0<end of list>!0");
      if (strcmp(str,"-")) return send_reply(550, str + 1);
      else return send_reply(211, "!Bpipe error, aborting list!0");
   }
   if (!fgets(str,sizeof(str),stdin)) strlcpy(str, "-!Bpipe error!0", sizeof(str));
   return send_reply(*str == '+' ? 200 : 550, str + 1);
}


int cmd_site_whoami(char **p,int n)
{
   char str[32];
   g_snprintf(str, sizeof(str), "You are: !H%s!0", usr.name);
   return send_reply(200, str);
}


int cmd_site_ff(char **p, int n)
{
   if ((n < 1) || ( n > 2)) return cat_help("ff", 501);
   else {
      (n == 1) ? printf("DUP_FF \"%s\"\n", p[0]) :
                 printf("DUP_FF %s \"%s\"\n", p[0], p[1]);
   }
   return send_ftpd_reply(NULL);
}


int cmd_site_fullspeed(char **p, int n)
{
	if (!superuser()) {
		return send_reply(200, "You are not allowed to use this command.");
	}
	if (fullspeed == 1) {
		fullspeed = 0;
		return send_reply(200, "Speed limits enabled.");
	}
	else {
		fullspeed = 1;
		return send_reply(200, "Speed limits disabled.");
	}
}


int cmd_site_help(char**,int);

int cmd_site_xdupe(gchar **p, int n) {
  gchar str[256];
  gint mode = 0;

  /* if no mode number was supplied or it was not a digit (0-9) */
  if(n != n || !isdigit(*p[0]))
    return send_reply(501, "Syntax: XDUPE <mode>");

  mode = atoi(p[0]);

  /* for the time being, there are only five modes (0 being disable) */
  if(mode > 4)
    return send_reply(501, "Syntax: XDUPE <mode (0-4)>");

  printf("X_DUPE %d\n", mode);

  if(mode) {
    g_snprintf(str, sizeof(str), "Activated extended dupe mode %d", mode);
    return send_reply(200, str);
  }
  else
    return send_reply(200, "Disabled extended dupe mode");
}

struct sitecmd_s
{
   char command[10];
   int  (*function)(char**,int);
} site_command[] =
{
  { "ADDGROUP",   cmd_site_addgrp      },
  { "ADDGROUPOP", cmd_site_addgroupop  },
  { "ADDGRP",     cmd_site_addgrp      },
  { "ADDIP",      cmd_site_addip,      },
  { "ADDNUKER",   cmd_site_addnuker,   },
  { "ADDSITEOP",  cmd_site_addsiteop   },
  { "ADDUNDUPER", cmd_site_addunduper, },
  { "ADDUSER",    cmd_site_adduser,    },
  { "BIND",       cmd_site_bind,       },
  { "CHGRP",      cmd_site_chgrp,      },
  { "CHMOD",      cmd_site_chmod,      },
  { "CHMODR",     cmd_site_chmodr,     },
  { "CHOWN",      cmd_site_chown,      },
  { "CHOWNR",     cmd_site_chownr,     },
  { "CLOSE",      cmd_site_close,      },
  { "COLOR",      cmd_site_color,      },
  { "COUNT",      cmd_site_count,      },
  { "DEL",        cmd_site_del,        },
  { "DELGROUP",   cmd_site_delgrp,     },
  { "DELGROUPOP", cmd_site_delgroupop  },
  { "DELGRP",     cmd_site_delgrp,     },
  { "DELIP",      cmd_site_delip,      },
  { "DELNUKER",   cmd_site_delnuker,   },
  { "DELSITEOP",  cmd_site_delsiteop   },
  { "DELUNDUPER", cmd_site_delunduper, },
  { "DELUSER",    cmd_site_deluser,    },
  { "DISABLE",    cmd_site_disable,    },
  { "DUMP",       cmd_site_dump,       },
  /* { "DUPE",       cmd_site_ff,         }, */
  { "ENABLE",     cmd_site_enable,     },
  { "EXEC",       cmd_site_exec,       },
  /* { "FF",         cmd_site_ff,         }, */
  { "FULLSPEED",  cmd_site_fullspeed,  },
  { "GADDUSER",   cmd_site_gadduser,   },
  { "GINFO",      cmd_site_ginfo,      },
  { "GIVE",       cmd_site_give,       },
  { "GROUP",      cmd_site_group,      },
  { "GROUPS",     cmd_site_groups,     },
/*   { "HELP",       cmd_site_help,       }, */
  { "IDLE",       cmd_site_idle,       },
  { "INFO",       cmd_site_info,       },
  { "IPBAN",      cmd_site_ipban,      },
  { "KICK",       cmd_site_kick,       },
  { "KILL",       cmd_site_kill,       },
  { "LGROUP",     cmd_site_lgroup,     },
  { "LGRP",       cmd_site_lgroup,     },
  { "LUSER",      cmd_site_luser,      },
  { "MOD",        cmd_site_mod,        },
  { "NUKE",       cmd_site_nuke,       },
  { "OPEN",       cmd_site_open,       },
  { "PASSWD",     cmd_site_passwd,     },
  { "REG",        cmd_site_reg,        },
  { "RIGHTS",     cmd_site_rights,     },
  { "RUN",        cmd_site_run,        },
  { "SEC",        cmd_site_sec,        },
  { "SECTION",    cmd_site_section,    },
  { "SETGINFO",   cmd_site_setginfo,   },
  { "SETINFO",    cmd_site_setinfo,    },
  { "SHUTDOWN",   cmd_site_shutdown,   },
  { "STATS",      cmd_site_stats,      },
  { "SWHO",       cmd_site_swho,       },
  { "TAKE",       cmd_site_take,       },
  { "TRAFFIC",    cmd_site_traffic,    },
  { "UNDUPE",     cmd_site_undupe,     },
  { "UNFO",       cmd_site_info,       },
  { "UNNUKE",     cmd_site_unnuke,     },
  { "UPTIME",     cmd_site_uptime,     },
  { "USER",       cmd_site_user,       },
  { "USERS",      cmd_site_users,      },
  { "USR",        cmd_site_user,       },
  { "VERSION",    cmd_site_version,    },
  { "W",          cmd_site_w,          },
  { "WHO",        cmd_site_w,          },
  { "WHOAMI",     cmd_site_whoami,     },
  { "XDUPE",      cmd_site_xdupe,      }
};

/* int cmd_site_help(char** p, int n) */
/* { char str[256]; */
/*   int i; */
/*   char *po; */

/*   *str=0; */
/*   if (n == 0) return cat_help("help",501); */
/*   for (i = 0; i< n; ++i) */
/*     g_snprintf(str+strlen(str), sizeof(str) - strlen(str), "%s.",p[i]); */
/*   *(str+strlen(str)-1)=0; */

/*   / next line is in case someone try site help ../../... / */
/*   if(NULL != strchr(str,'/')) return cat_help("help",501); */

/*   / in case someon try site help help.privileged or things like this / */
/*   if(NULL != strstr(str,".privileged")) return cat_help("help",501); */

/*   po=str; */
/*   while(*po) { */
/*     *po=tolower(*po); */
/*     ++po; */
/*   } */
/*   return cat_help(str,214); */
/* }  */

/*
   SITE SITE <command> overwrites custom commands if an internal command
   of the same name exists
*/

int cmd_site(char** p, int n)
{ int   i = 0;
  int   j;
  int   k = (sizeof(site_command) / sizeof(struct sitecmd_s)) - 1;
  int   r;
  char* q;
  char str[256];
  struct customcmd_s* cmd;
  char buff[sizeof(cmd->param)];
  char* p2[n+(sizeof(cmd->param)/2)];

  if (!n) return cat_help("help",501);

#if defined (HAVE_PERL)
  g_assert (n == 0 || n == 1);

  if (! perl_call_subf ("FTP::cmd_site", (p[0] ? "s" : ""), p[0]))
    return 0;

  /* well... perl didn't handle it, so we must pass it on to the other builtins... */

  n++, p--; /* restore state... */
  if (n == 2)
    {
      n--;
      if (parse_command (*(p+1),p,31,&n))
	return send_reply(500, "Too many parameters.");
    }
  n--, p++; /* ... */
#endif /* defined (HAVE_PERL) */

  if (strcasecmp(*p, "SITE")) {
     for (i = 0, cmd = customcmd; i < customcmd_count; i++, cmd++) {
	if (!strncasecmp(cmd->command, *p, strlen(cmd->command))) {
           p2[0] = cmd->path;
	   if (strlen(cmd->param) > 0) {
              memset(buff,0,sizeof(buff));
              strlcpy(buff, cmd->param, sizeof(buff));
              i=0;
              q=buff;
              do {
                skipspace(&q);
                p2[i+1]=q;
                skipnspace(&q);
                *q=0; ++q; ++i;
              } while(*q);
	      for (j = 1; j < n; j++) {
	        p2[j+i] = p[j];
 	      }
	      p2[j+i+1] = NULL;
	      return cmd_site_run_script(p2, j+i, 0);
	   }
	   else {
	      for (j = 1; j < n; j++) {
	        p2[j] = p[j];
 	      }
	      p2[j+1] = NULL;
	      return cmd_site_run_script(p2, j, 0);
	   }
	}
     }
  }

  i = 0;
  if ((n >= 2) && !strcasecmp(*p,"SITE")) {
     p++;
     n--;
  }
  while (i <= k) {
     j = (i + k) >> 1;
     r = strcasecmp(*p, site_command[j].command);
     if (r < 0) k = j - 1; else if (r > 0) i = j + 1;
     else return site_command[j].function(p+1,n-1);
  }

  g_snprintf(str, sizeof(str), "!BUnknown SITE command: !H%s!0", *p);
  return send_reply(501, str);
}

/* EOF */
