/* glftpd -> OpenFTPD userfile converter (c) 2001 primemover */


#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>
#include <string.h>
#include "commondef.h"
#include "utils.h"
#include "../ftpd/definesd.h" /* hvr fixme -- $srcdir */

gchar *server_dir; /* workaround for mk_filename */

static int get_ident(char** ipr, char* ident)
{ 
   char* p;
   int	i;

   p = *ipr;
   i = 0;

   /* remove the ident if there is no @ in the whole ipstring */
   if ((char *)(strchr(*ipr, '@')) == NULL) { 
      ident[0]='\0';
      return 0;
   }
   while (*p && (*p!='@') && (*p!='.') && (*p!=',') && (*p!=':')) {i++; p++; }
   if (!i) return -1;
   if (*p=='@') { 
      *ipr=p+1;
      if (i>11) {p=p-i+11; i=11;}
      *p='\0';
      if (!sscanf(p-i,"%s",ident)) return -1;
   }
   if (ident[0]=='*') ident[0]='\0'; /* we dont need *@, so we remove the ident ... */
   return 0;
}


static int get_oneip(char** ipr, char* l, char* h)
{ 
   char  c;
   int   i;
   char* p;

   p = *ipr;
   if (!skipspace(&p)) return -1;
   if (*p == '*') { 
      *l = 0x00;
      *h = 0xFF; /* fixme: signed char overflow */
      p++;
      skipspace(&p);
      *ipr = p;
      return 0;
   }
   i = 0;
   while (*p && isdigit(*p) && (i<3)) { i++; p++; }
   if (!i) return -1;
   c = *p;
   *p = '\0';
   if (!sscanf(p-i,"%i",&i)) return -1;
   if ((i<0) || (i>255)) return -1;
   *l = i;
   *p = c;
   if (skipspace(&p) != '-') { 
      *h = i;
      *ipr = p;
      return 0;
   }
   p++;
   if (!skipspace(&p)) return -1;
   i = 0;
   while (*p && isdigit(*p) && (i<3)) { i++; p++; }
   if (!i || isdigit(*p)) return -1;
   c = *p;
   *p = '\0';
   if (!sscanf(p-i,"%i",&i)) return -1;
   if ((i<0) || (i>255)) return -1;
   *p = c;
   *h = i;
   skipspace(&p);
   *ipr = p;
   return 0;
}


static int get_iprange(char** ipr, char* ident)
{ 
   if (get_ident(ipr, ident) == -1) return -1;
   if (get_oneip(ipr,&ident[12],&ident[13]) == -1) return -1;
   if (**ipr != '.') return -1;
   (*ipr)++;
   if (get_oneip(ipr,&ident[14],&ident[15]) == -1) return -1;
   if (**ipr != '.') return -1;
   (*ipr)++;
   if (get_oneip(ipr,&ident[16],&ident[17]) == -1) return -1;
   if (**ipr != '.') return -1;
   (*ipr)++;
   if (get_oneip(ipr,&ident[18],&ident[19]) == -1) return -1;
   return 0;
}

struct uf_ident_s
{ char id[32];
  int  version;
  int  count;
  int  encrypted;
  int  week;
  int  month;
  int  year;
  int  day; /* This was temp1 changed now to day */
  int  temp2;
} u_id = { "OpenFTPD user database", 101, 0, 0, 0, 0, 0, 0, 0 }; 


struct gf_ident_s
{ 
   char id[32];
   int  version;
   int  count;
   int  encrypted;
   int  temp1;
} g_id = { "OpenFTPD group database", 101, 0, 0, 0 };


int main (int argc, char *argv[]) 
{
   int debug = 0; /* set this to 1 to enable debug mode */

   FILE *fin;
   FILE *fout;
   char buf[2048];
   char *s;
   gint32 i;
   gint64 l;
   char c;
   struct user_s *usr;
   struct group_s *grp;
   int usercount = 0;
   int groupcount = 0;
   struct tm* t;
   time_t     tim;
   
   if (argc < 3) {
      printf("Missing argument.\n");
      return -1;	
   }
      
   /* we get the number of users as argument, this value is
      written to the header of the binary userfile */
   if (sscanf(argv[1], "%d", &usercount) != 1) {
      printf("Invalid number of users.\n");	
      return -1;
   }
   if (sscanf(argv[2], "%d", &groupcount) != 1) {
      printf("Invalid number of groups.\n");	
      return -1;
   }
         
   fin = fopen("users.txt", "rt");
   if (fin == NULL) {
      printf("Unable to open sourcefile for import.\n");      
      return -1;
   }	

   /* open the output userfile for writing */
   fout = fopen("users", "wb");
   if (fout == NULL) {
      printf("Unable to open destination file.\n");      
      return -1;   	
   }
   
   /* write the fileheader */
   u_id.count = usercount;
   if (!fwrite(&u_id, sizeof(u_id), 1, fout)) { 
      printf("Unable to write file header.\n");
      fclose(fout);
      return -1;
   }

   time(&tim);
   t = localtime(&tim);
   t->tm_mon++;
   t->tm_year += 1900;
   if (t->tm_wday >= t->tm_yday) u_id.week = 53;
   else u_id.week = ((t->tm_yday-t->tm_wday) / 7) + 1;
   u_id.month = t->tm_mon;
   u_id.year = t->tm_year;
   u_id.day = t->tm_yday;

   if (debug) printf("Stats: day = %d, week = %d, month = %d\n", 
      u_id.day, u_id.week, u_id.month);


   /* read a line from the text userfile which was generated by
      glconv.pl from the original glftpd files */
   while (fgets(buf, 2048, fin)) {
      usr = g_malloc(sizeof(struct user_s));
      memset(usr, 0, sizeof(struct user_s));
      
      strcpy(usr->creator, "glconv");
      for (i=0; i<8; i++) usr->gids[i] = -1;

      usr->creation_time = tim;
	 
      /* username */
      s = (char *) strtok(buf, "\t");
      if (s) strncpy(usr->name, s, sizeof(usr->name) - 1);
      usr->name[strlen(usr->name)] = '\0';
      if (debug) printf("Name: %s\n", usr->name);
      
      /* password */
      s = (char *) strtok(NULL, "\t");
      if (s) strncpy(usr->password, s, sizeof(usr->password) - 1);
      usr->password[strlen(usr->password)] = '\0';
      if (debug) printf("Password: %s\n", usr->password);

      /* user id */
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, "%d", &i);      
      	 usr->uid = i;
         if (debug) printf("UID: %d\n", usr->uid);
      }

      /* group id */
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, "%d", &i);      
      	 usr->gid = i;
         if (debug) printf("GID: %d\n", usr->gid);
      }

      /* user class */
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, "%c", &c);
         usr->class = 0;
         switch (c) {
            case 'a': usr->class |= UC_ANON; 
                      break;
            case 's': usr->class |= UC_ROOT; 
	              break;
            case 'o': usr->class |= UC_SITEOP; 
       	              usr->rights = 0x7FFFFF7000F0000LL;;
	              break;
            case 'g': usr->class |= UC_GROUPOP; 
	              usr->rights = 0x08400B500000000LL;
		      break;
            default : usr->class |= UC_NORMAL; 
	              break;
         }
         if (debug) { 
	    printf("Class (%c)", c);
	    switch (usr->class & UC_MASK) { 
	       case UC_ROOT    : printf("SUPERUSER"); break;
	       case UC_ANON    : printf("ANONYMOUS"); break;
	       case UC_NORMAL  : printf("NORMAL"); break;
	       case UC_GROUPOP : printf("GROUPOP"); break;
	       case UC_SITEOP  : printf("SITEOP"); break;
	       case UC_TEMPLATE: printf("TEMPLATE"); break;
	       default         : printf("UNKNOWN"); break;
            }
	    printf("\n");
        }
      }

      /* ratio */
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, "%d", &i);      
      	 if (i == 0) usr->limit = 0;
      	 else usr->limit = 100 / i;
         if (debug) printf("Limit: %d\n", usr->limit);
      }

      /* credit */
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI64, &l);      
      	 usr->credit = (gint64)(l * 1024);
         if (debug) printf("Credit: "SI64" MB\n", usr->credit / 0x100000);
      }

      /* logins */
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, "%d", &i);      
      	 usr->class = (usr->class & ~UL_MASK) | ((gint32)(i) << 8);
         if (debug) printf("Logins: %d\n", (usr->class & UL_MASK) >> 8);
      }

      /* tagline */
      s = (char *) strtok(NULL, "\t");
      if (s) strncpy(usr->info, s, sizeof(usr->info) - 1);
      usr->info[strlen(usr->info)] = '\0';
      if (debug) printf("Tagline: %s\n", usr->info);

      /* max upload speed */
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI32, &i);      
      	 usr->ul_speed = 0;
         usr->ul_speed = (usr->ul_speed & 0xFFFF) | (i << 16);
         if (debug) printf("UL Max: "SI32"\n", (usr->ul_speed >> 16 )& 0xFFFF);
      }

      /* max download speed */
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI32, &i);      
      	 usr->dl_speed = 0;
         usr->dl_speed = (usr->dl_speed & 0xFFFF) | (i << 16);
         if (debug) printf("DL Max: "SI32"\n", (usr->dl_speed >> 16 )& 0xFFFF);
      }

      /* allup */
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI32, &i);      
      	 usr->ul_f_total = i;
         if (debug) printf("Total files UL: "SI32"\n", usr->ul_f_total);
      }
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI64, &l);      
      	 usr->ul_total = (gint64)(l * 1024);
         if (debug) printf("Total MB UL: "SI64"\n", usr->ul_total / 0x100000);
      }
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI64, &l);      
      	 usr->ul_tm_total = l;
         if (debug) printf("Total time UL: "SI64"\n", usr->ul_tm_total);
      }

      /* alldn */
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI32, &i);      
      	 usr->dl_f_total = i;
         if (debug) printf("Total files DL: "SI32"\n", usr->dl_f_total);
      }
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI64, &l);      
      	 usr->dl_total = (gint64)(l * 1024);
         if (debug) printf("Total MB DL: "SI64"\n", usr->dl_total / 0x100000);
      }
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI64, &l);      
      	 usr->dl_tm_total = l;
         if (debug) printf("Total time UL: "SI64"\n", usr->dl_tm_total);
      }

      /* dayup */
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI32, &i);      
      	 usr->ul_f_day = i;
         if (debug) printf("Today files UL: "SI32"\n", usr->ul_f_day);
      }
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI64, &l);      
      	 usr->ul_day = l * 1024;
         if (debug) printf("Today bytes UL: "SI64"\n", usr->ul_day);
      }
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI64, &l);      
      	 usr->ul_tm_day = l;
         if (debug) printf("Today time UL: "SI64"\n", usr->ul_tm_day);
      }

      /* daydn */
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI32, &i);      
      	 usr->dl_f_day = i;
         if (debug) printf("Today files DL: "SI32"\n", usr->dl_f_day);
      }
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI64, &l);      
      	 usr->dl_day = l * 1024;
         if (debug) printf("Today bytes DL: "SI64"\n", usr->dl_day);
      }
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI64, &l);      
      	 usr->dl_tm_day = l;
         if (debug) printf("Today time UL: "SI64"\n", usr->dl_tm_day);
      }

      /* weekup */
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI32, &i);      
      	 usr->ul_f_week = i;
         if (debug) printf("Week files UL: "SI32"\n", usr->ul_f_week);
      }
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI64, &l);      
      	 usr->ul_week = l * 1024;
         if (debug) printf("Week bytes UL: "SI64"\n", usr->ul_week);
      }
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI64, &l);      
      	 usr->ul_tm_week = l;
         if (debug) printf("Week time UL: "SI64"\n", usr->ul_tm_week);
      }

      /* weekdn */
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI32, &i);      
      	 usr->dl_f_week = i;
         if (debug) printf("Week files DL: "SI32"\n", usr->dl_f_week);
      }
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI64, &l);      
      	 usr->dl_week = l * 1024;
         if (debug) printf("Week bytes DL: "SI64"\n", usr->dl_week);
      }
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI64, &l);      
      	 usr->dl_tm_week = l;
         if (debug) printf("Week time UL: "SI64"\n", usr->dl_tm_week);
      }

      /* monthup */
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI32, &i);      
      	 usr->ul_f_month = i;
         if (debug) printf("Month files UL: "SI32"\n", usr->ul_f_month);
      }
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI64, &l);      
      	 usr->ul_month = l * 1024;
         if (debug) printf("Month bytes UL: "SI64"\n", usr->ul_month);
      }
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI64, &l);      
      	 usr->ul_tm_month = l;
         if (debug) printf("Month time UL: "SI64"\n", usr->ul_tm_month);
      }

      /* monthdn */
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI32, &i);      
      	 usr->dl_f_month = i;
         if (debug) printf("Month files DL: "SI32"\n", usr->dl_f_month);
      }
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI64, &l);      
      	 usr->dl_month = l * 1024;
         if (debug) printf("Month bytes DL: "SI64"\n", usr->dl_month);
      }

      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, SI64, &l);      
      	 usr->dl_tm_month = l;
         if (debug) printf("Month time UL: "SI64"\n", usr->dl_tm_month);
      }

      /* color */
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 if (s[0] == '1') {
            usr->class |= UF_COLOR; 
	 }
         if (debug) printf("Color: %s\n", (usr->class & UF_COLOR) ? "yes" : "no");
      }

      /* always login */
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 if (s[0] == '1') {
            usr->rights |= R_ALWAYS_LOGIN;
	 }
         if (debug) printf("Always login: %s\n", (usr->rights & R_ALWAYS_LOGIN) ? "yes" : "no");
      }

      /* nuker */
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 if (s[0] == '1') {
            usr->rights |= R_NUKE;
	 }
         if (debug) printf("Nuke: %s\n", (usr->rights & R_NUKE) ? "yes" : "no");
      }

      /* undupe */
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 if (s[0] == '1') {
            usr->rights |= R_UNDUPE;
	 }
         if (debug) printf("Undupe: %s\n", (usr->rights & R_UNDUPE) ? "yes" : "no");
      }

      /* ip masks */
      i = 0;
      while((s = (char *) strtok(NULL, "\t"))) {
         if (s[0] != '\n') {
  	    if (debug) printf("IP%d: %s\n", i, s);
            if (i < 10) get_iprange(&s, usr->idents[i]);
            i++;
	 }   
      }
            
      /* write the user record */
      fwrite(usr, sizeof(struct user_s), 1, fout);
      g_free(usr);
      if (debug) printf ("-----------------------\n");
   }
   fclose(fout);
   fclose(fin);

   
   /* group file */   
   fin = fopen("groups.txt", "rt");
   if (fin == NULL) {
      printf("Unable to open groupfile for import.\n");      
      return -1;
   }	

   /* open the output groupfile for writing */
   fout = fopen("groups", "wb");
   if (fout == NULL) {
      printf("Unable to open destination group file.\n");      
      return -1;   	
   }

   /* write the fileheader */
   g_id.count = groupcount;
   if (!fwrite(&g_id, sizeof(g_id), 1, fout)) { 
      printf("Unable to write groupfile header.\n");
      fclose(fin);
      fclose(fout);
      return -1;
   }
   /* read a line from the text groupfile which was generated by
      glconv.pl from the original glftpd files */
   while (fgets(buf, 2048, fin)) {
      grp = g_malloc(sizeof(struct group_s));
      memset(grp, 0, sizeof(struct group_s));

      /* groupname */
      s = (char *) strtok(buf, "\t");
      if (s) strncpy(grp->name, s, sizeof(grp->name) - 1);
      grp->name[strlen(grp->name)] = '\0';

      /* tagline */
      s = (char *) strtok(NULL, "\t");
      if (s) strncpy(grp->info, s, sizeof(grp->info) - 1);
      grp->info[strlen(grp->info)] = '\0';

      /* group id */
      s = (char *) strtok(NULL, "\t");
      if (s) {
      	 sscanf(s, "%d", &i);      
      	 grp->gid = i;
      }
      if (debug) printf("group processed: %04d %-10s %-32s\n", 
         grp->gid, grp->name, grp->info);

      /* write the group record */
      fwrite(grp, sizeof(struct group_s), 1, fout);
      g_free(grp);
   }
   fflush(fout);
   fclose(fin);
   fclose(fout);
   return 0;
}

/* EOF */
