#!/usr/bin/perl -w

my $standarddir   = $ENV{'F4A_STD'};

# ----------------------------------------------------------------------------
#  the directory var/log/dupe file
# ----------------------------------------------------------------------------
my $directorylog = $standarddir."var/log/directory.log";

##############################################################################
# end of configuration (don't edit below unless you know what you are doing  #
##############################################################################

use strict;
use Cwd;
use Time::Local;
use File::Basename;

use FindBin;
use lib "$FindBin::Bin/";
use primetools;

# ----------------------------------------------------------------------------
#  strdifftime($earlier, $later) - calc the diff between two times as string
# ----------------------------------------------------------------------------
sub strdifftime {
   my $time1 = $_[0]; my $time2 = $_[1];
   my $diffstr = "";  my $diff = $time2 - $time1;
   my $seconds = $diff % 60; $diff = ($diff - $seconds) / 60;
   my $minutes = $diff % 60; $diff = ($diff - $minutes) / 60;
   my $hours   = $diff % 24; $diff = ($diff - $hours) / 24;
   my $days    = $diff % 7; 
   my $weeks   = ($diff - $days) / 7;
   if ($weeks != 0) { $diffstr .= $weeks . " weeks "; }
   if ($days  != 0) { $diffstr .= $days . " days "; }
   if ($hours != 0 and $weeks == 0) { 
      $diffstr .= $hours . " hours "; 
   }
   if ($minutes != 0 and $days == 0 and $weeks == 0) { 
      $diffstr .= $minutes . " min "; 
   }
   if ($seconds != 0 and $hours == 0 and $days == 0) { 
      $diffstr .= $seconds . " sec "; 
   }
   return $diffstr;
}

sub print_header {
  my $filename = ${standarddir} . "/msg/dupe_top";
  
  open(HEADERFILE, "< $filename") or die "Unable to open headerfile $filename";

  while(<HEADERFILE>) {
    print $_;
  }

  close(HEADERFILE);
}

sub print_footer {
  my $filename = ${standarddir} . "/msg/dupe_tail";

  open(FILE, "< $filename") or die "Unable to open $filename";
   
  while(<FILE>) {
    print $_;
  }
   
  close(FILE);
}

##############################################################################
# main program                                                               #
##############################################################################

my $message;
my $nothing = 1;

my $syntax = "DUPE: search for directories in the dupelist\n" .
             "Syntax: dupe <string>\n" .
	           "Example: dupe dir_xyz\n";

if ($#ARGV != 0) { 
   print $syntax; 
   exit; 
}

my $dirname = $ARGV[0];

our %registry = read_registry(); 
if (not exists $registry{basedir}) { die "unable to find basedir\n"; }

open(DIRLOG, "< $directorylog") or exit 0;

&print_header();

while (<DIRLOG>) {
  if (/(\d\d\d\d)-(\d\d)-(\d\d) (\d\d):(\d\d):(\d\d) ([\w\-_]+) +([\w\-_]+) +(.+)/) {
    my $loggeddir = $9; my $usr = $7; my $grp = $8;
    
    my $basedir = basename($loggeddir);
    
    if ($basedir =~ /\Q$dirname\E/i) {
      $nothing = 0;

      my $dupetime = stringtotime($_);
      my $diffstr = strdifftime($dupetime, time);
      
      $message = sprintf(" !C| %-78s !C|!0\n !C|       %-78s !C|!0\n", 
      "!BDUPE: !h!H$basedir!h", "!hCreated !H$diffstr!hago by !H$usr !hof !H$grp!h.");
      printcolor($message);
    }
  }      
}

if ($nothing) {
  $message = sprintf(" !C| %-72s !C|!0\n", "!hNo dupes found.");
  printcolor($message);
}
&print_footer();

close(DIRLOG);

exit 1;

