#!/usr/bin/perl -w

##############################################################################
#                                                                            #
#  GLFTPD --> OpenFTPD Userfile converter (c) 2001 primemover                #
#                                                                            #
# ---------------------------------------------------------------------------#
#                                                                            #
#  This program is free software; you can redistribute it and/or modify      #
#  it under the terms of the GNU General Public License as published by      #
#  the Free Software Foundation; either version 2 of the License, or         #
#  (at your option) any later version.                                       #
#                                                                            #
##############################################################################

use strict;
my $debug = 1;

print "GlFTPD -> OpenFTPD userfile converter (c) 2001 primemover\n";

if ($#ARGV < 1) {
   print "Syntax: glconv <password file> <group file> <user directory>\n\n";
   print "        password file  ... path to the glftpd passwd file\n";
   print "                           usually in /glftpd/etc/passwd)\n\n";
   print "        group file     ... path to the glftpd group file\n";
   print "                           usually in /glftpd/etc/group\n\n";
   print "        user directory ... path to the glftpd user directory\n";
   print "                           usually in /glftpd/ftp-date/users\n";
   exit;
}

if (not -e "glconv") {
   print "Error: glconv binary was not found, please compile it first!\n";
   exit;
}

# open the glftpd user file ...
my @users;
if (-e "$ARGV[0]") {
   open(INPUT, "$ARGV[0]");
   while(my $line = <INPUT>) { push(@users, $line); }
   close(INPUT);
}
else {
   print "Unable to open glftpd passwd file ($ARGV[0]).\n";
   exit;
}

# open the glftpd group file ...
my @groups;
if (-e "$ARGV[1]") {
   open(INPUT, "$ARGV[1]");
   while(my $line = <INPUT>) { push(@groups, $line); }
   close(INPUT);
}
else {
   print "Unable to open glftpd group file ($ARGV[1]).\n";
   exit;
}

# check for the user directory
if (not -d "$ARGV[2]") {
   print "Unable to find glftpd user directory ($ARGV[2]).\n";
   exit;
}


# for each single user ...
print "converting users:\n";
my $usercount = 0;
open(OUTPUT, "> users.txt");
foreach my $user (@users) {
my ($name, $pass, $uid, $gid, $class, $ratio, $credits, $logins,
    $tagline, $maxulspeed, $maxdlspeed, $allup_files, $allup_bytes, 
    $allup_seconds, $alldn_files, $alldn_bytes, $alldn_seconds, $dayup_files, 
    $dayup_bytes, $dayup_seconds, $daydn_files, $daydn_bytes, $daydn_seconds,
    $wkup_files, $wkup_bytes, $wkup_seconds, $wkdn_files, $wkdn_bytes,
    $wkdn_seconds, $monthup_files, $monthup_bytes, $monthup_seconds, 
    $monthdn_files, $monthdn_bytes, $monthdn_seconds, $color, $exempt, 
    $nuke, $undupe, $alotment,$flags, $nuke_times,$nuke_mb, $login_times,
    $last_on,$time_limit,$time_on_today) ;

	$name = "";
	$pass = "";
	$uid  = 0;
	$gid  = 0;
	$class = 0;
	$ratio = 0;
	$credits = 0;
	$logins = 0;
	$tagline = "";
	$maxulspeed = 0;
	$maxdlspeed = 0;
	$allup_files = 0;
	$allup_bytes = 0;
	$allup_seconds = 0;
	$alldn_files = 0;
	$alldn_bytes = 0;
	$alldn_seconds = 0;
	$dayup_files = 0;
	$dayup_bytes = 0;
	$dayup_seconds = 0;
	$daydn_files = 0;
	$daydn_bytes = 0;
	$daydn_seconds = 0;
        $wkup_files = 0;
        $wkup_bytes = 0;
        $wkup_seconds = 0;
        $wkdn_files = 0;
        $wkdn_bytes = 0;
        $wkdn_seconds = 0;
        $monthup_files = 0;
        $monthup_bytes = 0;
        $monthup_seconds = 0;
        $monthdn_files = 0;
        $monthdn_bytes = 0;
        $monthdn_seconds = 0;
	$color = 0;
	$exempt = 0;
  	$nuke = 0; 
	$undupe = 0; 
	$alotment = 0;
	$flags = 0;
	$nuke_times = 0;
	$nuke_mb = 0;
	$login_times = 0;
	$last_on = 0;
	$time_limit = 0;
	$time_on_today = 0;

   if ($user =~ /.+:.+:\d+:\d+:.+:.+:.+/) {
      my ($name, $pass, $uid, $gid, $createdate) = (split(/:/, $user))[0..4];
      print "processing (", $usercount + 1, ") $name ...\n";
      if ($debug) {
	 print "Username     : $name\n";
	 print "Password     : $pass\n";
	 print "User ID      : $uid\n";
	 print "Group ID     : $gid\n";
	 print "Created on   : $createdate\n";
      }      
      # search for the user record file ...
      if (-e "$ARGV[2]/$name") {
         
	 # init some stuff
	 my @lines   = ();
         my @ipmasks = ();
         my $tagline = "No tagline";
	 my $class   = "n";

         open(INPUT, "$ARGV[2]/$name");
         while(my $line = <INPUT>) { push(@lines, $line); }
        close(INPUT);        
         # print @lines;
         foreach my $line (@lines) {
            # general 
            if ($line =~ /^GENERAL\s+([\d,]+)\s+(\d+)\s+(\d+)\s+(\d+)/) {
               $alotment = $1;
               $logins = $2;
               $maxdlspeed = $3;
               $maxulspeed = $4;
               if ($debug) {
        	  print "Alotment     : $alotment\n";
        	  print "Logins       : $logins\n";
        	  print "Max dl speed : $maxdlspeed\n";
        	  print "Max ul speed : $maxulspeed\n";
               }
            }
            # ratio
            if ($line =~ /^RATIO\s+(\d+)/) {
               $ratio = $1;
               if ($debug) { print "Ratio        : $ratio\n"; }
            }
            # flags
            if ($line =~ /^FLAGS\s+(.+)/) {
               $flags = $1;
               if ($debug) { print "Flags        : $flags\n"; }
               if ($flags =~ /1/) { $class = "s"; }
               if ($flags =~ /2/) { $class = "g"; }
               if ($flags =~ /7/) { $class = "o"; }
               if ($flags =~ /8/) { $class = "a"; }
               if ($debug) { print "==> Class    : $class\n"; }
               
	       $color = 0; 
	       if ($flags =~ /5/) { 
	          $color = 1; 
	       }
               
	       $exempt = 0; 
	       if ($flags =~ /4/) { 
	          $exempt = 1; 
	       }
               
	       $nuke = 0; 
	       if ($flags =~ /A/) { 
	          $nuke = 1; 
	       }
               
	       $undupe = 0; 
	       if ($flags =~ /C/) { 
	          $undupe = 1; 
	       }
            }

            # credits
            if ($line =~ /^CREDITS\s+(\d+)/) {
               $credits = $1;
               if ($debug) { print "Credits      : $credits\n"; }
            }
            # tagline
            if ($line =~ /^(?:TAGLINE|UNFO)\s+(.+)/) {
               $tagline = $1;
               if ($debug) { print "Tagline      : $tagline\n"; }
            }
            # alltime uploads
            if ($line =~ /^ALLUP\s+(\d+)\s+(\d+)\s+(\d+)/) {
               $allup_files   = $1;
               $allup_bytes   = $2;
               $allup_seconds = $3;
               if ($debug) { print "allup        : ", 
	          $allup_bytes / 0x100000, " MB in $allup_files files\n"; }
            }
            # alltime downloads
            if ($line =~ /^ALLDN\s+(\d+)\s+(\d+)\s+(\d+)/) {
               $alldn_files   = $1;
               $alldn_bytes   = $2;
               $alldn_seconds = $3;
               if ($debug) { print "alldn        : ", $alldn_bytes / 0x100000, 
	          " MB in $alldn_files files\n"; }
            }
            # week uploads
            if ($line =~ /^WKUP\s+(\d+)\s+(\d+)\s+(\d+)/) {
               $wkup_files   = $1;
               $wkup_bytes   = $2;
               $wkup_seconds = $3;
               if ($debug) { print "wkup         : ", $wkup_bytes / 0x100000, 
	          " MB in $wkup_files files\n"; }
            }
            # week downloads
            if ($line =~ /^WKDN\s+(\d+)\s+(\d+)\s+(\d+)/) {
               $wkdn_files   = $1;
               $wkdn_bytes   = $2;
               $wkdn_seconds = $3;
               if ($debug) { print "wkdn         : ", $wkdn_bytes / 0x100000, 
	          " MB in $wkdn_files files\n"; }
            }
            # month uploads
            if ($line =~ /^MONTHUP\s+(\d+)\s+(\d+)\s+(\d+)/) {
               $monthup_files   = $1;
               $monthup_bytes   = $2;
               $monthup_seconds = $3;
               if ($debug) { print "monthup      : ", $monthup_bytes / 0x100000, 
	          " MB in $monthup_files files\n"; }
            }
            # month downloads
            if ($line =~ /^MONTHDN\s+(\d+)\s+(\d+)\s+(\d+)/) {
               $monthdn_files   = $1;
               $monthdn_bytes   = $2;
               $monthdn_seconds = $3;
               if ($debug) { print "monthdn      : ", $monthdn_bytes / 0x100000, 
	          " MB in $monthdn_files files\n"; }
            }
            # day uploads
            if ($line =~ /^DAYUP\s+(\d+)\s+(\d+)\s+(\d+)/) {
               $dayup_files   = $1;
               $dayup_bytes   = $2;
               $dayup_seconds = $3;
               if ($debug) { print "dayup         : ", $dayup_bytes / 0x100000, 
	          " MB in $dayup_files files\n"; }
            }
            # day downloads
            if ($line =~ /^DAYDN\s+(\d+)\s+(\d+)\s+(\d+)/) {
               $daydn_files   = $1;
               $daydn_bytes   = $2;
               $daydn_seconds = $3;
               if ($debug) { print "daydn         : ", $daydn_bytes / 0x100000, 
	          " MB in $daydn_files files\n"; }
            }
            # nukes
            if ($line =~ /^NUKE\s+(\d+)\s+(\d+)\s+(\d+)/) {
            #   $nuke_last   = $1; # This was not in use?
               $nuke_times   = $2;
               $nuke_mb = $3;
               if ($debug) { print "nuke         : total nukes $nuke_times ($nuke_mb MB)\n"; }
            }
            # times
            if ($line =~ /^TIME\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)/) {
               $login_times = $1;
               $last_on = $2;
               $time_limit = $3;
               $time_on_today = $4;
               if ($debug) {
        	  print "Login times  : $login_times\n";
        	  print "Last on      : $last_on\n";
        	  print "Time limit   : $time_limit\n";
        	  print "Time on today: $time_on_today\n";
               }
            }

            # ip's
	    if ($line =~ /^IP\s+(.+)/) {
	       push(@ipmasks, $1);
	    }	    
         }
         if ($debug) { print "IP:", @ipmasks,"\n"; }
         # dump in tab seperated format, each line one user

         print OUTPUT $name, "\t", 
               $pass, "\t", 
               $uid, "\t", 
               $gid, "\t", 
               $class, "\t",
               $ratio ? $ratio : 0, "\t", 
               $credits, "\t", 
               $logins, "\t",
               $tagline, "\t",
               $maxulspeed, "\t",
               $maxdlspeed, "\t",
               $allup_files,   "\t",
               $allup_bytes,   "\t",
               $allup_seconds, "\t",
               $alldn_files,   "\t",
               $alldn_bytes,   "\t",
               $alldn_seconds, "\t",
               $dayup_files,   "\t",
               $dayup_bytes,   "\t",
               $dayup_seconds, "\t",
               $daydn_files,   "\t",
               $daydn_bytes,   "\t",
               $daydn_seconds, "\t",
               $wkup_files,   "\t",
               $wkup_bytes,   "\t",
               $wkup_seconds, "\t",
               $wkdn_files,   "\t",
               $wkdn_bytes,   "\t",
               $wkdn_seconds, "\t",
               $monthup_files,   "\t",
               $monthup_bytes,   "\t",
               $monthup_seconds, "\t",
               $monthdn_files,   "\t",
               $monthdn_bytes,   "\t",
               $monthdn_seconds, "\t",
               $color, "\t",
	       $exempt, "\t",
               $nuke, "\t",
	       $undupe, "\t";
               foreach my $ipmask (@ipmasks) {
		   # This will convert those nasty *@10.* ips to
		   # real ones like *@10.*.*.*
		   $ipmask =~ s/(.+@\*$)/$1\.\*\.\*\.\*/;
		   $ipmask =~ s/(.+@\d{1,3}\.\*$)/$1\.*\.*/g;
		   $ipmask =~ s/(.+@\d{1,3}\.\d{1,3}\.\*$)/$1\.*/g;
		   print OUTPUT "$ipmask\t";
	       }
	       print OUTPUT "\n";
	       
         $usercount++;
      }
   }   
}
close(OUTPUT);

print "converting groups:\n";
my $groupcount = 0;
open(OUTPUT, "> groups.txt");
foreach my $group (@groups) {
   my $tagline = "No Tagline";
   if ($group =~ /^(.+):(.*):(\d+):\d*/) {
      $group = $1;
      print "processing (", $groupcount, ") $group ...\n";
      if ($2) { $tagline = $2; }
      my $gid = $3;
      print OUTPUT $group, "\t", $tagline, "\t", $gid, "\n";
      $groupcount++;
   }
   else { printf "WRONG SYNTAX: $group"; }
}
close(OUTPUT);

# run glconv with number of users and path to group file as arguments
my @args = ("./glconv", $usercount, $groupcount);
my $rc = system(@args);

if ($rc == 0) {
   print("$usercount users and $groupcount groups successfully converted.\n")
}

if ($debug == 0) {
   unlink 'users.txt', 'groups.txt';
}

my $key = 0;
while ($key !~ /^(y|n)$/) {
   print "We will now install the converted userbase, please shutdown\n",
         "OpenFTPD if it is currently running.\n\n";
   print ">> Do you want to overwrite you current userbase ? (y/n): ";
   $key = <STDIN>; chomp($key);
}
if ($key eq "y") {
   `mv -f users ../etc`;
   `mv -f groups ../etc`;
}

print "TIP: To create an initial root superuser, start ftpd with -s\n";
print "done.\n";

# EOF
